@@ -19,7 +19,8 @@ import subprocess
 import sys
 import traceback
 from pathlib import Path
-from typing import Optional, List
+from typing import Any, Optional, List, Union, TYPE_CHECKING
+from typing_extensions import Protocol
 
 import click
 import rich
@@ -27,12 +28,16 @@ import rich
 from . import DefaultNames, __version__, get_home_config_path, is_version_unreleased
 from . import init as initmod
 from . import logs, version_tuple
-from .job import start_job
+from .job import start_job, Job
 from .localenv import LocalEnv, Project
 from .logs import Levels
 from .support import Status
 from .util import filter_env, get_package_digest
 
+if TYPE_CHECKING:
+    from repo import RepoView
+    from yamlmanifest import YamlManifest
+
 _latestJobs = []  # for testing
 _args: List[str] = []  # for testing
 
@@ -170,7 +175,7 @@ readonlyJobControlOptions = option_group(
         "--commit",
         default=False,
         is_flag=True,
-        help="Commit modified files to the instance repository. (Default: false)",
+        help="Commit modified files to the ensemble repository. (Default: false)",
     ),
     click.option(
         "--dirty",
@@ -220,6 +225,13 @@ commonJobFilterOptions = option_group(
         default=None,
         help="Run this job in the given environment.",
     ),
+    click.option(
+        "--var",
+        nargs=2,
+        type=click.Tuple([str, str]),
+        multiple=True,
+        help="name/value pair to pass to job (multiple times ok).",
+    ),
 )
 destroyUnmanagedOption = click.option(
     "--destroyunmanaged",
@@ -306,7 +318,7 @@ def _remote_cmd(runtime_, cmd_line, local_env, version_check):
 
     if envvar_filter:
         addOnly = kind == "docker"
-        env = filter_env(local_env.map_value(envvar_filter), addOnly=addOnly)
+        env = filter_env(local_env.map_value(envvar_filter, None), addOnly=addOnly)
     else:
         env = None
 
@@ -434,11 +446,12 @@ def _run_remote(runtime, options, localEnv):
         sys.exit(rv)
 
 
-def _print_summary(job, options):
-    jsonSummary = {}
+def _print_summary(job: "Job", options) -> str:
+    jsonSummary: Any = {}
+    text = ""
     summary = options.get("output")
     if summary == "text" and not job.jobOptions.planOnly:
-        job.print_summary_table()
+        text = job.print_summary_table()
     elif summary == "json":
         if job.jobOptions.planOnly:
             jsonSummary = job._json_plan_summary()
@@ -458,7 +471,9 @@ def _print_summary(job, options):
             else:
                 click.echo(result)
     if jsonSummary:
-        click.echo(json.dumps(jsonSummary, indent=2))
+        text = json.dumps(jsonSummary, indent=2)
+        click.echo(text)
+    return text
 
 
 def yesno(prompt):
@@ -471,7 +486,10 @@ def yesno(prompt):
         return False
 
 
-def _stop_logging(job, options, verbose, tmplogfile):
+def _stop_logging(job, options, verbose, tmplogfile, summary):
+    if summary:
+        with open(tmplogfile, "a") as f:
+            f.write(summary)
     if job and job.log_path:
         log_path = job.log_path
         dir = os.path.dirname(log_path)
@@ -495,6 +513,7 @@ def _run_local(ensemble: str, options):
     tmplogfile = options["logfile"]
     job, rendered, proceed = start_job(ensemble, options)
     _latestJobs.append(job)  # testing only
+    summary = ""
     if job:
         declined = False
         if not job.unexpectedAbort and not job.planOnly and proceed:
@@ -507,11 +526,11 @@ def _run_local(ensemble: str, options):
             if verbose > 0:
                 raise job.unexpectedAbort
         elif not declined:
-            _print_summary(job, options)
+            summary = _print_summary(job, options)
     else:
         click.echo("Unable to create job")
 
-    _stop_logging(job, options, verbose, tmplogfile)
+    _stop_logging(job, options, verbose, tmplogfile, summary)
     return _exit(job, options)
 
 
@@ -720,6 +739,13 @@ def plan(ctx, ensemble=None, **options):
     "--use-deployment-blueprint",
     help="Use this deployment blueprint.",
 )
+@click.option(
+    "--var",
+    nargs=2,
+    type=click.Tuple([str, str]),
+    multiple=True,
+    help="name/value pair to pass to skeleton (multiple times ok).",
+)
 def init(ctx, projectdir, ensemble_name=None, **options):
     """
     Create a new project or, if [project_dir] exists or is inside a project, create a new ensemble.
@@ -857,7 +883,7 @@ def runtime(ctx, project_folder, init=False, update=False, **options):
 
         packages = [
             req
-            for req in importlib.metadata.requires("unfurl")
+            for req in importlib.metadata.requires("unfurl")  # type: ignore
             if "extra ==" not in req
         ]
         error = subprocess.run(["pipenv", "install"] + packages, env=env).returncode
@@ -955,7 +981,7 @@ def git(ctx, gitargs, dir="."):
             for repo in localEnv.get_manifest().repositories.values()
             if repo.repo
         }
-    elif localEnv.project.project_repoview.repo:
+    elif localEnv.project and localEnv.project.project_repoview.repo:
         repo = localEnv.project.project_repoview.repo
         repos = {os.path.relpath(repo.working_dir, os.path.abspath(dir)): repo}
     else:
@@ -973,22 +999,32 @@ def git(ctx, gitargs, dir="."):
         _status, stdout, stderr = repo.run_cmd(gitargs)
         click.echo(stdout + "\n")
         if stderr.strip():
-            click.echo(stderr + "\n", color="red")
+            click.secho(stderr + "\n", fg="red")
         if _status != 0:
             status = _status
 
     return status
 
 
-def get_commit_message(committer):
+class Committer(Protocol):
+    def commit(self, msg: str, add_all: bool = False) -> int:
+        ...
+
+    def add_all(self) -> None:
+        ...
+
+    def get_repo_status(self, dirty=False) -> str:
+        ...
+
+
+def get_commit_message(committer, default_message):
     statuses = committer.get_repo_status(True)
     if not statuses:
         click.echo("Nothing to commit!")
         return None
-    defaultMsg = committer.get_default_commit_message()
     MARKER = "# Everything below is ignored\n"
     statusMsg = "# " + "\n# ".join(statuses.rstrip().splitlines())
-    message = click.edit(defaultMsg + "\n\n" + MARKER + statusMsg)
+    message = click.edit(default_message + "\n\n" + MARKER + statusMsg)
     if message is not None:
         return message.split(MARKER, 1)[0].rstrip("\n")
     else:
@@ -1012,12 +1048,26 @@ def get_commit_message(committer):
     is_flag=True,
     help="Don't add files for committing (user must add using git)",
 )
+@click.option(
+    "--all-repositories",
+    default=False,
+    is_flag=True,
+    help="Commit all repositories the ensemble accesses",
+)
 @click.option(
     "--use-environment",
     default=None,
     help="Use this environment.",
 )
-def commit(ctx, project_or_ensemble_path, message, skip_add, no_edit, **options):
+def commit(
+    ctx,
+    project_or_ensemble_path,
+    message,
+    skip_add,
+    no_edit,
+    all_repositories,
+    **options,
+):
     """Commit any changes to the given project or ensemble."""
     options.update(ctx.obj)
     localEnv = LocalEnv(
@@ -1026,20 +1076,36 @@ def commit(ctx, project_or_ensemble_path, message, skip_add, no_edit, **options)
         can_be_empty=True,
         override_context=options.get("use_environment") or "",
     )
-    if localEnv.manifestPath:
-        committer = localEnv.get_manifest()
+    if localEnv.manifestPath and len(os.path.abspath(project_or_ensemble_path)) >= len(
+        localEnv.manifestPath
+    ):
+        ensemble = localEnv.get_manifest()
+        default_commit_message = ensemble.get_default_commit_message()
+        if all_repositories:
+            committer: Committer = ensemble
+        else:
+            committer = ensemble.repositories["self"]
     else:
-        committer = localEnv.project.project_repoview
+        default_commit_message = "Commit by Unfurl"
+        # otherwise commit the whole project
+        if all_repositories:
+            click.echo("aborting: --all-repositories requires an ensemble path")
+            return
+        else:
+            assert localEnv.project
+            committer = localEnv.project.project_repoview
 
     if not skip_add:
         committer.add_all()
+
     if not message:
         if no_edit:
-            message = committer.get_default_commit_message()
+            message = default_commit_message
         else:
-            message = get_commit_message(committer)
+            message = get_commit_message(committer, default_commit_message)
             if not message:
                 return  # aborted
+
     committed = committer.commit(message, False)
     click.echo(f"committed to {committed} repositories")
 
@@ -1068,8 +1134,9 @@ def git_status(ctx, project_or_ensemble_path, dirty, **options):
         override_context=options.get("use_environment") or "",
     )
     if localEnv.manifestPath:
-        committer = localEnv.get_manifest()
+        committer: Union["YamlManifest", "RepoView"] = localEnv.get_manifest()
     else:
+        assert localEnv.project
         committer = localEnv.project.project_repoview
     statuses = committer.get_repo_status(dirty)
     if not statuses:
@@ -1105,6 +1172,7 @@ def export(ctx, project_or_ensemble_path, format, file, **options):
         project_or_ensemble_path,
         options.get("home"),
         override_context=options.get("use_environment") or "",
+        readonly=True,
     )
     exporter = getattr(to_json, "to_" + format)
     jsonSummary = exporter(localEnv, file)
@@ -1135,12 +1203,17 @@ def status(ctx, ensemble, **options):
         ensemble,
         options.get("home"),
         override_context=options.get("use_environment") or "",
+        readonly=True,
     )
+    logger = logging.getLogger("unfurl")
     manifest = localEnv.get_manifest()
-    click.echo(manifest.status_summary())
+    summary = manifest.status_summary()
+    click.echo(summary)
+    logger.debug("Status summary:\n%s", summary)
     query = options.get("query")
     if query:
         trace = options.get("trace")
+        assert manifest.rootResource
         result = manifest.rootResource.query(query, trace=trace)
         click.echo("query: " + query)
         if result is None:
@@ -1182,8 +1255,10 @@ def version(ctx, semver=False, remote=False, **options):
         click.echo(f"unfurl version {__version__(True)} ({get_package_digest()})")
 
     if remote and not options.get("no_runtime"):
-        click.echo("Remote:")
-        _run(get_home_config_path(options.get("home")), options)
+        home_path = get_home_config_path(options.get("home"))
+        if home_path:
+            click.echo("Remote:")
+            _run(home_path, options)
 
 
 @cli.command()
@@ -1205,7 +1280,12 @@ def help(ctx, cmd=""):
 
 @cli.command()
 @click.pass_context
-@click.argument("project_or_ensemble_path", default=".", type=click.Path(exists=True))
+@click.argument(
+    "project_or_ensemble_path",
+    default=".",
+    envvar="UNFURL_SERVE_PATH",
+    type=click.Path(exists=True),
+)
 @click.option("--port", default=8081, help="Port to listen on")
 @click.option("--address", default="localhost", help="Host to listen on")
 @click.option(
@@ -1217,13 +1297,26 @@ def help(ctx, cmd=""):
     "--clone-root",
     default=".",
     help="Where to clone all repositories",
+    envvar="UNFURL_CLONE_ROOT",
     type=click.Path(exists=True),
 )
-def serve(ctx, port, address, secret, clone_root, project_or_ensemble_path, **options):
+@click.option(
+    "--cors",
+    envvar="UNFURL_SERVE_CORS",
+    help='enable CORS with origin (e.g. "*")',
+)
+def serve(
+    ctx, port, address, secret, clone_root, project_or_ensemble_path, cors, **options
+):
     options.update(ctx.obj)
-    from .server import serve
+    # env vars need to set before importing serve
+    if cors:
+        os.environ["UNFURL_SERVE_CORS"] = cors
+    os.environ["UNFURL_SERVE_PATH"] = project_or_ensemble_path
+    os.environ["UNFURL_CLONE_ROOT"] = clone_root
+    from .server import serve as _serve
 
-    serve(address, port, secret, clone_root, project_or_ensemble_path, options)
+    _serve(address, port, secret, clone_root, project_or_ensemble_path, options)
 
 
 def main():
