@@ -18,12 +18,12 @@ import six
 from collections.abc import Mapping
 import os
 import copy
-
-from unfurl.logs import UnfurlLogger, Levels, LogExtraLevels
+from toscaparser.properties import Property
+from .logs import UnfurlLogger, Levels, LogExtraLevels
 
 if TYPE_CHECKING:
-    from unfurl.manifest import Manifest
-    from unfurl.job import ConfigTask
+    from .manifest import Manifest, ChangeRecordRecord
+    from .job import ConfigTask
 
 
 from .support import Status, ResourceChanges, Priority, set_context_vars
@@ -140,6 +140,8 @@ class Configurator:
 
     exclude_from_digest: Tuple[str, ...] = ()
 
+    attribute_output_metadata_key: Optional[str] = None
+
     @classmethod
     def set_config_spec_args(klass, kw: dict, target: None) -> dict:
         return kw
@@ -150,7 +152,7 @@ class Configurator:
     def get_generator(self, task: "TaskView") -> Generator:
         return self.run(task)
 
-    def render(self, task: "TaskView") -> None:
+    def render(self, task: "TaskView") -> Any:
         """
         This method is called is called during the planning phase to give the configurator an
         opportunity to do early validation and error detection and generate any plan information or configuration files that the user may want to review before the running the deployment task.
@@ -204,7 +206,7 @@ class Configurator:
         """
         return True
 
-    def should_run(self, task: "TaskView") -> bool:
+    def should_run(self, task: "TaskView") -> Union[bool, Priority]:
         """Does this configuration need to be run?"""
         return self.configSpec.should_run()
 
@@ -255,7 +257,7 @@ class Configurator:
         )
         return digest
 
-    def check_digest(self, task: "TaskView", changeset: ChangeRecord) -> bool:
+    def check_digest(self, task: "TaskView", changeset: "ChangeRecordRecord") -> bool:
         """
         Examine the previous :class:`ChangeRecord` generated by the previous time this operation
         was performed on the target instance and return whether it should be rerun or not.
@@ -308,7 +310,7 @@ class Configurator:
                 results.append(task.inputs._getresult(key))
 
         newDigest = get_digest(results, manifest=task._manifest)
-        # I can't find somewhere where digestValue is defined on ChangeRecord, but I'll assume the use is correct
+        # note: digestValue attribute is set in Manifest.load_config_change
         mismatch = changeset.digestValue != newDigest  # type: ignore
         if mismatch:
             task.logger.verbose(
@@ -340,6 +342,9 @@ class _ConnectionsMap(dict):
         }
         return by_type.values()
 
+    def copy(self):
+        return self
+
     def __missing__(self, key: object) -> object:
         # the more specific connections are inserted first so this should find
         # the most relevant connection of the given type
@@ -365,10 +370,9 @@ class TaskLoggerAdapter(logging.LoggerAdapter, LogExtraLevels):
         # we override this instead of process() because we need to change the log level
         # note that `self.extra` does NOT replace kwargs['extra']
         if self.isEnabledFor(level):
-            # task for {resource name}: {operation} (reason: {resource})
-            # e.g. Task for test-node: configure (reason: missing)
-            task: TaskView = self.extra
-            task_id = f"for {task.target.name}: {task.configSpec.operation}"
+            task = cast(TaskView, self.extra)
+            # e.g. Task configure for test-node (reason: missing)
+            task_id = f"{task.configSpec.operation} for {task.target.name}"
             if task.reason:
                 task_id += f" (reason: {task.reason})"
             if task._rendering:
@@ -376,10 +380,12 @@ class TaskLoggerAdapter(logging.LoggerAdapter, LogExtraLevels):
                 if level >= Levels.VERBOSE:
                     level = Levels.VERBOSE
             else:
-                msg = f"Running task {task_id}: {msg}"
+                msg = f"Task {task_id}: {msg}"
             self.logger.log(level, msg, *args, **kwargs)
 
 
+_initializing_environ = object()
+
 class TaskView:
     """The interface presented to configurators.
 
@@ -414,7 +420,6 @@ class TaskView:
             UnfurlTaskError
         ] = []  # UnfurlTaskError objects appends themselves to this list
         self._inputs: Optional[ResultsMap] = None
-        self._environ = None
         self._manifest = manifest
         self.messages: List[object] = []
         self._addedResources: List[NodeInstance] = []
@@ -423,9 +428,44 @@ class TaskView:
         self._resourceChanges = ResourceChanges()
         self._workFolders: dict = {}
         self._rendering = False
+        self._environ: object = None
         # public:
         self.operation_host = find_operation_host(target, configSpec.operation_host)
 
+    @property
+    def environ(self) -> Dict[str, str]:
+        if self._inputs is None or self._environ is _initializing_environ:
+            # not ready yet
+            return self.target.environ
+        if self._environ is None:
+            self._environ = _initializing_environ
+            self._environ = self.get_environment(False)
+        return cast(Dict[str, str], self._environ)
+
+    def set_envvars(self):
+        """
+        Update os.environ with the task's environ and save the current one so it can be restored by ``restore_envvars``
+        """
+        # self.logger.trace("update os.environ with %s", self.environ)
+        for key in os.environ:
+            current = self.environ.get(key)
+            if current is None:
+                del os.environ[key]
+        for key, value in self.environ.items():
+            if value is not None:
+                os.environ[key] = str(value)
+
+    def restore_envvars(self):
+        # restore the environ set on root resource
+        # self.logger.trace("restoring os.environ with %s", self.target.root.environ)
+        for key in os.environ:
+            current = self.target.root.environ.get(key)
+            if current is None:
+                del os.environ[key]
+        for key, value in self.target.environ.items():
+            if value is not None:
+                os.environ[key] = str(value)
+
     @property
     def inputs(self) -> ResultsMap:
         """
@@ -441,8 +481,7 @@ class TaskView:
             assert self._attributeManager  # type: ignore
             assert self.target.root.attributeManager is self._attributeManager  # type: ignore
             ctx = RefContext(self.target, task=self)
-            # deepcopy because ResultsMap might modify interior maps and lists
-            inputs = copy.deepcopy(self.configSpec.inputs)
+            inputs = self.configSpec.inputs
             relationship = None
             if isinstance(self.target, RelationshipInstance):
                 relationship = self.target
@@ -455,7 +494,6 @@ class TaskView:
             HOST = (target.parent or target).attributes
             ORCHESTRATOR = target.root.find_instance_or_external("localhost")
             vars = dict(
-                inputs=inputs,
                 task=self.get_settings(),
                 connections=self._get_connections(),
                 SELF=self.target.attributes,
@@ -477,6 +515,7 @@ class TaskView:
             elif self.configSpec.base_dir:
                 ctx.base_dir = self.configSpec.base_dir
             self._inputs = ResultsMap(inputs, ctx)
+            vars["inputs"] = self._inputs
         return self._inputs
 
     @property
@@ -545,7 +584,7 @@ class TaskView:
 
         return env
 
-    def get_environment(self, addOnly: bool) -> dict:
+    def get_environment(self, addOnly: bool, env: Optional[dict] = None) -> dict:
         """Return a dictionary of environment variables applicable to this task.
 
         Args:
@@ -560,22 +599,21 @@ class TaskView:
         2. Variables set by the connections that are available to this operation.
         3. Variables declared in the operation's ``environment`` section.
         """
+        if env is None:
+            env = self.target.environ
 
-        env = os.environ.copy()
         # build rules by order of preference (last overrides first):
         # 1. ensemble's environment
         # 2. variables set by connections
         # 3. operation's environment
 
-        # we use merge.copy() to preserve basedir
-        rules = merge.copy(self.target.root.envRules)
-
-        rules.update(self._find_relationship_env_vars())
-
+        rules = self._find_relationship_env_vars()
         if self.configSpec.environment:
             rules.update(self.configSpec.environment)
 
         # apply rules
+        # NB: Context.environ() calls TaskView.environ() which calls this method during initiation
+        # If an expression being evaluated below invokes Context.environ a task error will be raised
         rules = serialize_value(
             map_value(rules, self.inputs.context), resolveExternal=True
         )
@@ -605,7 +643,7 @@ class TaskView:
                     [
                         r.tosca_id
                         for r in self.target.source.get_requirements(
-                            self.target.requirement.template.name
+                            self.target.template.name
                         )
                     ]
                 )
@@ -680,7 +718,8 @@ class TaskView:
     #     outputs to share operation outputs so far
     #     """
 
-    def _set_outputs(self, outputs):
+    def _set_outputs(self, outputs: Dict[str, Any]) -> None:
+        metadata_key = self.configurator.attribute_output_metadata_key  # type: ignore
         for key, value in outputs.items():
             mapping = self.configSpec.outputs.get(key)
             if mapping:
@@ -700,6 +739,10 @@ class TaskView:
                     )
                 else:
                     self.target.attributes[mapping] = value
+            elif metadata_key:
+                attr_def = self.target.template.attributeDefs.get(key)
+                if attr_def and attr_def.schema.get("metadata", {}).get(metadata_key):
+                    self.target.attributes[key] = value      
 
     def done(
         self,
@@ -759,7 +802,7 @@ class TaskView:
     # other configurations maybe modify those changes, triggering a configuration change
     def query(
         self,
-        query: str,
+        query: Union[str, dict],
         dependency: bool = False,
         name: Optional[str] = None,
         required: bool = False,
@@ -774,7 +817,7 @@ class TaskView:
             result = Ref(query, vars=vars).resolve(
                 self.inputs.context, wantList, strict
             )
-        except:
+        except Exception:
             if not throw:
                 UnfurlTaskError(
                     self, f"error while evaluating query: {query}", logging.WARNING
@@ -796,7 +839,7 @@ class TaskView:
         name: Optional[str] = None,
         required: bool = True,
         wantList: bool = False,
-        target: Optional[NodeInstance] = None,
+        target: Optional[EntityInstance] = None,
     ) -> "Dependency":
         getter = getattr(expr, "as_ref", None)
         if getter:
@@ -832,7 +875,7 @@ class TaskView:
 
     def create_sub_task(
         self,
-        operation: Optional[str] = None,
+        operation: Optional[Union[str, ConfigurationSpec]] = None,
         resource: Optional[EntityInstance] = None,
         inputs: Optional[dict] = None,
         persist: bool = False,
@@ -855,7 +898,7 @@ class TaskView:
 
         if not operation:
             operation = f"{self.configSpec.interface}.{self.configSpec.operation}"
-        if isinstance(operation, six.string_types):
+        if isinstance(operation, str):
             taskRequest = create_task_request(
                 self.job.jobOptions,  # type: ignore
                 operation,
@@ -918,6 +961,23 @@ class TaskView:
                     truncate_str(value),
                 )
             updated = True
+
+        template = resourceSpec.get("template")
+        if template:
+            assert self._manifest.tosca
+            if isinstance(template, dict):
+                tname = existingResource.template.name
+                existingResource.template = self._manifest.tosca.add_node_template(tname, template)
+            elif isinstance(template, str) and template != existingResource.template.name:
+                nodeSpec = self._manifest.tosca.nodeTemplates.get(template)
+                if not nodeSpec:
+                    raise UnfurlTaskError(
+                        self,
+                        f"couldn't update TOSCA template for {existingResource.name}: '{template}' is not defined",
+                    )
+                existingResource.template = nodeSpec
+            updated = True
+
         return updated
 
     def _parse_instances_tpl(
@@ -1110,10 +1170,20 @@ class Dependency(Operational):
         self.expected = expected
         self.schema = schema
         self._required = required
-        self.name = name or expr
+        if name:
+            self.name = name
+        elif isinstance(expr, str):
+            self.name = str(expr)
+        elif target:
+            self.name = f"{target.name}:{dict(expr)}"
+        else:
+            self.name = f"{dict(expr)}"
         self.wantList = wantList
         self.target = target
 
+    def __repr__(self) -> str: 
+        return f"Dependency('{self.name}')"
+
     @property
     def local_status(self) -> Optional[Status]:
         if (
@@ -1131,6 +1201,13 @@ class Dependency(Operational):
     def priority(self) -> Priority:
         return Priority.required if self._required else Priority.optional
 
+    def _is_unexpected(self) -> bool:
+        if self.target is None:
+            return True
+        wantList: Union[str, bool] = "result" if isinstance(self.expected, Result) else self.wantList
+        result = Ref(self.expr).resolve(RefContext(self.target), wantList=wantList)
+        return result != self.expected
+
     def refresh(self, config: "ConfigTask") -> None:
         if self.expected is not None:
             changeId = config.changeId
@@ -1148,7 +1225,8 @@ class Dependency(Operational):
         value = Ref(self.expr).resolve(RefContext(self.target))
         if isinstance(self.schema, dict):
             return not validate_schema(value, self.schema)
-        else:  # ProperyDef
+        else:
+            assert isinstance(self.schema, Property), type(self.schema)
             try:
                 self.schema._validate(value)
             except Exception:
@@ -1174,7 +1252,9 @@ class Dependency(Operational):
             return False
         return False  # assuming this is what is meant by the function
 
-    def has_changed(self, config: "ChangeRecord") -> bool:
+    def has_changed(self, config: Optional["ChangeRecord"] = None) -> bool:
+        if config is None:
+            return self._is_unexpected()
         changeId = config.changeId
         # Manifest.load_config_change sets config.target
         context = RefContext(config.target, dict(val=self.expected, changeId=changeId))  # type: ignore
