@@ -1,12 +1,15 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
-from ..configurator import Configurator
-from ..result import ResultsMap
+import os
+from ..eval import Ref, map_value
+from ..configurator import Configurator, TaskView
+from ..result import Results, ResultsMap
 from ..util import register_short_names
 from ..support import Status
 from ..planrequests import set_default_command
 import importlib
-from typing import Tuple
+from typing import Sequence, Tuple
+from collections.abc import Mapping
 
 # need to define these now because these configurators are lazily imported
 # and so won't register themselves through AutoRegisterClass
@@ -39,7 +42,7 @@ class PythonPackageCheckConfigurator(Configurator):
 class TemplateConfigurator(Configurator):
     exclude_from_digest: Tuple[str, ...] = ("resultTemplate", "done")
 
-    def process_result_template(self, task, result):
+    def process_result_template(self, task: "TaskView", result):
         """
         for both the ansible and shell configurators
         result can include: "returncode", "msg", "error", "stdout", "stderr"
@@ -47,34 +50,41 @@ class TemplateConfigurator(Configurator):
         """
         # get the resultTemplate without evaluating it
         resultTemplate = task.inputs._attributes.get("resultTemplate")
-        errors = []
-        if resultTemplate:  # evaluate it now with the result
-            if isinstance(resultTemplate, str):
-                query = dict(template=resultTemplate)
-            else:
-                query = resultTemplate
-
-            # workaround for jinja template processing setting Result when getting items
-            if not isinstance(result, ResultsMap):
-                vars = ResultsMap(result, task.inputs.context)
-                vars.doFullResolve = True
-            else:
-                vars = result
-            task.logger.trace("evaluated result template with %s", result)
+        errors: Sequence[Exception] = []
+        current_status = task.target.local_status
+        if resultTemplate:  # evaluate it now with the result    
+            if os.getenv("UNFURL_TEST_DEBUG_EX"):
+                task.logger.error("evaluated result template (%s) with %s %s", type(resultTemplate), type(result), result)
+            if isinstance(resultTemplate, Results):
+                resultTemplate = resultTemplate._attributes
             try:
-                results = task.query({"eval": query}, vars=vars, throw=True)
+                if os.getenv("UNFURL_TEST_DEBUG_EX"):
+                    task.logger.error("result %s template is %s", type(resultTemplate), resultTemplate)   
+                    trace = 2
+                else:
+                    trace = 0
+                if Ref.is_ref(resultTemplate):
+                    results = task.query({"eval": resultTemplate}, vars=result, throw=True)
+                else:
+                    # lazily evaluated by update_instances() below
+                    results = Results._map_value(resultTemplate, task.inputs.context.copy(vars=result, trace=trace))
             except Exception as e:
                 results = None
                 errors = [e]
+                task.logger.debug("exception when processing resultTemplate", exc_info=True)
             else:
                 if results:
-                    jr, errors = task.update_instances(results)
+                    jr, errors = task.update_instances(results)  # type: ignore
             if errors:
                 task.logger.warning(
                     "error processing resultTemplate: %s",
                     errors[0],
                 )
-        return errors
+        if task.target.local_status != current_status: 
+            new_status = task.target.local_status
+        else:
+            new_status = None
+        return errors, new_status
 
     def can_dry_run(self, task):
         return not not task.inputs.get("dryrun")
@@ -88,7 +98,7 @@ class TemplateConfigurator(Configurator):
             runResult = task.inputs.get("run")
         return runResult
 
-    def done(self, task, **kw):
+    def done(self, task: "TaskView", **kw):
         # this is called by derived classes like ShellConfigurator to allow the user
         # to override the default logic for updating the state and status of a task.
         done = task.inputs.get_copy("done", {})
@@ -97,16 +107,19 @@ class TemplateConfigurator(Configurator):
             kw.update(done)  # "done" overrides kw
         return task.done(**kw)
 
-    def run(self, task):
+    def run(self, task: "TaskView"):
         runResult = task.rendered
         done = task.inputs.get_copy("done", {})
         if "result" not in done:
-            if not isinstance(runResult, dict):
+            if not isinstance(runResult, Mapping):
                 done["result"] = {"run": runResult, "outputs": done.get("outputs")}
             else:
                 done["result"] = runResult
-        if self.process_result_template(task, done.get("result") or {}):
+        errors, new_status = self.process_result_template(task, done.get("result") or {})
+        if errors:
             done["success"] = False  # returned errors
+        if new_status is not None:
+            done["status"] = new_status
         yield task.done(**done)
 
 
@@ -143,6 +156,6 @@ class DelegateConfigurator(Configurator):
         else:
             subtaskRequest = task.rendered
             assert subtaskRequest
-            # note: this will call canRun() and if needed canDryRun() on subtask but not shouldRun()
+            # note: this will call can_run_task() for the subtask but not shouldRun()
             subtask = yield subtaskRequest
             yield subtask.result
