@@ -3,6 +3,7 @@
 from __future__ import absolute_import
 import re
 from base64 import b64encode
+from typing import Dict, Union
 from ..configurator import Configurator
 from ..support import Status, Priority
 from ..runtime import RelationshipInstance
@@ -28,11 +29,12 @@ def _get_connection_config(instance):
     # https://github.com/ansible-collections/kubernetes.core/blob/7f7008fecc9e5d16340e9b0bff510b7cde2f2cfd/plugins/connection/kubectl.py
     if not instance:
         return {}
-    connection = {}
+    connection : Dict[str, Union[str, bool]] = {}
     if isinstance(instance, RelationshipInstance):
         connect = instance.attributes
         # the parent of a relationship will be the capability it connects to
         # or root if relationship is a default connection
+        assert instance.parent
         if instance.parent is not instance.root:
             endpoint = instance.parent.attributes
         else:
@@ -205,6 +207,9 @@ class ClusterConfigurator(Configurator):
             return "Configurator can't perform this operation (only supports check and discover)"
         return True
 
+    def can_dry_run(self, task):
+        return True
+
     def run(self, task):
         cluster = task.target
         # the endpoint on the cluster will have connection info
@@ -277,7 +282,8 @@ def mark_sensitive(task, resource):
                 if containers:
                     for container in containers:
                         for env in container.get("env") or []:
-                            env["value"] = task.sensitive(env["value"])
+                            if "value" in env:
+                                env["value"] = task.sensitive(env["value"])
 
 
 class ResourceConfigurator(AnsibleConfigurator):
