@@ -44,7 +44,6 @@ import logging
 
 if TYPE_CHECKING:
     from .configurator import TaskView
-    from .runtime import NodeInstance
 
 
 logger = cast(UnfurlLogger, logging.getLogger("unfurl.eval"))
@@ -52,15 +51,15 @@ logger = cast(UnfurlLogger, logging.getLogger("unfurl.eval"))
 
 def map_value(
     value: Any,
-    resourceOrCxt: Union["RefContext", "NodeInstance"],
+    resourceOrCxt: Union["RefContext", "ResourceRef"],
     applyTemplates: bool = True,
-) -> Optional[Union[ResultsList, Result, List[Result]]]:
+) -> Any:
     if not isinstance(resourceOrCxt, RefContext):
         resourceOrCxt = RefContext(resourceOrCxt)
     return _map_value(value, resourceOrCxt, False, applyTemplates)
 
 
-def _map_value(value, ctx, wantList=False, applyTemplates=True):
+def _map_value(value: Any, ctx: "RefContext", wantList: bool = False, applyTemplates: bool = True) -> Any:
     from .support import is_template, apply_template
 
     if Ref.is_ref(value):
@@ -130,6 +129,7 @@ class RefContext:
     """
 
     DefaultTraceLevel = 0
+    _Funcs: Dict = {}
 
     def __init__(
         self,
@@ -166,6 +166,13 @@ class RefContext:
         else:
             return self._strict
 
+    @property
+    def environ(self):
+        if self.task:
+            return self.task.environ
+        else:
+            return self.currentResource.environ
+
     def copy(
         self,
         resource: Optional["ResourceRef"] = None,
@@ -178,7 +185,7 @@ class RefContext:
             self._lastResource, ResourceRef
         ):
             resource = self._lastResource
-        copy = RefContext(
+        copy = self.__class__(
             resource or self.currentResource,
             self.vars,
             self.wantList,
@@ -199,7 +206,7 @@ class RefContext:
 
     def trace(self, *msg: Any) -> None:
         if self._trace:
-            log = logger.info if self._trace == 2 else logger.trace
+            log = logger.info if self._trace >= 2 else logger.trace
             log(f"{' '.join(str(a) for a in msg)} (ctx: {self._lastResource})")  # type: ignore
 
     def add_external_reference(self, external: ExternalValue) -> Result:
@@ -286,7 +293,7 @@ class Expr:
         # XXX vars
         return f"Expr('{self.source}')"
 
-    def resolve(self, context):
+    def resolve(self, context) -> List[Result]:
         # returns a list of Result
         currentResource = context.currentResource
         if not self.paths[0].key and not self.paths[0].filters:  # starts with "::"
@@ -334,7 +341,7 @@ class Ref:
         ctx: RefContext,
         wantList: Union[bool, str] = True,
         strict: bool = _defaultStrictness,
-    ) -> Optional[Union[ResultsList, Result, List[Result]]]:
+    ) -> Optional[Union[ResultsList, Result, List[Result], Any]]:
         """
         If wantList=True (default) returns a ResultList of matches
         Note that values in the list can be a list or None
@@ -474,7 +481,7 @@ def eq_func(arg, ctx):
 
 def validate_schema_func(arg, ctx):
     args = map_value(arg, ctx)
-    assert_form(args, MutableSequence, len(args) == 2)
+    assert_form(args, MutableSequence, len(args) == 2)  # type: ignore
     return validate_schema(args[0], args[1])
 
 if sys.version_info >= (3, 9):
@@ -524,7 +531,10 @@ def _for_each(
                 elif key is Continue:
                     continue
             assert valExp
-            evalResults = eval_ref(valExp, ictx)
+            if isinstance(valExp, str):
+                evalResults = eval_ref(valExp, ictx)
+            else:
+                evalResults = _eval_ref_results(valExp, ictx)
             if not evalResults:
                 continue
             if len(evalResults) == 1:
@@ -558,20 +568,17 @@ def for_each(
     return ResultsList(result, ctx)  # ResultsList[List[ResultOrResultList]]]
 
 
-def for_each_func(foreach: Union[Mapping, str], ctx: RefContext):
+def for_each_func(foreach: Union[Mapping, str], ctx: RefContext) -> List[ResultOrResultList]:
     results = ctx.currentResource
-    if results:
-        if isinstance(results, Mapping):
-            return _for_each(foreach, results.items(), ctx)
-        elif isinstance(results, MutableSequence):
-            return _for_each(foreach, enumerate(results), ctx)
-        else:
-            return _for_each(foreach, [(0, results)], ctx)
+    if isinstance(results, Mapping):
+        return _for_each(foreach, results.items(), ctx)
+    elif isinstance(results, MutableSequence):
+        return _for_each(foreach, enumerate(results), ctx)
     else:
-        return results
+        return _for_each(foreach, [(0, results)], ctx)
 
 
-_Funcs = {
+_CoreFuncs = {
     "if": if_func,
     "and": and_func,
     "or": or_func,
@@ -581,17 +588,24 @@ _Funcs = {
     "validate": validate_schema_func,
     "foreach": for_each_func,
 }
+RefContext._Funcs = _CoreFuncs.copy()
 _FuncsTop = ["q"]
 
 
+class SafeRefContext(RefContext):
+    _Funcs = _CoreFuncs.copy()
+
+
 def get_eval_func(name):
-    return _Funcs.get(name)
+    return RefContext._Funcs.get(name)
 
 
-def set_eval_func(name, val, topLevel=False):
-    _Funcs[name] = val
+def set_eval_func(name, val, topLevel=False, safe=False):
+    RefContext._Funcs[name] = val
     if topLevel:
         _FuncsTop.append(name)
+    if safe:
+        SafeRefContext._Funcs[name] = val
 
 
 def eval_ref(
@@ -608,7 +622,7 @@ def eval_ref(
 
     if isinstance(val, Mapping):
         for key in val:
-            func = _Funcs.get(key)
+            func = ctx._Funcs.get(key)
             if func:
                 args = val[key]
                 ctx.kw = val
@@ -630,6 +644,9 @@ def eval_ref(
                     else:
                         return [Result(val)]
                 break
+        else:
+            if "ref" not in val and "eval" not in val:
+                raise UnfurlError(f"Function missing in {dict(val)}")
     elif isinstance(val, six.string_types):
         if is_template(val, ctx):
             return [Result(apply_template(val, ctx))]
@@ -638,7 +655,9 @@ def eval_ref(
             results = expr.resolve(ctx)  # returns a list of Result
             ctx.trace("expr.resolve", results)
             return results
+    return _eval_ref_results(val, ctx)
 
+def _eval_ref_results(val, ctx):
     mappedVal = Results._map_value(val, ctx)
     if isinstance(mappedVal, Result):
         return [mappedVal]
@@ -662,7 +681,7 @@ Segment = collections.namedtuple("Segment", ["key", "test", "modifier", "filters
 defaultSegment = Segment("", [], "", [])
 
 
-def eval_test(value, test, context):
+def eval_test(value, test, context) -> bool:
     comparor = test[0]
     key = test[1]
     try:
@@ -681,12 +700,11 @@ def eval_test(value, test, context):
     return False
 
 
-# given a Result and a key, return None or new Result
-def lookup(result, key, context):
+def lookup(result: Result, key: Any, context: RefContext) -> Optional[Result]:
     try:
         # if key == '.':
         #   key = context.currentKey
-        if context and isinstance(key, six.string_types) and key.startswith("$"):
+        if context and isinstance(key, str) and key.startswith("$"):
             key = context.resolve_var(key)
 
         if isinstance(result.resolved, ResourceRef):
@@ -806,7 +824,7 @@ def recursive_eval(v, exp, context):
                     return
 
 
-def eval_exp(start, paths, context):
+def eval_exp(start, paths, context) -> List[Result]:
     "Returns a list of Result"
     context.trace("evalexp", start, paths)
     assert_form(start, MutableSequence)
