@@ -9,7 +9,7 @@ import os.path
 import shutil
 import sys
 import json
-from typing import Any
+from typing import Any, Dict, Optional
 import uuid
 import logging
 from jinja2.loaders import FileSystemLoader
@@ -17,7 +17,7 @@ from pathlib import Path
 
 from . import DefaultNames, __version__, get_home_config_path, is_version_unreleased
 from .localenv import LocalEnv, Project, LocalConfig
-from .repo import GitRepo, Repo, is_url_or_git_path, split_git_url, commit_secrets
+from .repo import GitRepo, Repo, is_url_or_git_path, split_git_url, commit_secrets, sanitize_url
 from .util import UnfurlError, get_random_password, substitute_env
 from .yamlloader import make_yaml, make_vault_lib
 
@@ -146,7 +146,7 @@ def write_service_template(projectdir):
 
 
 def write_ensemble_manifest(
-    destDir, manifestName, specRepo, specDir=None, extraVars=None, templateDir=None
+    destDir: str, manifestName: str, specRepo, specDir=None, extraVars=None, templateDir=None
 ):
     if specDir:
         specDir = os.path.abspath(specDir)
@@ -195,6 +195,7 @@ def render_project(
     use_context=None,
     mono=False,
     no_secrets=False,
+    skeleton_vars: Optional[Dict[str, Any]]=None,
 ):
     """
     Creates a folder named `projectdir` with a git repository with the following files:
@@ -226,12 +227,19 @@ def render_project(
         manifestName = names.Ensemble
         ensemblePath = os.path.join(ensembleDir, manifestName)
 
-    vaultpass = get_random_password()
-    # use project name plus a couple of random digits to avoid collisions
-    vaultid = (
-        Project.get_name_from_dir(projectdir) + get_random_password(2, "", "").upper()
-    )
-    vars = dict(vaultpass=vaultpass, vaultid=vaultid)
+    if skeleton_vars:
+        vars = skeleton_vars.copy()
+    else:
+        vars = {}
+    if "vaultpass" not in vars:
+        vars["vaultpass"] = get_random_password()
+    vaultpass = vars["vaultpass"]
+    if "vaultid" not in vars:
+        # use project name plus a couple of random digits to avoid collisions
+        vars["vaultid"] = (
+            Project.get_name_from_dir(projectdir) + get_random_password(2, "", "").upper()
+        )
+    vaultid = vars["vaultid"]
 
     # only commit external ensembles references if we are creating a mono repo
     # otherwise record them in the local config:
@@ -250,7 +258,8 @@ def render_project(
         templateDir,
     )
 
-    _warn_about_new_password(localProjectConfig)
+    if not skeleton_vars or "vaultpass" not in skeleton_vars:
+        _warn_about_new_password(localProjectConfig)
 
     localInclude = "+?include-local: " + os.path.join("local", localConfigFilename)
 
@@ -270,6 +279,8 @@ def render_project(
 
     # note: local overrides secrets
     vars = dict(include=secretsInclude + "\n" + localInclude, vaultid=vaultid)
+    if skeleton_vars:
+        vars.update(skeleton_vars)
     if use_context and not localExternal:
         # since this is specified while creating the project set this as the default context
         vars["default_context"] = use_context
@@ -305,6 +316,8 @@ def render_project(
             # include the ensembleTemplate in the root of the specDir
             ensembleTemplate=names.EnsembleTemplate,
         )
+        if skeleton_vars:
+            extraVars.update(skeleton_vars)
         # write ensemble/ensemble.yaml
         write_ensemble_manifest(
             ensembleDir,
@@ -397,6 +410,7 @@ def create_project(
 ):
     create_context = kw.get("create_environment")
     use_context = kw.get("use_environment")
+    skeleton_vars = dict((n, v) for n, v in kw.get("var", []))
     if existing:
         repo = _find_project_repo(projectdir)
     else:
@@ -447,6 +461,8 @@ def create_project(
         names,
         create_context or use_context,
         mono,
+        False,
+        skeleton_vars,
     )
     if homePath and create_context:
         newProject = find_project(projectConfigPath, homePath)
@@ -636,6 +652,7 @@ class EnsembleBuilder:
         assert not self.templateVars
 
         # source is a path into the project relative to the current directory
+        assert self.source_path
         source_path = os.path.join(self.source_project.projectRoot, self.source_path)
         self.templateVars = _get_ensemble_paths(
             source_path,
@@ -667,6 +684,7 @@ class EnsembleBuilder:
         return destDir, manifestName
 
     def _get_inputs_template(self):
+        assert self.source_project
         local_template = os.path.join(
             self.source_project.projectRoot, DefaultNames.InputsTemplate
         )
@@ -680,6 +698,7 @@ class EnsembleBuilder:
 
         specProject = self.dest_project
         assert project
+        assert self.source_project
         sourceDir = os.path.normpath(
             os.path.join(
                 self.source_project.projectRoot, self.templateVars["sourceDir"]
@@ -774,22 +793,13 @@ class EnsembleBuilder:
             )
         else:
             destProject.register_ensemble(manifest.path, context=self.environment)
-        environments_json_path = os.path.join(
-            destProject.projectRoot, "environments.json"
-        )
-        if os.path.exists(environments_json_path):
-            # temp hack for unfurl.cloud integration
-            from .to_json import set_deploymentpaths
-
-            envjson = set_deploymentpaths(destProject, environments_json_path)
-            with open(environments_json_path, "w") as f:
-                f.write(json.dumps(envjson, indent=2))
-
+        destProject.project_repoview.repo.commit_files([destProject.localConfig.config.path], "Add ensemble")
         self.manifest = manifest
         return destDir
 
     def clone_local_project(self, currentProject, sourceProject, dest_dir):
         self.source_path = sourceProject.get_relative_path(self.input_source)
+        assert self.source_path
         assert not self.source_path.startswith(
             ".."
         ), f"{self.source_path} should be inside the project"
@@ -834,14 +844,16 @@ class EnsembleBuilder:
 
         targetDir = os.path.join(destDir, filePath)
         sourceProjectRoot = Project.find_path(targetDir, destDir)
-        self.logger.debug(f'cloned {self.input_source}" to "{destDir}"')
+        clean_url = sanitize_url(self.input_source)
+        self.logger.debug(f'cloned "{clean_url}" to "{destDir}"')
         if not sourceProjectRoot:
             raise UnfurlError(
-                f'Error: cloned "{self.input_source}" to "{destDir}" but couldn\'t find an Unfurl project'
+                f'Error: cloned "{clean_url}" to "{destDir}" but couldn\'t find an Unfurl project'
             )
 
         self.source_project = find_project(sourceProjectRoot, self.home_path)
         # set source to point to the cloned project
+        assert self.source_project
         self.source_path = self.source_project.get_relative_path(targetDir)
         return self.source_project
 
@@ -888,9 +900,8 @@ class EnsembleBuilder:
         self.dest_path = relDestDir
 
     def has_existing_ensemble(self, sourceProject):
-        from unfurl import yamlmanifest
-
         if self.source_project is not sourceProject and not self.shared_repo:
+            assert self.dest_project
             if "localEnv" in self.templateVars and os.path.exists(
                 Path(self.dest_project.projectRoot) / self.dest_path
             ):
@@ -908,6 +919,7 @@ class EnsembleBuilder:
     def set_ensemble(self, isRemote, existingSourceProject, existingDestProject):
         sourceWasCloned = self.source_project is not existingSourceProject
         destIsNew = not existingDestProject
+        assert self.dest_project
         if destIsNew and self.has_existing_ensemble(existingSourceProject):
             # if dest_project is new (we just cloned it)
             # check if we cloned the ensemble already
@@ -1035,8 +1047,8 @@ def _create_local_config(clonedProject, logger, vars):
         vars["generate_new_vault_password"] = (
             lambda: _warn_about_new_password(dest) or get_random_password()
         )
-        # replace ${var}
-        contents = substitute_env(contents, vars)
+        # replace ${var} or preserve if not set
+        contents = substitute_env(contents, vars, True)
 
         _write_file(
             os.path.join(clonedProject.projectRoot, "local"),
@@ -1117,7 +1129,7 @@ def _run_pip_env(do_install, pipenv_project, kw):
         def noexit(code):
             retcode = code
 
-        sys.exit = noexit
+        sys.exit = noexit  # type: ignore
 
         do_install(pipenv_project, **kw)
     finally:
@@ -1196,7 +1208,7 @@ def create_venv(projectDir, pipfileLocation, unfurlLocation):
             return f'Pipfile location is not a valid directory: "{pipfileLocation}"'
         copy_pipfiles(pipfileLocation, projectDir)
 
-        kw = dict(python=pythonPath, extra_index_url=[])
+        kw: Dict = dict(python=pythonPath, extra_index_url=[])
         pipenv_project = PipEnvProject()
         # need to run without args first so lock isn't overwritten
         retcode = _run_pip_env(do_install, pipenv_project, kw)
