@@ -9,6 +9,7 @@ By convention, the "home" project defines a localhost instance and adds it to it
 """
 import os
 import os.path
+import re
 from typing import (
     Any,
     Iterable,
@@ -19,23 +20,22 @@ from typing import (
     Tuple,
     Union,
     TYPE_CHECKING,
+    cast,
 )
-
 from ansible.parsing.vault import VaultLib
-import six
-from six import Iterator
 from unfurl.runtime import NodeInstance
 
-from .repo import GitRepo, Repo, split_git_url, RepoView, normalize_git_url_hard
+from .repo import GitRepo, Repo, add_user_to_url, split_git_url, RepoView, normalize_git_url_hard
 from .util import (
     UnfurlError,
+    filter_env,
     substitute_env,
     wrap_sensitive_value,
     save_to_tempfile,
     is_sensitive,
 )
 from .merge import merge_dicts
-from .yamlloader import YamlConfig, make_vault_lib_ex, make_yaml
+from .yamlloader import UnfurlVaultLib, YamlConfig, make_vault_lib_ex, make_yaml
 from . import DefaultNames, get_home_config_path
 from urllib.parse import urlparse, urlunsplit, urlsplit
 from ruamel.yaml.comments import CommentedMap
@@ -52,7 +52,6 @@ from .logs import getLogger
 
 logger = getLogger("unfurl")
 
-
 class Project:
     """
     A Unfurl project is a folder that contains at least a local configuration file (unfurl.yaml),
@@ -64,16 +63,17 @@ class Project:
         path: str,
         homeProject: Optional["Project"] = None,
         overrides: Optional[dict] = None,
+        readonly: Optional[bool] = False,
     ):
-        assert isinstance(path, six.string_types), path
+        assert isinstance(path, str), path
         self.projectRoot = os.path.abspath(os.path.dirname(path))
         self.overrides = overrides or {}
         if os.path.exists(path):
             self.localConfig = LocalConfig(
-                path, yaml_include_hook=self.load_yaml_include
+                path, yaml_include_hook=self.load_yaml_include, readonly=readonly
             )
         else:
-            self.localConfig = LocalConfig()
+            self.localConfig = LocalConfig(readonly=readonly)
         self._set_repos()
         # XXX this updates and saves the local config to disk -- constructing a Project object shouldn't do that
         # especially since we localenv might not have found the ensemble yet
@@ -160,7 +160,9 @@ class Project:
         current = os.path.abspath(testPath)
         if not stopPath:
             stopPath = os.sep
-        while current and current != stopPath:
+        else:
+            stopPath = os.path.abspath(stopPath)
+        while current and len(current) >= len(stopPath):
             test = os.path.join(current, DefaultNames.LocalConfig)
             if os.path.exists(test):
                 return test
@@ -169,6 +171,8 @@ class Project:
             )
             if os.path.exists(test):
                 return test
+            if current == os.sep:
+                break
             current = os.path.dirname(current)
         return None
 
@@ -204,8 +208,8 @@ class Project:
                                     paths.append(path)
         return paths
 
-    def _get_git_repos(self) -> List[Repo]:
-        repos = [r.repo for r in self.workingDirs.values()]
+    def _get_git_repos(self) -> List[GitRepo]:
+        repos = [r.repo for r in self.workingDirs.values() if r.repo]
         if self.project_repoview.repo and self.project_repoview.repo not in repos:
             repos.append(self.project_repoview.repo)
         return repos
@@ -227,6 +231,12 @@ class Project:
         return None
 
     def get_relative_path(self, path: str) -> str:
+        # TODO: consider asserting that `path` is an absolute path so this method is not ambiguous
+        # assert os.path.isabs(path)
+        # return os.path.relpath(path, self.projectRoot)
+
+        # NOTE: `os.path.abspath` constructs the absolute path from the current working directory
+        # This may not be the desired behavior (unfurl-server being a counterexample)
         return os.path.relpath(os.path.abspath(path), self.projectRoot)
 
     def is_path_in_project(self, path: str) -> bool:
@@ -248,11 +258,13 @@ class Project:
         self, repoURL: str, revision: Optional[str] = None
     ) -> Union[GitRepo, str]:
         # if repo isn't found, return the repo url, possibly rewritten to include server credentials
+        apply_credentials = self.overrides.get("apply_url_credentials")
         candidate = None
-        # repourl_parts = urlsplit(repoURL)
+        repourl_parts = urlsplit(repoURL)
         normalized = normalize_git_url_hard(repoURL)
         for dir, repository in self.workingDirs.items():
             repo = repository.repo
+            assert repo
             if repoURL.startswith("git-local://"):
                 initialCommit = urlparse(repoURL).netloc.partition(":")[0]
                 match = initialCommit == repo.get_initial_revision()
@@ -266,16 +278,16 @@ class Project:
                     candidate = repo
                 else:
                     return repo
-            # elif not repourl_parts.username:  # if url doesn't have credentials already
-            #     candidate_parts = urlsplit(repo.url)
-            #     if (
-            #         candidate_parts.password
-            #         and candidate_parts.hostname == repourl_parts.hostname
-            #     ):
-            #         # rewrite repoUrl to add credentials
-            #         repoURL = urlunsplit(
-            #             repourl_parts._replace(netloc=candidate_parts.netloc)
-            #         )
+            elif apply_credentials:
+                # if an existing repository has the same hostname as the new repository's url
+                # and has credentials, update the new url with those credentials
+                candidate_parts = urlsplit(repo.url)
+                password = candidate_parts.password
+                if password:
+                    if (candidate_parts.hostname == repourl_parts.hostname
+                       and candidate_parts.port == repourl_parts.port):
+                        # rewrite repoUrl to add credentials
+                        repoURL = add_user_to_url(repoURL, candidate_parts.username, password)
         return candidate or repoURL
 
     def find_git_repo(
@@ -295,6 +307,7 @@ class Project:
         candidate = None
         for dir in sorted(self.workingDirs.keys()):
             repo = self.workingDirs[dir].repo
+            assert repo
             filePath = repo.find_repo_path(path)
             if filePath is not None:
                 return repo, filePath, repo.revision, False
@@ -321,7 +334,7 @@ class Project:
 
     def find_or_clone(self, repo: GitRepo) -> GitRepo:
         for dir, repository in self.workingDirs.items():
-            if repo.get_initial_revision() == repository.repo.get_initial_revision():
+            if repository.repo and repo.get_initial_revision() == repository.repo.get_initial_revision():
                 return repository.repo
 
         gitUrl = repo.url
@@ -343,7 +356,7 @@ class Project:
         return repo
 
     def get_manifest_path(
-        self, localEnv: "LocalEnv", manifestPath: str, can_be_empty: bool
+        self, localEnv: "LocalEnv", manifestPath: Optional[str], can_be_empty: bool
     ) -> Tuple[Optional["Project"], str, str]:
         if manifestPath:
             # at this point a named manifest
@@ -445,6 +458,10 @@ class Project:
                 yield vaultId, password
 
     def make_vault_lib(self, contextName: Optional[str] = None) -> Optional[VaultLib]:
+        if self.overrides.get("UNFURL_SKIP_VAULT_DECRYPT"):
+            vault = UnfurlVaultLib(secrets=None)
+            vault.skip_decode = True
+            return vault
         secrets = list(self.get_vault_passwords(contextName))
         if secrets:
             return make_vault_lib_ex(secrets)
@@ -453,24 +470,32 @@ class Project:
     @staticmethod
     def _find_ensemble_by_path(ensembles, projectRoot, path: str) -> Optional[dict]:
         path = os.path.abspath(path)
+        match = None
         for tpl in ensembles:
             file = tpl.get("file")
             if file and os.path.normpath(os.path.join(projectRoot, file)) == path:
-                return tpl
-        return None
+                if not match:
+                    match = tpl
+                else:  # merge duplicate ensembles
+                    match.update(tpl)
+        return match
 
     def find_ensemble_by_path(self, path: str) -> Optional[dict]:
         return self._find_ensemble_by_path(
             self.localConfig.ensembles, self.projectRoot, path
         )
 
-    def find_ensemble_by_name(self, name: str) -> Optional[dict]:
-        for tpl in self.localConfig.ensembles:
+    @staticmethod
+    def _find_ensemble_by_name(ensembles, name: str) -> Optional[dict]:
+        for tpl in ensembles:
             alias = tpl.get("alias")
             if alias and alias == name:
                 return tpl
         return None
 
+    def find_ensemble_by_name(self, name: str) -> Optional[dict]:
+        return self._find_ensemble_by_name(self.localConfig.ensembles, name)
+
     @property
     def name(self) -> str:
         return self.get_name_from_dir(self.projectRoot)
@@ -480,6 +505,11 @@ class Project:
         dirname, name = os.path.split(projectRoot)
         if name == DefaultNames.ProjectDirectory:
             name = os.path.basename(dirname)
+        # make sure this conforms to project name syntax in json-schema:
+        #       ^[A-Za-z._][A-Za-z0-9._:\\-]*$
+        name = re.sub(r"[^A-Za-z0-9._:\\-]", "_", name)
+        if not name[0].isalpha() and name[0] not in "._":
+            name = "_" + name
         return name
 
     def get_default_context(self) -> Any:
@@ -547,9 +577,10 @@ class Project:
         if managedBy or project:
             assert not (managedBy and project)
             self.register_project(managedBy or project, changed=True)
-        elif context:
+        else:
             self.localConfig.config.config["ensembles"] = self.localConfig.ensembles
-            self.localConfig.config.save()
+            if not self.localConfig.config.readonly:
+                self.localConfig.config.save()
 
     def register_project(
         self, project, for_context=None, changed=False, save_project=True
@@ -579,7 +610,7 @@ class Project:
             key = templatePath
             merge = None
 
-        url_vars = os.environ.copy()
+        url_vars = {}
         url_vars.update(self.overrides)
         if action:
             if isinstance(action, str):
@@ -607,6 +638,10 @@ class Project:
                     ensemble_tpl = self._find_ensemble_by_path(
                         ensembles, self.projectRoot, url_vars["manifest_path"]
                     )
+                elif "manifest_alias" in url_vars:
+                    ensemble_tpl = self._find_ensemble_by_name(
+                        ensembles, url_vars["manifest_alias"]
+                    )
                 else:
                     ensemble_tpl = LocalConfig._get_default_manifest_tpl(ensembles)
                 if ensemble_tpl:
@@ -634,7 +669,7 @@ class LocalConfig:
         "instances",
     ]
 
-    def __init__(self, path=None, validate=True, yaml_include_hook=None):
+    def __init__(self, path=None, validate=True, yaml_include_hook=None, readonly=False):
         defaultConfig = {"apiVersion": "unfurl/v1alpha1", "kind": "Project"}
         self.config = YamlConfig(
             defaultConfig,
@@ -642,6 +677,7 @@ class LocalConfig:
             validate,
             os.path.join(_basepath, "unfurl-schema.json"),
             yaml_include_hook,
+            readonly=readonly
         )
         self.ensembles = self.config.expanded.get("ensembles") or []
         self.projects = self.config.expanded.get("projects") or {}
@@ -777,10 +813,11 @@ class LocalConfig:
                     changed = True
 
         if changed:
-            if key:
+            if key and not self.config.readonly:
                 self.config.save_include(key)
             self.config.config["ensembles"] = self.ensembles
-            self.config.save()
+            if not self.config.readonly:
+                self.config.save()
         return lock
 
 
@@ -842,6 +879,8 @@ class LocalEnv:
             manifestPath = ""
         else:
             assert manifestPath
+            # set this because the yaml loader needs access to this while the project is being instantiated
+            self.overrides["manifest_alias"] = manifestPath
             # if manifestPath doesn't point to a project or ensemble,
             # look for a project in the current directory and then see if the project has a manifest with that name
             project = self.find_project(".")
@@ -878,6 +917,8 @@ class LocalEnv:
         project: Optional[Project] = None,
         can_be_empty: bool = False,
         override_context: Optional[str] = None,
+        overrides: Optional[Dict[str, Any]] = None,
+        readonly: Optional[bool] = False,
     ) -> None:
         """
         If manifestPath is None find the first unfurl.yaml or ensemble.yaml
@@ -892,11 +933,16 @@ class LocalEnv:
         logger = logging.getLogger("unfurl")
         self.logger = logger
         self.manifest_context_name = None
-        self.overrides: Dict[str, Any] = {}
+        self.overrides: dict = overrides or (parent and parent.overrides.copy()) or {}
+        self.readonly = readonly
         if override_context is not None:
             # aka the --use-environment option
             # hackishly, "" is a valid option used by load_yaml_include
             self.overrides["ENVIRONMENT"] = override_context
+        if os.getenv("UNFURL_SKIP_VAULT_DECRYPT"):
+            self.overrides["UNFURL_SKIP_VAULT_DECRYPT"] = True
+        if os.getenv("UNFURL_SKIP_UPSTREAM_CHECK"):
+            self.overrides["UNFURL_SKIP_UPSTREAM_CHECK"] = True
 
         if parent:
             self.parent = parent
@@ -1009,7 +1055,7 @@ class LocalEnv:
         path = Project.normalize_path(path)
         project = self._projects.get(path)
         if not project:
-            project = Project(path, homeProject, self.overrides)
+            project = Project(path, homeProject, self.overrides, readonly=self.readonly)
             self._projects[path] = project
         return project
 
@@ -1037,7 +1083,10 @@ class LocalEnv:
             project = self.homeProject.localConfig.projects.get(projectName)
             # allow "home" to refer to the home project
             if not project and projectName == "home":
-                repo = self.homeProject.project_repoview
+                repoview = self.homeProject.project_repoview
+                if not repoview:
+                    return None
+                repo = repoview.repo
                 file = ""
 
         if project:
@@ -1078,9 +1127,10 @@ class LocalEnv:
                         message = f"Can't find an Unfurl ensemble or project in folder '{manifestPath}'"
                         raise UnfurlError(message)
         else:
-            test = os.path.join(manifestPath, DefaultNames.LocalConfig)
-            if os.path.exists(test):
-                return None, self.get_project(test, self.homeProject)
+            assert os.path.isfile(manifestPath)
+            if os.path.basename(manifestPath) == DefaultNames.LocalConfig:
+                # assume unfurl.yaml is a project file
+                return None, self.get_project(manifestPath, self.homeProject)
             # assume its a pointing to an ensemble
             return manifestPath, None
 
@@ -1096,7 +1146,7 @@ class LocalEnv:
 
     # NOTE this currently isn't used and returns repos outside of this LocalEnv
     # (every repo found in localRepositories)
-    def _get_git_repos(self) -> List[Repo]:
+    def _get_git_repos(self) -> List[GitRepo]:
         if self.project:
             repos = self.project._get_git_repos()
         else:
@@ -1223,7 +1273,7 @@ class LocalEnv:
             logger.debug(
                 "Using existing repository at %s for %s", repo.working_dir, repoURL
             )
-            if not repo.is_dirty():
+            if not self.overrides.get("UNFURL_SKIP_UPSTREAM_CHECK") and not repo.is_dirty():
                 repo.pull(revision=revision)
         else:
             assert isinstance(
@@ -1278,16 +1328,18 @@ class LocalEnv:
                 return repo, filePath, revision, bare
         return None, None, None, None
 
-    def map_value(self, val: Any) -> Any:
+    def map_value(self, val: Any, env_rules: Optional[dict]) -> Any:
         """
         Evaluate using project home as a base dir.
         """
         from .runtime import NodeInstance
-        from .eval import map_value
+        from .eval import map_value, RefContext
 
         instance = NodeInstance()
         instance._baseDir = self.config.config.get_base_dir()
-        return map_value(val, instance)
+        if env_rules is not None:
+            instance._environ = filter_env(env_rules, instance.environ)
+        return map_value(val, RefContext(instance))
 
     def get_paths(self) -> List[str]:
         """
