@@ -6,9 +6,7 @@ import os
 import tempfile
 import time
 import types
-from typing import Any, Dict, Tuple, Union
-from typing_extensions import NotRequired, TypedDict
-from abc import ABC, abstractmethod
+from typing import Any, Union, cast
 
 import rich
 from rich.console import Console
@@ -20,8 +18,8 @@ except ImportError:
 
 HIDDEN_MSG_LOGGER = "unfurl.metadata"
 
-
-def truncate(s: str, max: int = 748) -> str:
+DEFAULT_TRUNCATE_LENGTH = 748
+def truncate(s: str, max: int = DEFAULT_TRUNCATE_LENGTH) -> str:
     if not s:
         return ""
     if len(s) > max:
@@ -86,7 +84,7 @@ class UnfurlLogger(logging.Logger, LogExtraLevels):
 
 
 def getLogger(name: str) -> UnfurlLogger:
-    return logging.getLogger(name)  # type: ignore
+    return cast(UnfurlLogger, logging.getLogger(name))
 
 
 class HiddenOutputLogHandler(logging.StreamHandler):
@@ -95,11 +93,17 @@ class HiddenOutputLogHandler(logging.StreamHandler):
         rich.print(record.msg, end="\x1b[2K\r", flush=True)
 
 
-def getConsole(**kwargs):
+PY_COLORS = os.environ.get("PY_COLORS") != "0"
+
+
+def getConsole(**kwargs) -> Console:
+    global PY_COLORS
+    PY_COLORS = os.environ.get("PY_COLORS") != "0"
     # Settings needed for emulated terminals, like gitlab CI:
     # - Soft wrap prevents rich from breaking lines automatically
     # - force_terminal to display colors
-    return Console(soft_wrap=True, force_terminal=os.environ.get("PY_COLORS") != "0", **kwargs)
+    return Console(soft_wrap=True, force_terminal=PY_COLORS, **kwargs)
+
 
 class ColorHandler(logging.StreamHandler):
     # https://rich.readthedocs.io/en/stable/appendix/colors.html
@@ -114,7 +118,11 @@ class ColorHandler(logging.StreamHandler):
     }
 
     def emit(self, record: logging.LogRecord) -> None:
-        message = truncate(self.format(record))
+        message = self.format(record)
+        if not record.exc_info:
+            truncate_length = getattr(record, "truncate", DEFAULT_TRUNCATE_LENGTH)
+            if truncate_length:
+                message = truncate(message, truncate_length)
         # Hide meta job output because it seems to also be logged captured by
         # the root logger.
         if record.name.startswith(HIDDEN_MSG_LOGGER):
@@ -210,6 +218,9 @@ def initialize_logging() -> None:
             os.getenv("UNFURL_LOGGING").upper()  # type: ignore
         ]
     logging.config.dictConfig(LOGGING)
+    if os.getenv("UNFURL_LOG_FORMAT"):
+        formatter = logging.Formatter(os.getenv("UNFURL_LOG_FORMAT"))
+        logging.getLogger().handlers[0].setFormatter(formatter)
 
 
 def set_console_log_level(log_level: int) -> None:
@@ -234,7 +245,8 @@ def add_log_file(filename: str, console_level: Levels = Levels.INFO):
 
     handler = logging.FileHandler(filename)
     f = SensitiveFilter()
-    formatter = logging.Formatter("[%(asctime)s] %(name)s:%(levelname)s: %(message)s")
+    fmt = os.getenv("UNFURL_LOG_FORMAT") or "[%(asctime)s] %(name)s:%(levelname)s: %(message)s"
+    formatter = logging.Formatter(fmt)
     handler.setFormatter(formatter)
     log_level = min(console_level, Levels.DEBUG)
     handler.setLevel(log_level)
