@@ -4,8 +4,9 @@ from collections.abc import Mapping
 import os.path
 import hashlib
 import json
-from typing import Tuple, Any
+from typing import Dict, List, Optional, Any, Sequence, TYPE_CHECKING, cast
 from ruamel.yaml.comments import CommentedMap
+
 from .tosca import ToscaSpec, TOSCA_VERSION, ArtifactSpec
 
 from .support import (
@@ -17,46 +18,54 @@ from .support import (
     Imports,
 )
 from .runtime import (
+    HasInstancesInstance,
     OperationalInstance,
     NodeInstance,
     CapabilityInstance,
     RelationshipInstance,
     ArtifactInstance,
+    TopologyInstance
 )
-from .util import UnfurlError, to_enum, sensitive_str, get_base_dir
-from .repo import RevisionManager, split_git_url, RepoView
+from .util import UnfurlError, to_enum, sensitive_str, get_base_dir, taketwo
+from .repo import split_git_url, RepoView, GitRepo
+from .packages import Package, PackageSpec, PackagesType
 from .merge import merge_dicts
-from .yamlloader import YamlConfig, yaml, ImportResolver
-from .job import ConfigChange
+from .result import ChangeRecord
+from .yamlloader import yaml, ImportResolver, yaml_dict_type
+from .logs import getLogger
 import toscaparser.imports
 from toscaparser.repositories import Repository
 
-# from .configurator import Dependency
-import logging
+if TYPE_CHECKING:
+    from .localenv import LocalEnv
+    from ansible.parsing.yaml.loader import AnsibleLoader
 
-logger = logging.getLogger("unfurl")
+logger = getLogger("unfurl")
 
 _basepath = os.path.abspath(os.path.dirname(__file__))
 
 
-def relabel_dict(context, localEnv, key):
-    connections = context.get(key)
+def relabel_dict(environment: Dict, localEnv: "LocalEnv", key: str) -> Dict:
+    """Retrieve environment dictionary and remap any values that are strings in the dictionary by treating them as keys into an environment.
+    """
+    connections = environment.get(key)
     if not connections:
         return {}
-    contexts = {}
+    assert isinstance(connections, dict)
+    environments: Dict[str, Dict] = {}
     if localEnv:
         project = localEnv.project or localEnv.homeProject
         if project:
-            contexts = project.contexts
+            environments = project.contexts  # type: ignore
 
     # handle items like newname : oldname to rename merged connections
     def follow_alias(v):
         if isinstance(v, str):
             env, sep, name = v.partition(":")
             if sep:  # found a ":"
-                v = contexts[env][key][name]
+                v = environments[env][key][name]
             else:  # look in current dict
-                v = connections[env]
+                v = connections[env]  # type: ignore
             return follow_alias(v)  # follow
         else:
             return v
@@ -64,36 +73,55 @@ def relabel_dict(context, localEnv, key):
     return dict((n, follow_alias(v)) for n, v in connections.items())
 
 
+class ChangeRecordRecord(ChangeRecord):
+    target: str = ""
+    operation: str = ""
+    dependencies: Sequence = ()
+
+
 class Manifest(AttributeManager):
     """
     Base class for managing an ensemble.
     Derived classes handle loading and serialization.
     """
 
-    rootResource = None
+    rootResource: Optional[TopologyInstance] = None
 
-    def __init__(self, path, localEnv=None):
+    def __init__(self, path: Optional[str], localEnv: Optional["LocalEnv"]=None):
         super().__init__(yaml)
         self.localEnv = localEnv
         self.path = path
         self.repo = self._find_repo()
         self.currentCommitId = self.repo and self.repo.revision
-        self.revisions = RevisionManager(self)
-        self.changeSets = None
+        # self.revisions = RevisionManager(self)
+        self.changeSets: Optional[Dict[str, ChangeRecordRecord]] = None
         self.tosca = None
         self.specDigest = None
-        self.repositories = {}
+        self.repositories: Dict[str, RepoView] = {}
+        self.package_specs: Dict[str, PackageSpec] = {}
+        self.packages: PackagesType = {}
         if self.localEnv:
             # before we start parsing the manifest, add the repositories in the environment
             self._add_repositories_from_environment()
         self.imports = Imports()
         self.imports.manifest = self
-        self._importedManifests = {}
+        self._importedManifests: Dict = {}
 
     def _add_repositories_from_environment(self):
         assert self.localEnv
         context = self.localEnv.get_context()
-        repositories = relabel_dict(context, self.localEnv, "repositories")
+        repositories = {}
+        for key, value in relabel_dict(context, self.localEnv, "repositories").items():
+            if "." in key:  # assume it's a package not a repository name
+                assert isinstance(value, dict)
+                self.package_specs[key] = PackageSpec(key, value.get("url"), value.get("revision"))
+            else:
+                repositories[key] = value
+        env_package_spec = context.get("variables", {}).get(
+            "UNFURL_PACKAGE_RULES", os.getenv("UNFURL_PACKAGE_RULES"))
+        if env_package_spec:
+            for key, value in taketwo(env_package_spec.split()):
+                self.package_specs[key] = PackageSpec(key, value, None)
         resolver = self.get_import_resolver(self)
         for name, tpl in repositories.items():
             toscaRepository = resolver.get_repository(name, tpl)
@@ -111,13 +139,15 @@ class Manifest(AttributeManager):
         )
         self.specDigest = self.get_spec_digest(spec)
 
-    def _find_repo(self):
+    def _find_repo(self) -> Optional["GitRepo"]:
         # check if this path exists in the repo
         repo = self.localEnv and self.localEnv.instanceRepo
         if repo:
-            path = repo.find_path(self.path)[0]
-            if path and (path, 0) in repo.repo.index.entries:
-                return repo
+            # this check is expensive and not that important so skip
+            return repo
+            # path = repo.find_path(self.path)[0]
+            # if path and (path, 0) in repo.repo.index.entries:
+            #     return repo
         return None
 
     def _load_spec(self, spec, path, repositories, more_spec, skip_validation=False):
@@ -143,8 +173,9 @@ class Manifest(AttributeManager):
                 more_spec,
                 replaceKeys=["node_templates", "relationship_templates"],
             )
-        if not isinstance(toscaDef, CommentedMap):
-            toscaDef = CommentedMap(toscaDef.items())
+        yaml_dict_cls = yaml_dict_type(bool(self.localEnv and self.localEnv.readonly))
+        if not isinstance(toscaDef, yaml_dict_cls):
+            toscaDef = yaml_dict_cls(toscaDef.items())
         if getattr(toscaDef, "base_dir", None) and (
             not path or toscaDef.base_dir != os.path.dirname(path)
         ):
@@ -156,6 +187,7 @@ class Manifest(AttributeManager):
 
     def get_spec_digest(self, spec):
         m = hashlib.sha1()  # use same digest function as git
+        assert self.tosca
         t = self.tosca.template
         for tpl in [spec, t.topology_template.custom_defs, t.nested_tosca_tpls]:
             m.update(json.dumps(tpl, sort_keys=True).encode("utf-8"))
@@ -176,7 +208,7 @@ class Manifest(AttributeManager):
         return "."
 
     @property
-    def loader(self):
+    def loader(self) -> Optional["AnsibleLoader"]:
         return (
             self.rootResource
             and self.rootResource.templar
@@ -188,12 +220,14 @@ class Manifest(AttributeManager):
 
     def load_template(self, name, lastChange=None):
         if lastChange:
-            try:
-                return self.revisions.get_revision(lastChange).tosca.get_template(name)
-            except:
-                return None
+            return None
+            # XXX implement revisions
+            # try:
+            #     return self.revisions.get_revision(lastChange).tosca.get_template(name)
+            # except Exception:
+            #     return None
         else:
-            return self.tosca.get_template(name)
+            return self.tosca and self.tosca.get_template(name)
 
     #  load instances
     #    create a resource with the given template
@@ -236,20 +270,22 @@ class Manifest(AttributeManager):
                 ]
         return resourceChanges
 
-    def load_config_change(self, changeSet):
+    def load_config_change(self, changeSet: dict) -> ChangeRecordRecord:
         """
         Reconstruct the Configuration that was applied in the past
         """
         from .configurator import Dependency
 
-        configChange = ConfigChange()
+        configChange = ChangeRecordRecord()
         Manifest.load_status(changeSet, configChange)
         configChange.changeId = changeSet.get("changeId", 0)
         configChange.previousId = changeSet.get("previousId")
-        configChange.target = changeSet.get("target")
-        configChange.operation = changeSet.get("implementation", {}).get("operation")
+        configChange.target = changeSet.get("target", "")
+        assert isinstance(configChange.target, str)
+        configChange.operation = changeSet.get("implementation", {}).get("operation", "")
+        assert isinstance(configChange.operation, str)
 
-        configChange.inputs = changeSet.get("inputs")
+        configChange.inputs = changeSet.get("inputs")  # type: ignore
         # 'digestKeys', 'digestValue' but configurator can set more:
         for key in changeSet.keys():
             if key.startswith("digest"):
@@ -269,12 +305,12 @@ class Manifest(AttributeManager):
             )
 
         if "changes" in changeSet:
-            configChange.resourceChanges = self.load_resource_changes(
+            configChange.resourceChanges = self.load_resource_changes(  # type: ignore
                 changeSet["changes"]
             )
 
-        configChange.result = changeSet.get("result")
-        configChange.messages = changeSet.get("messages", [])
+        configChange.result = changeSet.get("result")  # type: ignore
+        configChange.messages = changeSet.get("messages", [])  # type: ignore
 
         # XXX
         # ('action', ''),
@@ -302,7 +338,7 @@ class Manifest(AttributeManager):
     #         postConditions=spec.get("postConditions"),
     #     )
 
-    def _create_requirement(self, key, val):
+    def _create_requirement(self, key, val) -> RelationshipInstance:
         # parent will be the capability, should have already been created
         capabilityId = val.get("capability")
         if not capabilityId:
@@ -316,7 +352,7 @@ class Manifest(AttributeManager):
             capability._relationships = []
         return self._create_entity_instance(RelationshipInstance, key, val, capability)
 
-    def create_node_instance(self, rname, resourceSpec, parent=None):
+    def create_node_instance(self, rname, resourceSpec, parent=None) -> NodeInstance:
         # if parent property is set it overrides the parent argument
         pname = resourceSpec.get("parent")
         if pname:
@@ -335,8 +371,9 @@ class Manifest(AttributeManager):
             for req in resourceSpec["requirements"]:
                 key, val = next(iter(req.items()))
                 requirement = self._create_requirement(key, val)
-                requirement.source = resource
-                resource.requirements.append(requirement)
+                requirement._source = resource
+                # XXX investigate and re-enable this assert
+                assert requirement in resource.requirements, f"{requirement} not in {resource.requirements} for {rname}"
 
         if resourceSpec.get("artifacts"):
             for key, val in resourceSpec["artifacts"].items():
@@ -352,7 +389,7 @@ class Manifest(AttributeManager):
             return None
         if not self.changeSets:  # XXX load changesets if None
             return None
-        jobId = ConfigChange.get_job_id(operational.last_config_change)
+        jobId = ChangeRecord.get_job_id(operational.last_config_change)
         return self.changeSets.get(jobId)
 
     def _create_entity_instance(self, ctor, name, status, parent):
@@ -413,10 +450,14 @@ class Manifest(AttributeManager):
                 created = ""
             output.append(f"{' ' * indent}{instance} {status} {state} {created}")
             indent += 4
-            for child in instance.instances:
-                summary(child, indent)
+            if isinstance(instance, HasInstancesInstance):
+                for rel in instance.requirements:
+                    if rel.local_status is not None:
+                        summary(rel, indent)
+                for child in instance.instances:
+                    summary(child, indent)
 
-        output = []
+        output: List[str] = []
         summary(self.rootResource, 0)
         return "\n".join(output)
 
@@ -429,7 +470,7 @@ class Manifest(AttributeManager):
             return self.localEnv.find_path_in_repos(path, importLoader)
         elif self.repo:
             repo = self.repo
-            filePath = repo.findRepoPath(path)
+            filePath = repo.find_repo_path(path)
             if filePath is not None:
                 return repo, filePath, repo.revision, False
         return None, None, None, None
@@ -446,8 +487,9 @@ class Manifest(AttributeManager):
         )
         return repo, filePath, revision, bare
 
-    def add_repository(self, repo, toscaRepository, file_name):
-        repository: RepoView = self.repositories.get(toscaRepository.name)
+    def add_repository(self, repo: Optional[GitRepo], toscaRepository: Repository, file_name: str) -> RepoView:
+        # add or replace the repository
+        repository: Optional[RepoView] = self.repositories.get(toscaRepository.name)
         if repository:
             # already exist, make sure it's the same repo
             repo = repo or repository.repo
@@ -457,13 +499,14 @@ class Manifest(AttributeManager):
                 raise UnfurlError(
                     f'Repository "{toscaRepository.name}" already defined'
                 )
-        self.repositories[toscaRepository.name] = RepoView(
+        repository = RepoView(
             toscaRepository, repo, file_name
         )
+        self.repositories[toscaRepository.name] = repository
         return repository
 
-    def _update_repositories(self, config, inlineRepositories=None, resolver=None):
-        # _update_repositories is called during parse time when including files
+    def _update_repositories(self, config, inlineRepositories=None, resolver: Optional[ImportResolver] = None) -> Dict[str, dict]:
+        # Called during parse time when including files
         if not resolver:
             resolver = self.get_import_resolver(self)
         # we need to fetch this every call since the config might have changed:
@@ -487,7 +530,7 @@ class Manifest(AttributeManager):
         return self._set_repositories()
 
     @staticmethod
-    def _get_repositories(tpl):
+    def _get_repositories(tpl) -> Dict:
         repositories = ((tpl.get("spec") or {}).get("service_template") or {}).get(
             "repositories"
         ) or {}
@@ -495,7 +538,7 @@ class Manifest(AttributeManager):
         repositories.update((tpl.get("environment") or {}).get("repositories") or {})
         return repositories
 
-    def _set_repositories(self):
+    def _set_repositories(self) -> Dict[str, Dict]:
         repositories = self.repositories
         if "unfurl" not in repositories:
             # add a repository that points to this package
@@ -504,6 +547,7 @@ class Manifest(AttributeManager):
                 None,
                 "",
             )
+            repository.package = False
             repositories["unfurl"] = repository
         if "self" not in repositories:
             # this is called too early to use self.getBaseDir()
@@ -518,23 +562,26 @@ class Manifest(AttributeManager):
                 self.repo,
                 path,
             )
+            repository.package = False
             repositories["self"] = repository
 
         inProject = False
         if self.localEnv and self.localEnv.project:
             if self.localEnv.project is self.localEnv.homeProject:
-                inProject = self.localEnv.project.projectRoot in self.path
+                inProject = self.localEnv.project.projectRoot in self.path  # type: ignore
             else:
                 inProject = True
         if inProject and "project" not in repositories:
-            repositories["project"] = self.localEnv.project.project_repoview
+            repositories["project"] = self.localEnv.project.project_repoview  # type: ignore
 
         if "spec" not in repositories:
             # if not found assume it points the project root or self if not in a project
             if inProject:
-                repositories["spec"] = self.localEnv.project.project_repoview
+                assert self.localEnv and self.localEnv.project
+                repositories["spec"] = self.localEnv.project.project_repoview  # type: ignore
             else:
                 repositories["spec"] = repositories["self"]
+        repositories["spec"].package = False
         return {name: repo.repository.tpl for name, repo in self.repositories.items()}
 
     def load_yaml_include(
@@ -577,7 +624,13 @@ class Manifest(AttributeManager):
                 return reponame in repositories
             return True
 
-        artifact = ArtifactSpec(artifactTpl, path=baseDir)
+        if not artifactTpl["file"]:
+            msg = f"document include {templatePath} missing file (base: {baseDir})"
+            if warnWhenNotFound:
+                logger.warning(msg)
+                return "", None
+            raise UnfurlError(msg)
+
         tpl = CommentedMap()
         resolver = self.get_import_resolver(warnWhenNotFound, config=expanded)
         tpl["repositories"] = self._update_repositories(
@@ -585,33 +638,49 @@ class Manifest(AttributeManager):
         )
         loader = toscaparser.imports.ImportsLoader(
             None,
-            artifact.base_dir,
+            get_base_dir(baseDir),
             tpl=tpl,
             resolver=resolver,
         )
-        path, doc = loader._load_import_template(None, artifact.as_import_spec())
+        import_spec = dict(file=artifactTpl["file"], repository=artifactTpl.get("repository"))
+        path, doc = loader._load_import_template(None, import_spec)
         if doc is None:
             logger.warning(
                 f"document include {templatePath} does not exist (base: {baseDir})"
             )
         return path, doc
 
-    def get_import_resolver(self, ignoreFileNotFound=False, expand=False, config=None):
+    def get_import_resolver(self, ignoreFileNotFound=False, expand=False, config=None) -> ImportResolver:
         return ImportResolver(self, ignoreFileNotFound, expand, config)
 
+    def last_commit_time(self) -> float:
+        # return seconds (0 if not found)
+        repo = self.repo
+        if not repo:
+            return 0
+        try:
+            # find the revision that last modified this file before or equal to the current revision
+            # (use current revision to handle branches)
+            commits = list(repo.repo.iter_commits(repo.revision, self.path or "", max_count=1))
+        except ValueError:
+            return 0
+        if commits:
+            return commits[0].committed_date
+        return 0
 
-class SnapShotManifest(Manifest):
-    def __init__(self, manifest, commitId):
-        super().__init__(manifest.path, self.localEnv)
-        self.commitId = commitId
-        oldManifest = manifest.repo.show(manifest.path, commitId)
-        self.repo = manifest.repo
-        self.localEnv = manifest.localEnv
-        self.manifest = YamlConfig(
-            oldManifest, manifest.path, loadHook=self.load_yaml_include
-        )
-        self._update_repositories(oldManifest)
-        expanded = self.manifest.expanded
-        # just needs the spec, not root resource
-        self._set_spec(expanded.get("spec", {}))
-        self._ready(None)
+# unused
+# class SnapShotManifest(Manifest):
+#     def __init__(self, manifest, commitId):
+#         super().__init__(manifest.path, self.localEnv)
+#         self.commitId = commitId
+#         oldManifest = manifest.repo.show(manifest.path, commitId)
+#         self.repo = manifest.repo
+#         self.localEnv = manifest.localEnv
+#         self.manifest = YamlConfig(
+#             oldManifest, manifest.path, loadHook=self.load_yaml_include
+#         )
+#         self._update_repositories(oldManifest)
+#         expanded = self.manifest.expanded
+#         # just needs the spec, not root resource
+#         self._set_spec(expanded.get("spec", {}))
+#         self._ready(None)
