@@ -1,9 +1,9 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
-from typing import Optional
+from typing import Dict, List, Optional
 import six
 
-from .runtime import NodeInstance
+from .runtime import InstanceKey, NodeInstance, EntityInstance
 from .util import UnfurlError
 from .result import ChangeRecord
 from .support import Status, NodeState, Reason
@@ -17,8 +17,10 @@ from .planrequests import (
     ConfigurationSpec,
     find_parent_resource,
     find_resources_from_template_name,
+    _find_implementation,
+    get_success_status,
 )
-from .tosca import find_standard_interface, EntitySpec
+from .tosca import NodeSpec, Workflow, find_standard_interface, EntitySpec, ToscaSpec
 from .logs import getLogger
 
 
@@ -46,16 +48,6 @@ def is_external_template_compatible(
     return False
 
 
-def get_success_status(workflow):
-    if workflow == "deploy":
-        return Status.ok
-    elif workflow == "stop":
-        return Status.pending
-    elif workflow == "undeploy":
-        return Status.absent
-    return None
-
-
 class Plan:
     @staticmethod
     def get_plan_class_for_workflow(workflow):
@@ -70,7 +62,7 @@ class Plan:
 
     interface: Optional[str] = None
 
-    def __init__(self, root, toscaSpec, jobOptions):
+    def __init__(self, root, toscaSpec: ToscaSpec, jobOptions):
         self.jobOptions = jobOptions
         self.workflow = jobOptions.workflow
         self.root = root
@@ -185,7 +177,6 @@ class Plan:
                 # Operation to pre-configure the target endpoint.
                 for relationship in capability.relationships:
                     # we're the target, source may not have been created yet
-                    # XXX if not relationship.source create the instance
                     yield from self._run_operation(
                         NodeState.configuring,
                         "Configure.pre_configure_target",
@@ -211,6 +202,7 @@ class Plan:
         )
 
         if sourceConfigOps:
+            # we're the source and we just ran configure, now configure any relationships
             for requirement in resource.requirements:
                 yield from self._run_operation(
                     NodeState.configuring,
@@ -286,9 +278,6 @@ class Plan:
     def execute_default_undeploy(self, resource, reason=None, inputs=None):
         # XXX run check if joboption set?
         # XXX don't delete if dirty
-        # XXX remove_target: Operation called on source when a target instance is removed
-        # XXX remove_source: Operation called on target when a source instance is removed
-
         if (
             resource.state in [NodeState.starting, NodeState.started]
             or self.workflow == "stop"
@@ -301,6 +290,32 @@ class Plan:
         if self.workflow == "stop":
             return
 
+        targetConfigOps = resource.template.get_capability_interfaces()
+        # test for targetConfigOps to avoid creating unnecessary instances
+        if targetConfigOps:
+            # if there are relationships targeting us we need to remove them first
+            for capability in resource.capabilities:
+                for relationship in capability.relationships:
+                    # we're the target that is about to be deleted
+                    yield from self._run_operation(
+                        NodeState.configuring,
+                        "Configure.remove_target",
+                        relationship,
+                        reason,
+                    )
+
+        sourceConfigOps = resource.template.get_requirement_interfaces()
+        if sourceConfigOps:
+            # before we are deleted, remove any relationships we have
+            if resource.template.get_requirement_interfaces():
+                for relationship in resource.requirements:
+                    yield from self._run_operation(
+                        NodeState.configuring,
+                        "Configure.remove_source",
+                        relationship,
+                        reason,
+                    )
+
         if resource.created or self.jobOptions.destroyunmanaged:
             nodeState = NodeState.deleting
             op = "Standard.delete"
@@ -311,12 +326,39 @@ class Plan:
         yield from self._run_operation(nodeState, op, resource, reason, inputs)
 
     def execute_default_install_op(self, operation, resource, reason=None, inputs=None):
+        # add target or add source
         req = create_task_request(
             self.jobOptions, "Install." + operation, resource, reason, inputs
         )
         if req:
             yield req
 
+        if operation != "check":
+            # we're the source, target has already started
+            sourceConfigOps = resource.template.get_requirement_interfaces()
+            if sourceConfigOps:
+                if resource.template.get_requirement_interfaces():
+                    # we're the source
+                    for relationship in resource.requirements:
+                        req = create_task_request(
+                            self.jobOptions, "Configure.add_source", relationship, reason, inputs
+                        )
+                        if req:
+                            yield req
+
+            targetConfigOps = resource.template.get_capability_interfaces()
+            # test for targetConfigOps to avoid creating unnecessary instances
+            if targetConfigOps:
+                for capability in resource.capabilities:
+                    for relationship in capability.relationships:
+                        # we're the target
+                        req = create_task_request(
+                            self.jobOptions, "Configure.add_target", relationship, reason, inputs
+                        )
+                        if req:
+                            yield req
+
+
     def generate_delete_configurations(self, include):
         for resource in get_operational_dependents(self.root):
             # reverse to teardown leaf nodes first
@@ -365,24 +407,28 @@ class Plan:
     def _get_connection_task(self, taskRequest):
         if not self._checked_connection_task:
             self._checked_connection_task = True
-            if self.tosca.topology.primary_provider:
+            if self.tosca.topology and self.tosca.topology.primary_provider:
                 for rel in self.root.requirements:
                     if rel.name == "primary_provider":
                         yield from self.execute_default_install_op("check", rel)
                         return
                 assert False, self.root.requirements
 
-    def _generate_configurations(self, resource, reason, workflow=None):
+    def _generate_configurations(self, resource: "EntityInstance", reason: str, workflow: Optional[str] = None):
+        # note: workflow parameter might be an installOp
         workflow = workflow or self.workflow
         # check if this workflow has been delegated to one explicitly declared
         configGenerator = self.execute_workflow(workflow, resource)
-        if not configGenerator:
+        if configGenerator:
+            custom_workflow = True
+        else:
             configGenerator = self._get_default_generator(workflow, resource, reason)
             if not configGenerator:
                 raise UnfurlError("can not get default for workflow " + workflow)
+            custom_workflow = False
 
         # if the workflow is one that can modify a target, create a TaskRequestGroup
-        if get_success_status(workflow):
+        if workflow != "check":
             group = TaskRequestGroup(resource, workflow)
         else:
             group = None
@@ -392,7 +438,10 @@ class Plan:
                 task_found = True
                 yield from self._get_connection_task(taskRequest)
                 if group and not isinstance(taskRequest, JobRequest):
-                    group.children.append(taskRequest)
+                    # if taskRequest.target == group.target:
+                    group.add_request(taskRequest)
+                    if isinstance(taskRequest, TaskRequest):
+                        taskRequest.set_final_for_workflow(bool(self.is_last_workflow_op(taskRequest)))
                 else:
                     yield taskRequest
         if not task_found:
@@ -400,9 +449,11 @@ class Plan:
                 f'No operations for workflow "{workflow}" defined for instance "{resource.name}"'
             )
         if group:
+            if custom_workflow:
+                group.set_final_for_workflow(True)
             yield group
 
-    def execute_workflow(self, workflowName, resource):
+    def execute_workflow(self, workflowName: str, resource):
         workflow = self.tosca.get_workflow(workflowName)
         if not workflow:
             return None
@@ -423,7 +474,7 @@ class Plan:
         finally:
             pass  # pop _workflow_inputs
 
-    def execute_steps(self, workflow, steps, resource):
+    def execute_steps(self, workflow: Workflow, steps: List, resource: EntityInstance):
         queue = steps[:]
         while queue:
             step = queue.pop()
@@ -445,9 +496,9 @@ class Plan:
             except StopIteration:
                 pass
 
-    def execute_step(self, step, resource, workflow):
+    def execute_step(self, step, resource: EntityInstance, workflow: Workflow):
         logger.debug("executing step %s for %s", step.name, resource.name)
-        reqGroup = TaskRequestGroup(resource, workflow)
+        reqGroup = TaskRequestGroup(resource, workflow.workflow.name)
         for activity in step.activities:
             if activity.type == "inline":
                 # XXX inputs
@@ -456,7 +507,7 @@ class Plan:
                     continue
                 for result in workflowGenerator:
                     if result:
-                        reqGroup.children.append(result)
+                        reqGroup.add_request(result)
             elif activity.type == "call_operation":
                 # XXX need to pass operation_host (see 3.6.27 Workflow step definition p188)
                 # if target is a group can be value can be node_type or node template name
@@ -469,9 +520,9 @@ class Plan:
                     activity.inputs,
                 )
                 if req:
-                    reqGroup.children.append(req)
+                    reqGroup.add_request(req)
             elif activity.type == "set_state":
-                reqGroup.children.append(SetStateRequest(resource, activity.set_state))
+                reqGroup.add_request(SetStateRequest(resource, activity.set_state))
             elif activity.type == "delegate":
                 # XXX inputs
                 configGenerator = self._get_default_generator(
@@ -479,7 +530,8 @@ class Plan:
                 )
                 if not configGenerator:
                     continue
-                reqGroup.children.extend(filter(None, configGenerator))
+                for req in filter(None, configGenerator):
+                    reqGroup.add_request(req)
 
         # XXX  yield step.on_failure  # list of steps
         yield reqGroup
@@ -507,6 +559,9 @@ class Plan:
     def include_not_found(self, template):
         return True
 
+    def is_last_workflow_op(self, taskrequest: TaskRequest) -> str:
+        return ""
+
     def _generate_workflow_configurations(self, instance, oldTemplate):
         yield from self._generate_configurations(instance, self.workflow)
 
@@ -685,6 +740,21 @@ class DeployPlan(Plan):
         else:
             yield from self._generate_configurations(instance, reason)
 
+    def is_last_workflow_op(self, taskrequest: TaskRequest) -> str:
+        interface = taskrequest.configSpec.interface
+        if interface == "Standard":
+            order = ["create", "configure", "start"]
+        elif interface == "Configure":
+            order = ["pre_configure_target", "pre_configure_source", "post_configure_target", "post_configure_source"]
+        else:
+            return ""
+        implemented = [op for op in order if _find_implementation(interface, op, taskrequest.target.template)]
+        if taskrequest.configSpec.operation in implemented:
+            index = implemented.index(taskrequest.configSpec.operation)
+            if len(implemented) == index + 1:
+                return implemented[index]
+        return ""
+
 
 class UndeployPlan(Plan):
     def execute_plan(self):
@@ -700,6 +770,24 @@ class UndeployPlan(Plan):
         # return value is used as "reason"
         return self.workflow
 
+    def is_last_workflow_op(self, taskrequest: TaskRequest) -> str:
+        interface = taskrequest.configSpec.interface
+        if interface == "Standard":
+            if self.workflow == "stop":
+                order = ["stop"]
+            else:
+                order = ["stop", "delete"]
+        elif interface == "Configure":
+            order = ["remove_target", "remove_source"]
+        else:
+            return ""
+        implemented = [op for op in order if _find_implementation(interface, op, taskrequest.target.template)]
+        if taskrequest.configSpec.operation in implemented:
+            index = implemented.index(taskrequest.configSpec.operation)
+            if len(implemented) == index + 1:
+                return implemented[index]
+        return ""
+
 
 class ReadOnlyPlan(Plan):
     interface = "Install"
@@ -721,7 +809,7 @@ class WorkflowPlan(Plan):
             if self.tosca.is_type_name(step.target):
                 templates = self.tosca.find_matching_templates(step.target)
             else:
-                template = self.tosca.findTemplate(step.target)
+                template = self.tosca.nodeTemplates.get(step.target)
                 if not template:
                     continue
                 templates = [template]
@@ -837,8 +925,7 @@ def interface_requirements_ok(spec, template):
     return True
 
 
-def order_templates(templates, filter=None, interface=None):
-    # templates is dict of NodeSpecs
+def order_templates(templates: Dict[str, NodeSpec], filter=None, interface=None):
     seen = set()
     for source in templates.values():
         if filter and source.name != filter:
@@ -860,7 +947,7 @@ def order_templates(templates, filter=None, interface=None):
             if spec:
                 if spec is not source:
                     # ancestor is required by source
-                    spec._isReferencedBy.append(source)
+                    spec._isReferencedBy.append(source)  # type: ignore
                 if spec in seen:
                     continue
                 seen.add(spec)
