@@ -2,13 +2,30 @@
 # SPDX-License-Identifier: MIT
 import collections
 import re
-from typing import Iterator, List, Optional, Sequence, Tuple
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Dict,
+    Iterator,
+    List,
+    Mapping,
+    Optional,
+    Sequence,
+    Set,
+    Tuple,
+    Union,
+    cast,
+)
 import six
 import shlex
 import sys
 import os
 import os.path
 
+if TYPE_CHECKING:
+    from .job import Job, ConfigTask, JobOptions
+    from .configurator import Dependency
+
 from .util import (
     lookup_class,
     load_module,
@@ -17,12 +34,13 @@ from .util import (
     UnfurlTaskError,
     to_enum,
 )
-from .result import serialize_value
-from .support import Defaults, NodeState, Priority
-from .runtime import EntityInstance
+from .result import Result, ResultsList, serialize_value
+from .support import Defaults, NodeState, Priority, Status
+from .runtime import EntityInstance, InstanceKey
+from .logs import getLogger
 import logging
 
-logger = logging.getLogger("unfurl")
+logger = getLogger("unfurl")
 
 
 # we want ConfigurationSpec to be independent of our object model and easily serializable
@@ -111,6 +129,7 @@ class ConfigurationSpec:
         if not klass:
             raise UnfurlError(f"Could not load configurator {self.className}")
         else:
+            assert callable(klass)
             return klass(self)
 
     def should_run(self):
@@ -143,20 +162,31 @@ class ConfigurationSpec:
 
 
 class PlanRequest:
-    error = None
-    future_dependencies: list = []
-    task = None
-    render_errors = None
+    error: Union[bool, str] = False
+    task: Optional["ConfigTask"] = None
+    render_errors: Optional[List[UnfurlTaskError]] = None
+    completed = False
+    group: Optional["TaskRequestGroup"] = None
+    required: Optional[bool] = None
+    is_final_for_workflow: Optional[bool] = None
 
     def __init__(self, target: "EntityInstance"):
         self.target = target
+        self.dependencies: List["Dependency"] = []
+
+    def set_error(self, msg: str):
+        self.error = msg
+        if self.task:
+            self.task.local_status = Status.error
+            self.task.result = msg
 
     @property
-    def root(self):
+    def root(self) -> Optional["EntityInstance"]:
         return self.target.root if self.target else None
 
-    def update_future_dependencies(self, completed):
-        return self.future_dependencies
+    def add_dependencies(self, dependencies: List["Dependency"]) -> List["Dependency"]:
+        self.dependencies.extend(dependencies)
+        return self.dependencies
 
     def get_operation_artifacts(self):
         return []
@@ -176,34 +206,112 @@ class PlanRequest:
             return True
         return False
 
-    def get_unfulfilled_refs(self):
+    def get_unfulfilled_refs(self, check_target="") -> Iterator["Dependency"]:
+        for dep in self.dependencies:
+            if check_target:
+                if not dep.validate():
+                    yield dep
+                if (
+                    check_target == "operational"
+                    and dep.target
+                    and dep.target != self.target
+                ):
+                    if dep.target.local_status not in [Status.ok, Status.degraded]:
+                        logger.trace(
+                            f"{dep} not operational with local status {dep.target.local_status and dep.target.local_status.name}"
+                        )
+                        yield dep
+                # otherwise, dep is fulfilled
+            else:
+                yield dep
         if not self.render_errors:
             return
         for error in self.render_errors:
-            dep = getattr(error, "dependency", None)
-            if dep:
-                yield dep
+            # result.py::_validation_error() creates a UnfurlTaskError with a Dependency
+            more_dep = cast(Optional["Dependency"], getattr(error, "dependency", None))
+            if more_dep:
+                yield more_dep
 
-    def update_unfulfilled_errors(self):
-        self.render_errors = list(self._get_unfulfilled_errors())
+    def update_unfulfilled_errors(self) -> bool:
+        # remove unfulfilled dependencies and errors caused by an unfulfilled dependency
+        self.render_errors = list(self._revalidate_unfulfilled_errors())
+        # note: this isn't called during the last round because we don't want the has_changed() check
+        self.dependencies = [
+            dep
+            for dep in self.dependencies
+            if not dep.validate() or not dep.has_changed()
+        ]
+        # logger.trace(
+        #     f"update_unfulfilled_errors {self.target.name}: {self.dependencies}"
+        # )
         return self.has_unfulfilled_refs()
 
-    def _get_unfulfilled_errors(self):
+    def _revalidate_unfulfilled_errors(self) -> Iterator["UnfurlTaskError"]:
         if not self.render_errors:
             return
         for error in self.render_errors:
-            dep = getattr(error, "dependency", None)
+            dep = cast(Optional["Dependency"], getattr(error, "dependency", None))
             if not dep or not dep.validate():
+                # if dep:
+                #     logger.trace(
+                #         f"in render_errors: invalid dep {dep} for {self.target.name}"
+                #     )
                 yield error
 
     @property
-    def not_ready(self):
-        return self.future_dependencies or self.has_unfulfilled_refs()
+    def previous(self) -> Optional["PlanRequest"]:
+        return self.group and self.group.get_previous(self) or None
+
+    @property
+    def not_ready(self) -> bool:
+        if self.previous and self.previous.not_ready:
+            return True
+        return bool(self.has_unfulfilled_refs())
 
     @property
     def name(self):
         return type(self).__name__
 
+    def get_notready_message(self) -> str:
+        start = "Never ran:"
+        msg = start
+        if self.render_errors:
+            render_deps = [
+                error.dependency.name  # type: ignore
+                for error in self.render_errors
+                if getattr(error, "dependency", None)
+            ]
+            msg += f" invalid dependencies: {render_deps}"
+        if self.dependencies:
+            not_operational = [
+                dep.name
+                for dep in self.dependencies
+                if dep.target
+                and dep.target.local_status not in [Status.ok, Status.degraded]
+            ]
+            invalid_deps = [
+                dep.name for dep in self.dependencies if dep not in not_operational
+            ]
+            if msg != start:
+                msg += " and "
+            if not_operational:
+                msg += f"non-operational dependencies: {not_operational}."
+            if msg != start:
+                msg += " and "
+            if invalid_deps:
+                msg += f" unfulfilled dependencies: {invalid_deps}."
+        if msg != start:
+            return msg
+        if self.group and self.group.target.name != self.target.name:
+            return f"Never ran: parent resource in error: {self.group.target.name}"
+        elif self.previous:
+            return f'Never ran: previous operation "{self.previous}" failed or could not run'
+        else:
+            return "Couldn't run"
+
+    def set_final_for_workflow(self, is_final: bool):
+        self.is_final_for_workflow = is_final
+
     def __str__(self) -> str:
         return f'Planned task {self.name} for "{self.target.name}"'
 
@@ -215,12 +323,12 @@ class TaskRequest(PlanRequest):
 
     def __init__(
         self,
-        configSpec,
-        target,
-        reason,
-        persist=False,
-        required=None,
-        startState=None,
+        configSpec: ConfigurationSpec,
+        target: EntityInstance,
+        reason: str,
+        persist: bool = False,
+        required: Optional[bool] = None,
+        startState: Optional[NodeState] = None,
     ):
         super().__init__(target)
         self.configSpec = configSpec
@@ -230,6 +338,63 @@ class TaskRequest(PlanRequest):
         self.error = configSpec.name == "#error"
         self.startState = startState
         self.task = None
+        self._completed = False
+
+    def __completed():  # type: ignore
+        def fget(self) -> bool:
+            return bool(self._completed or self.task and self.task.completed)
+
+        def fset(self, value: bool):
+            self._completed = value
+
+        return locals()
+
+    completed: bool = property(**__completed())  # type: ignore
+
+    def reassign_final_for_workflow(self) -> Optional["TaskRequest"]:
+        req = self
+        group = req.group
+        if group and req.is_final_for_workflow:
+            previous = group.get_previous(req)
+            while previous:
+                if (
+                    previous.target == req.target
+                    and previous.required != False
+                    and isinstance(previous, TaskRequest)
+                ):
+                    previous.is_final_for_workflow = True
+                    return previous
+                previous = group.get_previous(previous)
+        return None
+
+    def finish_workflow(self) -> None:
+        # This is called on the last task apply to a resource in a work flow.
+        # If the all the task succeeded and if target hasn't had its status set already,
+        # set the target status to workflow's desired status.
+        task = self.task
+        assert task
+        assert task.completed  # task.finished() has been called
+        if task.local_status == Status.error:
+            # if failed then task._update_status() set error state on the target
+            return
+        workflow = self.group and self.group.workflow
+        if not workflow:  # not in a group or not in a mutable workflow
+            return
+        explicit_status = task.result and task.result.status
+        if explicit_status is None:
+            # even though we didn't modify the target this is the last op and the task succeeded
+            # so assume the target is that the workflow's final state
+            explicit_status = get_success_status(workflow)
+            # target's status needs to change
+        task.logger.trace(
+            "finish workflow with %s status %s for workflow %s with local_status %s created: %s",
+            task.target.name,
+            "None" if explicit_status is None else explicit_status.name,
+            workflow,
+            task.target.local_status,  # type: ignore
+            task.target.created,
+        )
+        task._finished_workflow(explicit_status, workflow)
 
     def _get_artifact_plan(self, artifact):
         # the artifact has an interface so it needs to be installed on the operation_host
@@ -305,12 +470,6 @@ class TaskRequest(PlanRequest):
             return name + " (reason: " + self.reason + ")"
         return name
 
-    def update_future_dependencies(self, completed):
-        self.future_dependencies = [
-            fr for fr in self.future_dependencies if fr not in completed
-        ]
-        return self.future_dependencies
-
     def _summary_dict(self, include_rendered=True):
         summary = dict(
             operation=self.configSpec.operation or self.configSpec.name,
@@ -343,25 +502,25 @@ class SetStateRequest(PlanRequest):
     def _summary_dict(self):
         return dict(set_state=self.set_state)
 
+    def set_final_for_workflow(self, is_final: bool):
+        assert self.group
+        previous = self.group.get_previous(self)
+        if previous:
+            previous.set_final_for_workflow(is_final)
+
 
 class TaskRequestGroup(PlanRequest):
-    def __init__(self, target, workflow):
+    """
+    A TaskRequestGroup is a list of task requests that need to succeed to advance
+    the target NodeInstance to the workflow's desired status (e.g. ok or absent)
+    It is created by TOSCA's standard operations or by a custom workflow.
+    """
+
+    def __init__(self, target: EntityInstance, workflow: str):
         super().__init__(target)
         self.workflow = workflow
-        self.children = []
-
-    @property
-    def future_dependencies(self):
-        future_dependencies = []
-        for req in self.children:
-            future_dependencies.extend(req.future_dependencies)
-        return future_dependencies
-
-    def update_future_dependencies(self, completed):
-        future_dependencies = []
-        for req in self.children:
-            future_dependencies.extend(req.update_future_dependencies(completed))
-        return future_dependencies
+        self.starting_status = target.local_status
+        self.children: List[PlanRequest] = []
 
     def get_operation_artifacts(self):
         artifacts = []
@@ -369,20 +528,41 @@ class TaskRequestGroup(PlanRequest):
             artifacts.extend(req.get_operation_artifacts())
         return artifacts
 
-    def get_unfulfilled_refs(self):
-        for req in self.children:
-            yield from req.get_unfulfilled_refs()
+    def add_request(self, req: PlanRequest) -> None:
+        self.children.append(req)
+        req.group = self
 
-    def update_unfulfilled_errors(self):
+    def get_previous(self, req) -> Optional[PlanRequest]:
+        siblings = self.children
+        if req in siblings and siblings[0] != req:
+            return siblings[siblings.index(req) - 1]
+        return None
+
+    def has_errors(self):
+        for req in self.children:
+            if req.error or req.task and req.task.local_status == Status.error:
+                return True
+        return False
+
+    def set_final_for_workflow(self, is_final: bool):
+        if self.children:
+            self.children[-1].set_final_for_workflow(is_final)
+
+    # XXX unused
+    def get_unfulfilled_refs(self, check_target=""):
+        for req in self.children:
+            yield from req.get_unfulfilled_refs(check_target)
+
+    def update_unfulfilled_errors(self) -> bool:
         for req in self.children:
             req.update_unfulfilled_errors()
         return self.has_unfulfilled_refs()
 
     def __str__(self) -> str:
         if self.children:
-            return str(self.children)
+            return f"TaskRequestGroup{self.children}"
         else:
-            return f'Planned task {self.workflow} for "{self.target.name}"'
+            return f'Planned task group {self.workflow} for "{self.target.name}"'
 
     def __repr__(self):
         return f"TaskRequestGroup({self.target}:{self.workflow}:{self.children})"
@@ -393,11 +573,19 @@ class JobRequest:
     Yield this to run a child job.
     """
 
-    def __init__(self, resources, errors=None, update=None):
+    def __init__(
+        self,
+        resources: List[EntityInstance],
+        errors: Optional[Sequence[UnfurlError]] = None,
+        update=None,
+    ):
         self.instances = resources
         self.errors = errors or []
         self.update = update
 
+    def set_error(self, msg: str):
+        self.errors.append(UnfurlError(msg))  # type: ignore
+
     def get_instance_specs(self):
         if self.update:
             return [self.update]
@@ -456,47 +644,72 @@ def find_operation_host(target, operation_host):
 
 def get_render_requests(
     requests: Sequence[PlanRequest],
-) -> Iterator[Tuple[Optional[TaskRequestGroup], TaskRequest]]:
+) -> Iterator[PlanRequest]:
     # returns requests that can be rendered grouped by its top-most task group
     for req in requests:
         if isinstance(req, TaskRequestGroup):
-            for parent, child in get_render_requests(req.children):
-                yield req, child  # yields root as parent
-        elif isinstance(req, TaskRequest):
-            yield None, req
-        elif not isinstance(req, SetStateRequest):
-            # note: SetStateRequests don't render, only run
+            for child in get_render_requests(req.children):
+                yield child  # yields root as parent
+        elif isinstance(req, (TaskRequest, SetStateRequest)):
+            yield req
+        else:
             assert not req, f"unexpected type of request: {req}"
 
 
-def _get_deps(parent, req, liveDependencies, requests):
-    previous = None
-    for (root, r) in requests:
+def _get_deps(
+    req: PlanRequest,
+    live_dependencies: Dict[InstanceKey, List["Dependency"]],
+    requests: Sequence[PlanRequest],
+) -> None:
+    # iterate through the live attributes the given task depends
+    # if the attribute's instance might be modified by another task, mark that task as a dependency
+    req.dependencies = []  # reset
+    for r in requests:
         if req.target.key == r.target.key:
             continue  # skip self
+        if r.task and r.task.completed:
+            continue  # already ran
         if req.required is not None and not req.required:
+            logger.trace(f"skipping not required {req}")
             continue  # skip requests that aren't going to run
-        if r.target.key in liveDependencies:
-            if root:
-                if previous is root or parent is root:
-                    # only yield root once and
-                    # don't consider requests in the same root
-                    continue
-                previous = root
-            yield root or r
+        dependencies = live_dependencies.get(r.target.key)
+        if dependencies:
+            logger.trace(
+                "%s dependencies on future task %s: %s",
+                req.target.key,
+                r.target.key,
+                dependencies,
+            )
+            req.add_dependencies(dependencies)
 
 
-def set_fulfilled(requests, completed):
-    # requests, completed are top level requests,
-    # as is future_dependencies
+def set_fulfilled(
+    upcoming: List[PlanRequest], completed: List[PlanRequest]
+) -> Tuple[List[PlanRequest], List[PlanRequest]]:
+    # find the requests that are ready to run
+    # requests, completed are top level requests
+
+    # Note:
+    # to avoid circular dependencies we want to be eager and run as soon as a task changed required attribute to a valid state
+    # but then another task could change the attribute later in the job.
+    # to fix, the user can declare an explicit dependency on the resource
+    # so that the task won't run after all the resource operations have run (i.e. its task group finished)
+
     ready, notReady = [], []
-    for req in requests:
+    for req in upcoming:
+        if req.completed:
+            continue
+        assert req not in completed, f"{req} already completed"
         _not_ready = False
-        # always need to evaluate both:
-        if req.update_future_dependencies(completed):
-            _not_ready = True
-        if req.update_unfulfilled_errors():
+        previous = req.previous
+        if previous and previous.not_ready:
+            logger.trace(
+                f"Previous {previous} not ready: {previous.get_notready_message()}"
+            )
             _not_ready = True
+        # see if these changes fulfilled the dependencies on a pending task
+        elif req.update_unfulfilled_errors():
+            _not_ready = True  # still has unfulfilled dependencies
         if _not_ready:
             notReady.append(req)
         else:  # list is now empty so request is ready
@@ -504,8 +717,40 @@ def set_fulfilled(requests, completed):
     return ready, notReady
 
 
-def _prepare_request(job, req: TaskRequest, errors):
-    # req is a taskrequests, future_requests are (grouprequest, taskrequest) pairs
+def set_fulfilled_stragglers(
+    upcoming: List[PlanRequest], deploying: bool
+) -> Tuple[List[PlanRequest], List[PlanRequest]]:
+    # we ran all the tasks we could so now we can run left-over tasks
+    # that depend on live attributes that we no longer have to worry about changing
+    ready, notReady = [], []
+    for req in upcoming:
+        if req.completed:
+            continue
+        ok = True
+        previous = req.previous
+        if previous and previous.not_ready:
+            # logger.trace(
+            #     f"Previous {previous} not ready in final round: {previous.get_notready_message()}."
+            # )
+            ok = False
+        elif deploying:
+            for dep in req.get_unfulfilled_refs("operational"):
+                ok = False  # not ready yet
+                # logger.trace(
+                #     f"Dependency for {req} not ready in final round: {req.get_notready_message()}."
+                # )
+                break
+        if ok:
+            ready.append(req)
+            continue
+        # found one, so not ready
+        notReady.append(req)
+    return ready, notReady
+
+
+def _prepare_request(job: "Job", req: PlanRequest, errors: List) -> bool:
+    if not isinstance(req, TaskRequest):
+        return True
     if req.task:
         task = req.task
         task._attributeManager.attributes = {}
@@ -514,17 +759,20 @@ def _prepare_request(job, req: TaskRequest, errors):
         task = req.task = job.create_task(req.configSpec, req.target, reason=req.reason)
     error = None
     try:
+        task.set_envvars()
         proceed, msg = job.should_run_task(task)
         if not proceed:
             req.required = False
+            reassigned = req.reassign_final_for_workflow()
             if task._errors:
                 error = task._errors[0]
             logger.debug(
-                "skipping task %s for instance %s with state %s and status %s: %s",
+                "skipping task %s for instance %s with state %s and status %s (reassigned: %s): %s",
                 req.configSpec.operation,
                 req.target.name,
                 req.target.state,
                 req.target.status,
+                reassigned,
                 msg,
             )
             # we want to track and save skipped tasks
@@ -533,8 +781,10 @@ def _prepare_request(job, req: TaskRequest, errors):
         proceed = False
         # note: failed rendering may be re-tried later if it has dependencies
         error = UnfurlTaskError(task, "should_run_task failed", logging.DEBUG)
+    finally:
+        task.restore_envvars()
     if error:
-        task._inputs = None
+        task._reset()
         task._attributeManager.attributes = {}  # rollback changes
         errors.append(error)
     else:
@@ -542,7 +792,12 @@ def _prepare_request(job, req: TaskRequest, errors):
     return proceed
 
 
-def _render_request(job, parent, req, requests):
+def _render_request(
+    job: "Job",
+    req: PlanRequest,
+    requests: Sequence[PlanRequest],
+    check_target: str,
+) -> Tuple[Optional[bool], Optional[UnfurlError]]:
     # req is a taskrequests, future_requests are (grouprequest, taskrequest) pairs
     assert req.task
     task = req.task
@@ -555,12 +810,15 @@ def _render_request(job, parent, req, requests):
         task.logger.debug("rendering %s %s", task.target.name, task.name)
         task._rendering = True
         task.inputs
-        assert not task._inputs.context.strict
+        task.set_envvars()
+        assert task._inputs is not None and not task._inputs.context.strict
         task.rendered = task.configurator.render(task)
     except Exception:
         # note: failed rendering may be re-tried later if it has dependencies
         error_info = sys.exc_info()
         error = UnfurlTaskError(task, "Configurator render failed", False)
+    finally:
+        task.restore_envvars()
     if task._errors:
         # we turned off strictness so templating errors got saved here instead
         req.render_errors = task._errors
@@ -569,35 +827,34 @@ def _render_request(job, parent, req, requests):
     task._rendering = False
     task._attributeManager.mark_referenced_templates(task.target.template)
 
-    if parent and parent.workflow == "undeploy":
+    workflow = req.group and req.group.workflow
+    if workflow != "undeploy":
         # when removing an instance don't worry about depending values changing in the future
-        deps = []
-    else:
-        # key => (instance, list<attribute>)
+        # key => (instance, list<attribute name>)
         liveDependencies = task._attributeManager.find_live_dependencies()
-        task.logger.trace(f"liveDependencies for {task.target}: {liveDependencies}")
+        task.logger.trace(f"live dependencies for {task.target}: {liveDependencies}")
         # a future request may change the value of these attributes
-        deps = list(_get_deps(parent, req, liveDependencies, requests))
-
-    dependent_refs = [dep.expr for dep in req.get_unfulfilled_refs()]
-    if deps or dependent_refs:
-        req.future_dependencies = deps
+        _get_deps(req, liveDependencies, requests)
+        dependent_refs = [dep.name for dep in req.get_unfulfilled_refs(check_target)]
+    else:
+        dependent_refs = req.render_errors  # type: ignore
+    if dependent_refs:
         task.logger.debug(
             "%s:%s can not render yet, depends on %s",
             task.target.name,
-            req.configSpec.operation,
-            deps or dependent_refs,
+            req.name,
+            dependent_refs,
             exc_info=error_info,
         )
         # rollback changes:
         task._errors = []
-        task._inputs = None
+        task._reset()
         task._attributeManager.attributes = {}
         task.discard_work_folders()
-        return deps or dependent_refs, None
+        return bool(dependent_refs), None
     elif error:
         task.fail_work_folders()
-        task._inputs = None
+        task._reset()
         task.logger.warning("Configurator render failed", exc_info=error_info)
         task._attributeManager.attributes = {}  # rollback changes
         return None, error
@@ -607,65 +864,81 @@ def _render_request(job, parent, req, requests):
         return None, None
 
 
-def _add_to_req_list(reqs, parent, request):
-    if parent:  # only add if we haven't already
-        if not reqs or reqs[-1] is not parent:
-            reqs.append(parent)
-    else:
+def _add_to_req_list(reqs, request):
+    if request not in reqs:
         reqs.append(request)
 
 
 def _reevaluate_not_required(not_required, render_requests):
     # keep rendering if a not_required template was referenced and is now required
     new_not_required = []
-    for (parent, request) in not_required:
+    for request in not_required:
         if request.include_in_plan():
             request.target.validate()
-            render_requests.append((parent, request))
+            render_requests.append(request)
         else:
-            new_not_required.append((parent, request))
+            new_not_required.append(request)
     return new_not_required
 
 
 def do_render_requests(
-    job, requests: Sequence[PlanRequest]
+    job,
+    requests: Sequence[PlanRequest],
+    stragglers=False,
 ) -> Tuple[List[PlanRequest], List[PlanRequest], List[UnfurlError]]:
     ready: List[PlanRequest] = []
     notReady: List[PlanRequest] = []
     errors: List[UnfurlError] = []
     flattened_requests = list(
-        (p, r)
-        for (p, r) in get_render_requests(requests)
-        if _prepare_request(job, r, errors)
+        r for r in get_render_requests(requests) if _prepare_request(job, r, errors)
     )
     not_required = []
     render_requests = collections.deque(flattened_requests)
+    if stragglers:
+        if job.workflow == "deploy":
+            check_target = "operational"
+        else:
+            check_target = "any"
+    else:
+        check_target = ""
+
+    notready_group = None
     while render_requests:
-        parent, request = render_requests.popleft()
+        request = render_requests.popleft()
+        if request.completed:
+            continue
         # we dont require default templates that aren't referenced
         # (but skip this check if the job already specified specific instances)
         required = (
             job.workflow != "deploy" or job.is_filtered() or request.include_in_plan()
         )
         if not required:
-            not_required.append((parent, request))
+            logger.trace("request isn't required by the plan, skipping: %s", request)
+            not_required.append(request)
         else:
             request.target.validate()
-            deps, error = _render_request(job, parent, request, flattened_requests)
+            if notready_group and notready_group == request.group:
+                # group children run sequentially
+                _add_to_req_list(notReady, request)
+                continue
+            if not request.task:
+                # i.e (for now): a setstate request
+                _add_to_req_list(ready, request)
+                continue
+            deps, error = _render_request(
+                job, request, flattened_requests, check_target
+            )
             if error:
                 errors.append(error)
             elif deps:
-                # remove parent from ready if added it there
-                if parent and ready and ready[-1] is parent:
-                    ready.pop()
-                _add_to_req_list(notReady, parent, request)
-            elif not parent or not notReady or notReady[-1] is not parent:
-                # don't add if the parent was placed on the notReady list
-                _add_to_req_list(ready, parent, request)
+                notready_group = request.group
+                _add_to_req_list(notReady, request)
+            else:
+                _add_to_req_list(ready, request)
             not_required = _reevaluate_not_required(not_required, render_requests)
 
-    for (parent, request) in not_required:
-        _add_to_req_list(notReady, parent, request)
+    for request in not_required:
+        _add_to_req_list(notReady, request)
     return ready, notReady, errors
 
 
@@ -703,8 +976,6 @@ def _find_implementation(interface, operation, template):
         if iDef.interfacename == interface or iDef.type == interface:
             if iDef.name == operation:
                 return iDef
-            if iDef.name == "default":
-                default = iDef
     return default
 
 
@@ -756,7 +1027,7 @@ def create_instance_from_spec(_manifest, target, rname, resourceSpec):
         resourceSpec["created"] = target.key
 
     if "parent" not in resourceSpec and "template" in resourceSpec:
-        tname = resourceSpec['template']
+        tname = resourceSpec["template"]
         nodeSpec = _manifest.tosca.get_template(tname)
         if not nodeSpec:
             raise UnfurlError(
@@ -783,12 +1054,22 @@ def _maybe_mock(iDef, template):
     return iDef
 
 
+def get_success_status(workflow):
+    if workflow == "deploy":
+        return Status.ok
+    elif workflow == "stop":
+        return Status.pending
+    elif workflow == "undeploy":
+        return Status.absent
+    return None
+
+
 def create_task_request(
-    jobOptions,
-    operation,
-    resource,
-    reason=None,
-    inputs=None,
+    jobOptions: "JobOptions",
+    operation: str,
+    resource: "EntityInstance",
+    reason: Optional[str] = None,
+    inputs: Optional[Mapping] = None,
     startState=None,
     operation_host=None,
     skip_filter=False,
@@ -844,7 +1125,7 @@ def create_task_request(
         configSpec,
         resource,
         reason or action,
-        startState=startState,
+        startState=startState or iDef.entry_state,
     )
     if skip_filter:
         return req
