@@ -1,17 +1,24 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
+import abc
 import os
 import os.path
 from pathlib import Path
 import sys
+from functools import lru_cache
+from typing import TYPE_CHECKING, Dict, List, Optional, Union, cast
+from typing_extensions import Literal
 import git
-from .logs import getLogger
+import git.exc
+
+from .logs import getLogger, PY_COLORS
 from urllib.parse import urlparse
 from .util import UnfurlError, save_to_file
-import toscaparser.repositories
+from toscaparser.repositories import Repository
 from ruamel.yaml.comments import CommentedMap
 import logging
-
+if TYPE_CHECKING:
+    from .packages import Package
 
 logger = getLogger("unfurl")
 
@@ -21,7 +28,7 @@ def is_git_worktree(path, gitDir=".git"):
     return os.path.exists(os.path.join(path, gitDir))
 
 
-def _add_user_to_url(url, username, password):
+def add_user_to_url(url, username, password):
     assert username
     parts = urlparse(url)
     user, sep, host = parts.netloc.rpartition("@")
@@ -29,10 +36,11 @@ def _add_user_to_url(url, username, password):
         netloc = f"{username}:{password}@{host}"
     else:
         netloc = f"{username}@{host}"
+
     return parts._replace(netloc=netloc).geturl()
 
 
-def normalize_git_url(url, hard=0):
+def normalize_git_url(url: str, hard: int = 0):
     if url.startswith("git-local://"):  # truncate url after commit digest
         return "git-local://" + urlparse(url).netloc.partition(":")[0]
 
@@ -41,29 +49,44 @@ def normalize_git_url(url, hard=0):
             return "file://" + url
         elif "@" in url:  # scp style used by git: user@server:project.git
             # convert to ssh://user@server/project.git
-            return "ssh://" + url.replace(":", "/", 1)
+            url = "ssh://" + url.replace(":", "/", 1)
     if hard:
         parts = urlparse(url)
         # remove password and .git
         user, sep, host = parts.netloc.rpartition("@")
         if sep and hard == 1:
             netloc = f"{user.partition(':')[0]}@{host}"
-        else:
+        else:  # hard >= 2
             netloc = host
-        if hard == 2:
-            path = parts.path.rstrip(".git")
+        if hard == 3 and parts.path.endswith(".git"):
+            path = parts.path[:-4]
         else:
             path = parts.path
         return parts._replace(netloc=netloc, path=path).geturl()
     return url
 
 
+def sanitize_url(url: str) -> str:
+    if "://" in url and "@" in url:  # sanitize
+        parts = urlparse(url)
+        # XXXX out user and password
+        user, sep, host = parts.netloc.rpartition("@")
+        if user:
+            user, sep, password = user.partition(':')
+            netloc = f"XXXXX{':XXXXX' if password else ''}@{host}"
+            return parts._replace(netloc=netloc).geturl()
+    return url
+
+
 def normalize_git_url_hard(url):
     # remove scheme, .git and fragment
-    return normalize_git_url(url, hard=2).rpartition("://")[2].partition("#")[0]
+    return normalize_git_url(url, hard=3).rpartition("://")[2].partition("#")[0]
 
 
 def is_url_or_git_path(url):
+    if url.startswith("--"):
+        # security: see https://github.com/gitpython-developers/GitPython/issues/1517
+        return False
     if "://" in url and not url.startswith("file:"):
         return True
     if "@" in url:
@@ -79,6 +102,9 @@ def split_git_url(url):
     Returns (repoURL, filePath, revision)
     RepoURL will be an empty string if it isn't a path to a git repo
     """
+    if url.startswith("--"):
+        # security: see https://github.com/gitpython-developers/GitPython/issues/1517
+        return False
     parts = urlparse(url)
     if parts.scheme == "git-local":
         return parts.scheme + "://" + parts.netloc, parts.path[1:], parts.fragment
@@ -94,14 +120,34 @@ def split_git_url(url):
     return url, "", ""
 
 
+# git fetch <remote> 'refs/tags/*:refs/tags/*' if our clones are shallow 
+@lru_cache(None)  # XXX won't get up dates in server mode
+def get_remote_tags(url, pattern="*") -> List[str]:
+    # https://github.com/gitpython-developers/GitPython/issues/1071
+    # https://myshittycode.com/2020/10/02/git-querying-tags-without-cloning-the-repository/
+    # -v:refname is version sort in reverse order
+    # -c versionsort.suffix=- ensures 1.0.0-XXXXXX comes before 1.0.0.
+    blob = git.cmd.Git()(c="versionsort.suffix=-").ls_remote(url, pattern, sort='-v:refname', tags=True)
+    # len("b90df3d12413db22d051db1f7c7286cdd2f00b66\trefs/tags/") == 51
+    # filter out ^{} references (see https://stackoverflow.com/questions/12938972/what-does-mean-in-git)
+    tags = [line[51:] for line in blob.split('\n') if not line.endswith("^{}")]
+    logger.debug("got %s remote tags with pattern %s from %s", len(tags), pattern, url)
+    return tags
+
+
 class _ProgressPrinter(git.RemoteProgress):
+    gitUrl = ''
+
     def update(self, op_code, cur_count, max_count=None, message=""):
         # we use print instead of logging because we don't want to clutter logs with this message
         if message and logger.getEffectiveLevel() <= logging.INFO:
-            print(f"fetching from {self.gitUrl}, received: {message} ", file=sys.stderr)
+            url = self.gitUrl
+            print(f"fetching from {url}, received: {message} ", file=sys.stderr)
 
 
-class Repo:
+class Repo(abc.ABC):
+    url: str = ""
+
     @staticmethod
     def find_containing_repo(rootDir, gitDir=".git"):
         """
@@ -115,8 +161,8 @@ class Repo:
         return None
 
     @staticmethod
-    def find_git_working_dirs(rootDir, include_root, gitDir=".git"):
-        working_dirs = {}
+    def find_git_working_dirs(rootDir, include_root, gitDir=".git") -> Dict[str, "RepoView"]:
+        working_dirs: Dict[str, "RepoView"] = {}
         for root, dirs, files in os.walk(rootDir):
             if Repo.update_git_working_dirs(working_dirs, root, dirs, gitDir):
                 if not include_root or rootDir != root:
@@ -124,7 +170,7 @@ class Repo:
         return working_dirs
 
     @staticmethod
-    def update_git_working_dirs(working_dirs, root, dirs, gitDir=".git"):
+    def update_git_working_dirs(working_dirs, root, dirs, gitDir=".git") -> Optional[str]:
         if gitDir in dirs and is_git_worktree(root, gitDir):
             assert os.path.isdir(root), root
             repo = GitRepo(git.Repo(root))
@@ -143,6 +189,16 @@ class Repo:
                 return path
         return None
 
+    @property
+    @abc.abstractmethod
+    def working_dir(self) -> str:
+        ...
+
+    @property
+    @abc.abstractmethod
+    def revision(self) -> str:
+        ...
+
     def find_repo_path(self, path):
         localPath = self.find_path(path)[0]
         if localPath is not None and not self.is_path_excluded(localPath):
@@ -170,26 +226,32 @@ class Repo:
         return None, None, None
 
     def as_repo_view(self, name=""):
-        return RepoView(dict(name=name, url=self.url), self)
+        return RepoView(dict(name=name, url=self.url), cast(GitRepo, self))
 
     def is_local_only(self):
         return self.url.startswith("git-local://") or os.path.isabs(self.url)
 
     @staticmethod
-    def get_path_for_git_repo(gitUrl):
+    def get_path_for_git_repo(gitUrl: str, name_only=True) -> str:
         parts = urlparse(gitUrl)
         if parts.scheme == "git-local":
             # e.g. extract spec from git-local://0cfeee6571c4276ce1a63dc37aa8cbf8b8085d60:spec
             name = parts.netloc.partition(":")[1]
         else:
+            path = parts.path.strip("/")
             # e.g. extract tosca-parser from https://github.com/onecommons/tosca-parser.git
+            if name_only:
+                path = os.path.basename(path)
             name = (
-                os.path.splitext(os.path.basename(parts.path.strip("/")))[0]
+                os.path.splitext(path)[0]
                 or parts.netloc
             )
         assert not name.endswith(".git"), name
         return name
 
+    def project_path(self) -> str:
+        return self.get_path_for_git_repo(self.url, False)
+
     @classmethod
     def create_working_dir(cls, gitUrl, localRepoPath, revision=None):
         localRepoPath = localRepoPath or "."
@@ -198,15 +260,23 @@ class Repo:
                 raise UnfurlError(
                     f"couldn't create directory, it already exists and isn't empty: {localRepoPath}"
                 )
-        logger.info("Fetching %s %s to %s", gitUrl, revision or "", localRepoPath)
-        progress = _ProgressPrinter()
-        progress.gitUrl = gitUrl
+        parent_dir = os.path.dirname(localRepoPath)
+        if parent_dir.strip("/"):
+            os.makedirs(parent_dir, exist_ok=True)
+        cleanurl = sanitize_url(gitUrl)
+        logger.info("Fetching %s %s to %s", cleanurl, revision or "", localRepoPath)
+        kwargs = dict(recurse_submodules=True, depth=1, no_single_branch=True)
+        non_interactive = os.getenv("CI") or not PY_COLORS  # or color out disabled
+        if not non_interactive:
+            # we're running in an interactive session
+            progress = _ProgressPrinter()
+            progress.gitUrl = cleanurl
+            kwargs["progress"] = progress  # type: ignore
         try:
-            kwargs = dict(recurse_submodules=True)
             if revision:
                 kwargs["branch"] = revision
-            repo = git.Repo.clone_from(gitUrl, localRepoPath, progress, **kwargs)
-        except git.exc.GitCommandError as err:
+            repo = git.Repo.clone_from(gitUrl, localRepoPath, **kwargs)  # type: ignore
+        except git.exc.GitCommandError as err:  # type: ignore
             raise UnfurlError(
                 f'couldn\'t create working directory, clone failed: "{err._cmdline}"\nTry re-running that command to diagnose the problem.'
             )
@@ -249,24 +319,29 @@ def find_dirty_secrets(working_dir):
 class RepoView:
     # view of Repo optionally filtered by path
     # XXX and revision too
-    def __init__(self, repository, repo, path=""):
+    def __init__(self, repository: Union[dict, Repository], repo: Optional["GitRepo"], path="") -> None:
         if isinstance(repository, dict):
             # required keys: name, url
             tpl = repository.copy()
             name = tpl.pop("name")
             tpl["url"] = normalize_git_url(tpl["url"])
-            repository = toscaparser.repositories.Repository(name, tpl)
+            repository = Repository(name, tpl)
         assert repository or repo
-        self.repository = repository
+        self.repository: Repository = repository
+        self.yaml = None
+        self.revision: Optional[str] = None
+        self.set_repo_and_path(repo, path)
+        self.package: Optional[Union[Literal[False], "Package"]] = None
+
+    def set_repo_and_path(self, repo: Optional["GitRepo"], path: str):
         self.repo = repo
         self.path = path
-        if repo and path and repository:
-            self.repository.url = repo.get_url_with_path(path)
+        if repo and path and self.repository:
+            self.repository.url = repo.get_url_with_path(path, False, self.revision or "")
         self.readOnly = not repo
-        self.yaml = None
 
     @property
-    def working_dir(self):
+    def working_dir(self) -> str:
         if self.repo:
             return os.path.join(self.repo.working_dir, self.path)
         else:
@@ -282,7 +357,7 @@ class RepoView:
             url = self.repository.url
             if self.repository.credential:
                 credential = self.repository.credential
-                return _add_user_to_url(url, credential["user"], credential["token"])
+                return add_user_to_url(url, credential["user"], credential["token"])
             else:
                 return url
         else:
@@ -304,14 +379,14 @@ class RepoView:
         return ""
 
     def is_dirty(self):
-        if self.readOnly:
+        if self.readOnly or not self.repo:
             return False
         for filepath, dotsecrets in find_dirty_secrets(self.working_dir):
             return True
         return self.repo.is_dirty(untracked_files=True, path=self.path)
 
     def add_all(self):
-        assert not self.readOnly
+        assert not self.readOnly and self.repo
         self.repo.repo.git.add("--all", self.path or ".")
 
     def load_secrets(self, _loader):
@@ -344,26 +419,26 @@ class RepoView:
     def save_secrets(self):
         return commit_secrets(self.working_dir, self.yaml)
 
-    def commit(self, message, addAll=False):
+    def commit(self, msg: str, add_all: bool = False) -> int:
         assert not self.readOnly
+        assert self.repo
         if self.yaml:
             for saved in self.save_secrets():
                 local_path = str(saved.relative_to(self.repo.working_dir))
                 if not self.repo.is_path_excluded(local_path):
                     self.repo.repo.git.add(local_path)
-        if addAll:
+        if add_all:
             self.add_all()
-        self.repo.repo.index.commit(message)
+        self.repo.repo.index.commit(msg)
         return 1
 
-    def get_default_commit_message(self):
-        return "Commit by Unfurl"
-
     def git_status(self):
         assert not self.readOnly
+        assert self.repo
         return self.repo.run_cmd(["status", self.path or "."])[1]
 
     def _secrets_status(self):
+        assert self.repo
         modified = "\n   ".join(
             [
                 str(filepath.relative_to(self.repo.working_dir))
@@ -410,26 +485,43 @@ class RepoView:
         if self.name:
             record["name"] = self.name
         if self.origin:
-            record["origin"] = self.origin
+            record["origin"] = normalize_git_url(self.origin, 1)
         return record
 
 
 class GitRepo(Repo):
-    def __init__(self, gitrepo):
+    def __init__(self, gitrepo: git.Repo):
         self.repo = gitrepo
-        self.url = self.working_dir or gitrepo.git_dir
+        self.url = self.working_dir or str(gitrepo.git_dir)
         if gitrepo.remotes:
             # note: these might not look like absolute urls, e.g. git@github.com:onecommons/unfurl.git
             try:
                 remote = gitrepo.remotes["origin"]
-            except:
+            except Exception:
                 remote = gitrepo.remotes[0]
             self.url = remote.url
 
+    def add_transient_credentials(self, username, password):
+        assert "@" not in self.url, self.url
+        replacement = f'url."{username}:{password}@{self.url}".insteadOf "{self.url}"'
+        self.repo.git.set_persistent_git_options(c=replacement)
+
+    def set_url_credentials(self, username, password):
+        try:
+            remote = self.repo.remotes["origin"]
+        except Exception:
+            remote = self.repo.remotes[0]
+        new_url = add_user_to_url(remote.url, username, password)
+        # replace the url
+        remote.set_url(new_url, remote.url)
+
     @property
-    def working_dir(self):
+    def working_dir(self) -> str:
         dir = self.repo.working_tree_dir
-        if not dir or dir[-1] == "/":
+        if not dir:
+            return ""
+        dir = str(dir)
+        if dir[-1] == "/":
             return dir
         else:
             return dir + "/"
@@ -443,17 +535,23 @@ class GitRepo(Repo):
     def resolve_rev_spec(self, revision):
         try:
             return self.repo.commit(revision).hexsha
-        except:
+        except Exception:
             return None
 
-    def get_url_with_path(self, path, sanitize=False):
+    def get_url_with_path(self, path: str, sanitize: bool = False, revision: str = ""):
+        hard = 2 if sanitize else 0
         if is_url_or_git_path(self.url):
             if os.path.isabs(path):
                 # get path relative to repository's root
                 path = os.path.relpath(path, self.working_dir)
-            return normalize_git_url(self.url, sanitize) + "#:" + path
+                if path.startswith(".."):
+                    # outside of the repo, don't include it in the url
+                    if revision:
+                        revision = "#" + revision
+                    return normalize_git_url(self.url, hard) + revision
+            return normalize_git_url(self.url, hard) + "#" + revision + ":" + path
         else:
-            return self.get_git_local_url(path)
+            return self.get_git_local_url(path, revision=revision)
 
     def find_excluded_dirs(self, root):
         root = os.path.relpath(root, self.working_dir)
@@ -478,21 +576,24 @@ class GitRepo(Repo):
         # success error code means it's ignored
         return not self.run_cmd(["check-ignore", "-q", localPath])[0]
 
-    def run_cmd(self, args, **kw):
+    def reset(self, args: str = "--hard HEAD~1") -> bool:
+        return not self.run_cmd(("reset " + args).split())[0]
+
+    def run_cmd(self, args, with_exceptions=False, **kw):
         """
         :return:
           tuple(int(status), str(stdout), str(stderr))
         """
         gitcmd = self.repo.git
-        call = [gitcmd.GIT_PYTHON_GIT_EXECUTABLE]
+        call: List[str] = [gitcmd.GIT_PYTHON_GIT_EXECUTABLE or "git"]
         # add persistent git options
         call.extend(gitcmd._persistent_git_options)
         call.extend(list(args))
 
         # note: sets cwd to working_dir
-        return gitcmd.execute(
-            call, with_exceptions=False, with_extended_output=True, **kw
-        )
+        return gitcmd.execute(  # type: ignore
+            call, with_exceptions=with_exceptions, with_extended_output=True, **kw
+        )  
 
     def add_to_local_git_ignore(self, rule):
         path = os.path.join(self.repo.git_dir, "info")
@@ -553,12 +654,12 @@ class GitRepo(Repo):
         # https://gitpython.readthedocs.io/en/stable/reference.html?highlight=is_dirty#git.repo.base.Repo.is_dirty
         return self.repo.is_dirty(untracked_files=untracked_files, path=path or None)
 
-    def pull(self, remote="origin", revision=None, ff_only=True):
+    def pull(self, remote="origin", revision=None, ff_only=True, **kw):
         if remote in self.repo.remotes:
             cmd = ["pull", remote, revision or "HEAD"]
             if ff_only:
                 cmd.append("--ff-only")
-            code, out, err = self.run_cmd(cmd)
+            code, out, err = self.run_cmd(cmd, **kw)
             if code:
                 logger.info(
                     "attempt to pull latest from %s into %s failed: %s %s",
@@ -587,11 +688,18 @@ class GitRepo(Repo):
         Repo.ignore_dir(newPath)
         return GitRepo(cloned)
 
-    def get_git_local_url(self, path, name=""):
+    def get_git_local_url(self, path, name="", revision=""):
         if os.path.isabs(path):
             # get path relative to repository's root
             path = os.path.relpath(path, self.working_dir)
-        return f"git-local://{self.get_initial_revision()}:{name}/{path}"
+        if revision:
+            revision = "#" + revision
+        return f"git-local://{self.get_initial_revision()}:{name}/{path}{revision}"
+
+    def delete_dir(self, path, commit=None):
+        self.repo.index.remove(os.path.abspath(path), r=True, working_tree=True)
+        if commit:
+            self.repo.index.commit(commit)
 
     # XXX: def getDependentRepos()
     # XXX: def canManage()
@@ -634,22 +742,22 @@ class GitRepo(Repo):
     #     repo.git.commit("")
 
 
-class RevisionManager:
-    def __init__(self, manifest, localEnv=None):
-        self.manifest = manifest
-        self.revisions = None
-        self.localEnv = localEnv
+# class RevisionManager:
+#     def __init__(self, manifest, localEnv=None):
+#         self.manifest = manifest
+#         self.revisions = None
+#         self.localEnv = localEnv
 
-    def get_revision(self, change):
-        if self.revisions is None:
-            self.revisions = {self.manifest.specDigest: self.manifest}
-        digest = change["specDigest"]
-        commitid = change["startCommit"]
-        if digest in self.revisions:
-            return self.revisions[digest]
-        else:
-            from .manifest import SnapShotManifest
+#     def get_revision(self, change):
+#         if self.revisions is None:
+#             self.revisions = {self.manifest.specDigest: self.manifest}
+#         digest = change["specDigest"]
+#         commitid = change["startCommit"]
+#         if digest in self.revisions:
+#             return self.revisions[digest]
+#         else:
+#             from .manifest import SnapShotManifest
 
-            manifest = SnapShotManifest(self.manifest, commitid)
-            self.revisions[digest] = manifest
-            return manifest
+#             manifest = SnapShotManifest(self.manifest, commitid)
+#             self.revisions[digest] = manifest
+#             return manifest
