@@ -8,6 +8,7 @@ Each instance have a status; attributes that describe its state; and a TOSCA tem
 which describes its capabilities, relationships and available interfaces for configuring and interacting with it.
 """
 from datetime import datetime
+import os
 from typing import (
     Any,
     Dict,
@@ -17,6 +18,8 @@ from typing import (
     Union,
     TYPE_CHECKING,
     cast,
+    NewType,
+    AnyStr
 )
 import six
 from collections.abc import Mapping
@@ -35,26 +38,24 @@ from .tosca import (
     TopologySpec,
     ArtifactSpec,
 )
-from .logs import UnfurlLogger
-import logging
+from .logs import getLogger
 
 if TYPE_CHECKING:
     from unfurl.configurator import Dependency
     import toscaparser.repositories
 
-# Tell mypy the logger is of type UnfurlLogger
-logger = cast(UnfurlLogger, logging.getLogger("unfurl"))
+logger = getLogger("unfurl")
 
-# CF 3.4.1 Node States p61
 
+InstanceKey = NewType("InstanceKey", str)
 
 class Operational(ChangeAware):
     """
     This is an abstract base class for Jobs, Resources, and Configurations all have a Status associated with them
     and all use the same algorithm to compute their status from their dependent resouces, tasks, and configurations
     """
-
     # XXX3 add repairable, messages?
+    name = ""
 
     # core properties to override
     @property
@@ -69,6 +70,20 @@ class Operational(ChangeAware):
     def state(self) -> NodeState:
         return NodeState.initial
 
+    def has_state(self, state: NodeState) -> bool:
+        if state == NodeState.initial:
+            return True
+        if self.state is None:
+            # self.state is sometimes None even though it shouldn't be
+            return False
+        if state == NodeState.error:
+            return self.state == state
+        if state < NodeState.stopping:
+            return self.state < NodeState.stopping and self.state >= state
+        if self.state > NodeState.stopping:
+            return self.state >= state
+        return False
+
     def get_operational_dependencies(self) -> Iterable["Operational"]:
         """
         Return an iterator of `Operational` object that this instance depends on to be operational.
@@ -119,13 +134,14 @@ class Operational(ChangeAware):
         if not status:
             return Status.unknown
 
-        if status >= Status.error:
-            # return error, pending, or absent
+        if status == Status.error or status == Status.absent:
+            # return error, or absent
             return status
 
         dependentStatus = self.aggregate_status(
             self.get_operational_dependencies(), seen
         )
+
         if self.is_computed():
             # if local_status is a no op (status purely determined by dependents) set to ok
             status = Status.ok
@@ -151,15 +167,15 @@ class Operational(ChangeAware):
     # The computed attribute `lastChange` is whichever of the above two attributes are more recent.
 
     @property
-    def last_state_change(self) -> Optional[datetime]:
+    def last_state_change(self) -> Optional[str]:
         return None
 
     @property
-    def last_config_change(self) -> Optional[datetime]:
+    def last_config_change(self) -> Optional[str]:
         return None
 
     @property
-    def last_change(self) -> Optional[datetime]:
+    def last_change(self) -> Optional[str]:
         if not self.last_state_change:
             return self.last_config_change
         elif not self.last_config_change:
@@ -179,7 +195,7 @@ class Operational(ChangeAware):
         return False
 
     @staticmethod
-    def aggregate_status(statuses: Iterable["Operational"], seen) -> Optional[Status]:
+    def aggregate_status(statuses: Iterable["Operational"], seen: Dict[int, "Operational"]) -> Optional[Status]:
         """
         Returns: ok, degraded, pending or None
 
@@ -235,16 +251,16 @@ class OperationalInstance(Operational):
         status: Optional[Union["OperationalInstance", int, str]] = None,
         priority: Optional[Union[int, str, Priority]] = None,
         manualOveride: Optional[Union["OperationalInstance", int, str]] = None,
-        lastStateChange: Optional[datetime] = None,
-        lastConfigChange: Optional[datetime] = None,
+        lastStateChange: Optional[str] = None,
+        lastConfigChange: Optional[str] = None,
         state: NodeState = None,
     ) -> None:
         if isinstance(status, OperationalInstance):
             self._localStatus: Optional[Status] = status._localStatus
             self._manualOverideStatus: Optional[Status] = status._manualOverideStatus
             self._priority: Optional[Priority] = status._priority
-            self._lastStateChange: Optional[datetime] = status._lastStateChange
-            self._lastConfigChange: Optional[datetime] = status._lastConfigChange
+            self._lastStateChange: Optional[str] = status._lastStateChange
+            self._lastConfigChange: Optional[str] = status._lastConfigChange
             self._state: Optional[NodeState] = status._state
             self.dependencies: List["Operational"] = status.dependencies
         else:
@@ -304,11 +320,11 @@ class OperationalInstance(Operational):
     priority: Optional[Priority] = property(**__priority())  # type: ignore
 
     @property
-    def last_state_change(self) -> Optional[datetime]:
+    def last_state_change(self) -> Optional[str]:
         return self._lastStateChange
 
     @property
-    def last_config_change(self) -> Optional[datetime]:
+    def last_config_change(self) -> Optional[str]:
         return self._lastConfigChange
 
     def __state():  # type: ignore
@@ -340,14 +356,14 @@ class _ChildResources(Mapping):
 
 
 class EntityInstance(OperationalInstance, ResourceRef):
-    attributeManager = None
+    attributeManager: Optional[AttributeManager] = None
     created: Optional[Union[bool, str]] = None
     protected: Optional[bool] = None
     imports = None
     imported = None
-    envRules = None
     _baseDir = ""
-    templateType = EntitySpec
+    templateType = EntitySpec  # must defined by subtype
+    parentRelation = ""
 
     def __init__(
         self, name="", attributes=None, parent=None, template=None, status=Status.ok
@@ -358,11 +374,11 @@ class EntityInstance(OperationalInstance, ResourceRef):
         self.name = name
         self._attributes = attributes or {}
         self.parent = parent
-        if parent:
+        if parent and self.parentRelation:
             p = getattr(parent, self.parentRelation)
             p.append(self)
 
-        self.template: EntitySpec = template or self.templateType()
+        self.template: EntitySpec = template or self.templateType()  # type: ignore
         assert isinstance(self.template, self.templateType)
         self._properties = {}
 
@@ -382,15 +398,30 @@ class EntityInstance(OperationalInstance, ResourceRef):
             return super()._get_prop(name)
 
     def _resolve(self, key):
-        # might return a Result
-        self.attributes[key]  # force resolve
-        return self.attributes._attributes[key]
+        # force resolve, might return a Result
+        return self.attributes._getresult(key)
 
     def query(self, expr, vars=None, wantList=False, trace=None):
         from .eval import Ref, RefContext
 
         return Ref(expr).resolve(RefContext(self, vars=vars, trace=trace), wantList)
 
+    def __priority():  # type: ignore
+        doc = "The priority property."
+
+        def fget(self) -> Optional[Priority]:
+            if self._priority is None and self.template:
+                return Priority.required if self.template.required else Priority.ignore
+            else:
+                return self._priority
+
+        def fset(self, value):  # type: ignore
+            self._priority = to_enum(Priority, value)
+
+        return locals()
+
+    priority: Optional[Priority] = property(**__priority())  # type: ignore
+
     def __local_status():  # type: ignore
         doc = "The working_dir property."
 
@@ -420,8 +451,9 @@ class EntityInstance(OperationalInstance, ResourceRef):
         return self.template.aggregate_only()
 
     @property
-    def key(self):
-        return f"{self.parent.key}::.{self.parentRelation[1:]}::{self.name}"
+    def key(self) -> InstanceKey:
+        assert self.parent and self.parentRelation
+        return cast(InstanceKey, f"{self.parent.key}::.{self.parentRelation[1:]}::{self.name}")
 
     def as_ref(self, options=None):
         return {"ref": self.key}
@@ -545,10 +577,11 @@ class HasInstancesInstance(EntityInstance):
         if self.root is self:
             self._all = _ChildResources(self)
             self._templar = Templar(DataLoader())
+            self._environ = os.environ.copy()
 
     @property
-    def key(self):
-        return f"::{self.name}"
+    def key(self) -> InstanceKey:
+        return cast(InstanceKey, f"::{self.name}")
 
     def get_self_and_descendents(self):
         "Recursive descendent including self"
@@ -581,10 +614,14 @@ class HasInstancesInstance(EntityInstance):
             return self.imports.find_import(resourceid)
         return None
 
-    def get_requirements(self, match):
+    @property
+    def requirements(self) -> List["RelationshipInstance"]:
+        return []
+
+    def get_requirements(self, match) -> List["RelationshipInstance"]:
         if match is None:
             return self.requirements
-        if isinstance(match, six.string_types):
+        if isinstance(match, str):
             return [r for r in self.requirements if r.template.name == match]
         elif isinstance(match, NodeInstance):
             return [r for r in self.requirements if r.target == match]
@@ -628,17 +665,14 @@ class CapabilityInstance(EntityInstance):
                     )
                     assert rel in self._relationships
                     # find the node instance that uses this requirement
-                    if template.source:
-                        sourceNode = self.root.find_resource(template.source.name)
-                        if sourceNode:
-                            rel.source = sourceNode
 
         return self._relationships
 
     @property
-    def key(self):
+    def key(self) -> InstanceKey:
         # XXX implement something like _ChildResources to enable ::name instead of [.name]
-        return f"{self.parent.key}::.{'capabilities'}::[.name={self.name}]"
+        assert self.parent
+        return cast(InstanceKey, f"{self.parent.key}::.{'capabilities'}::[.name={self.name}]")
 
 
 class RelationshipInstance(EntityInstance):
@@ -648,7 +682,12 @@ class RelationshipInstance(EntityInstance):
     templateType = RelationshipSpec
     if TYPE_CHECKING:
         template = templateType()
-    source = None
+    _source: Optional["NodeInstance"] = None
+
+    def __init__(
+        self, name="", attributes=None, parent=None, template=None, status=Status.pending
+    ):
+        EntityInstance.__init__(self, name, attributes, parent, template, status)
 
     @property
     def target(self) -> Optional["NodeInstance"]:
@@ -658,19 +697,28 @@ class RelationshipInstance(EntityInstance):
         else:
             return None
 
+    @property
+    def source(self) -> Optional["NodeInstance"]:
+        if self._source is None:
+            if self.template.source:
+                sourceNode = self.root.find_instance(self.template.source.name)
+                if sourceNode:
+                    self._source = sourceNode
+        return self._source
+
     @property
     def capability(self) -> Optional[CapabilityInstance]:
         return self.parent
 
     @property
-    def key(self) -> str:
+    def key(self) -> InstanceKey:
         # XXX implement something like _ChildResources to enable ::name instead of [.name=name]
         if self.source:
-            return f"{self.source.key}::.requirements::[.name={self.name}]"
+            return cast(InstanceKey, f"{self.source.key}::.requirements::[.name={self.name}]")
         elif not self.parent or self.parent is self.root:  # it's a default relationship
-            return f"::.requirements::[.name={self.name}]"
+            return cast(InstanceKey, f"::.requirements::[.name={self.name}]")
         elif self.parent:  # capability is parent
-            return f"{self.parent.key}::.relationships::[.name={self.name}]"
+            return cast(InstanceKey, f"{self.parent.key}::.relationships::[.name={self.name}]")
 
     def merge_props(self, matchfn, delete_if_none=False) -> Dict[str, str]:
         env: Dict[str, str] = {}
@@ -692,6 +740,9 @@ class RelationshipInstance(EntityInstance):
                 env[name] = val
         return env
 
+    def __repr__(self):
+        return f"RelationshipInstance('{self.name}')"
+
 
 class ArtifactInstance(EntityInstance):
     parentRelation = "_artifacts"
@@ -730,6 +781,9 @@ class ArtifactInstance(EntityInstance):
         for d in self.dependencies:
             yield d
 
+    def __repr__(self):
+        return f"ArtifactInstance('{self.name}')"
+
 
 class NodeInstance(HasInstancesInstance):
     templateType = NodeSpec
@@ -765,9 +819,9 @@ class NodeInstance(HasInstancesInstance):
         assert relationship and relationship.capability and relationship.target
         # find the Capability instance that corresponds to this relationship template's capability
         targetNodeInstance = self.root.find_resource(relationship.target.name)
-        assert (
-            targetNodeInstance
-        ), f"target instance {relationship.target.name} should have been already created"
+        if not targetNodeInstance:
+            logger.warning(f"target instance {relationship.target.name} should have already been created -- is {self.name} out of sync with latest templates?")
+            return None
         for cap in targetNodeInstance.capabilities:
             if cap.template is relationship.capability:
                 for relInstance in cap.relationships:
@@ -788,9 +842,8 @@ class NodeInstance(HasInstancesInstance):
                 if id(template.relationship) not in instantiated:
                     relInstance = self._find_relationship(template.relationship)
                     if not relInstance:
-                        raise UnfurlError(
-                            f'can not find relation instance for requirement "{name}" on node "{self.name}"'
-                        )
+                        logger.warning(f'can not find relation instance for requirement "{name}" on node "{self.name}"')
+                        continue
                     assert template.relationship is relInstance.template
                     assert self.template is relInstance.template.source
                     self._requirements.append(relInstance)
@@ -844,7 +897,7 @@ class NodeInstance(HasInstancesInstance):
     def repository(self) -> Optional["toscaparser.repositories.Repository"]:
         if self.template.is_compatible_type("unfurl.nodes.Repository"):
             # this is a reified repository -- return the Repository with the same name
-            return self.template.spec.template.repositories.get(self.template.name)
+            return self.template.spec.get_repository(self.template.name)
         return None
 
     def _get_default_relationships(self, relation=None):
@@ -860,15 +913,16 @@ class NodeInstance(HasInstancesInstance):
 
     @property
     def sources(self):
-        dep = {}
+        dep: Dict[str, Union[EntityInstance, List[EntityInstance]]] = {}
         for cap in self.capabilities:
             for rel in cap.relationships:
                 if rel.source:
                     if rel.name in dep:
-                        if isinstance(dep[rel.name], list):
-                            dep[rel.name].append(rel.source)
+                        val = dep[rel.name]
+                        if isinstance(val, list):
+                            val.append(rel.source)
                         else:
-                            dep[rel.name] = [dep[rel.name], rel.source]
+                            dep[rel.name] = [val, rel.source]
                     else:
                         dep[rel.name] = rel.source
         return dep
@@ -914,8 +968,8 @@ class NodeInstance(HasInstancesInstance):
     def _resolve(self, key):
         # might return a Result
         try:
-            self.attributes[key]  # force resolve
-            return self.attributes._attributes[key]
+            # force resolve
+            return self.attributes._getresult(key)
         except KeyError:
             try:
                 inherit = self._interfaces.get("inherit")  # pre-loaded
@@ -934,6 +988,7 @@ class NodeInstance(HasInstancesInstance):
         yield from super().get_operational_dependencies()
 
         for instance in self.requirements:
+            # super() yields self.parent so skip that one
             if instance.target is not self.parent:
                 yield instance
 
@@ -969,7 +1024,7 @@ class NodeInstance(HasInstancesInstance):
         else:
             className = self._attributes[".interfaces"].get(name)
             if className:
-                instance = load_class(className)(name, self)
+                instance = load_class(className)(name, self)  # type: ignore
                 self._interfaces[name] = instance
                 return instance
         return None
