@@ -5,7 +5,6 @@ Internal classes supporting the runtime.
 """
 import collections
 from collections.abc import MutableSequence, Mapping
-import copy
 import os
 import sys
 import os.path
@@ -13,12 +12,18 @@ import six
 import re
 import ast
 import time
-from typing import TYPE_CHECKING, cast, Dict, Optional, Any
+from typing import TYPE_CHECKING, Iterator, List, MutableMapping, Tuple, Union, cast, Dict, Optional, Any, NewType
+from typing_extensions import Protocol
 from enum import Enum
 from urllib.parse import urlsplit
 
 
-from .eval import RefContext, set_eval_func, Ref, map_value
+if TYPE_CHECKING:
+    from .manifest import Manifest
+    from .runtime import EntityInstance, InstanceKey
+    from .configurator import Dependency
+
+from .eval import RefContext, set_eval_func, Ref, map_value, SafeRefContext
 from .result import (
     Results,
     ResultsList,
@@ -142,6 +147,7 @@ def eval_python(arg, ctx):
     else:
         funcName = arg
     func = load_class(funcName)
+    assert callable(func)
 
     kw = ctx.kw
     if "args" in kw:
@@ -162,13 +168,13 @@ set_eval_func("python", eval_python)
 
 set_eval_func(
     "sensitive",
-    lambda arg, ctx: wrap_sensitive_value(map_value(arg, ctx)),
+    lambda arg, ctx: wrap_sensitive_value(map_value(arg, ctx)), safe=True
 )
 
 
 set_eval_func(
     "portspec",
-    lambda arg, ctx: PortSpec.make(map_value(arg, ctx)),
+    lambda arg, ctx: PortSpec.make(map_value(arg, ctx)), safe=True
 )
 
 
@@ -255,6 +261,8 @@ def is_template(val, ctx=None):
 
 
 class _VarTrackerDict(dict):
+    ctx: Optional[RefContext] = None
+
     def __getitem__(self, key):
         try:
             val = super().__getitem__(key)
@@ -263,6 +271,7 @@ class _VarTrackerDict(dict):
             raise
 
         try:
+            assert self.ctx
             return self.ctx.resolve_reference(key)
         except KeyError:
             return val
@@ -291,7 +300,20 @@ def _wrap_sequence(v):
 unsafe_proxy._wrap_sequence = _wrap_sequence
 
 
-def apply_template(value, ctx, overrides=None):
+def _sandboxed_template(value: str, ctx: SafeRefContext, _UnfurlUndefined):
+    from jinja2.sandbox import SandboxedEnvironment
+    from jinja2.nativetypes import NativeCodeGenerator, native_concat
+
+    SandboxedEnvironment.code_generator_class = NativeCodeGenerator
+    SandboxedEnvironment.concat = staticmethod(native_concat)  # type: ignore
+    env = SandboxedEnvironment()
+
+    if not ctx.strict:
+        env.undefined = _UnfurlUndefined
+    return env.from_string(value).render(ctx.vars)
+
+
+def apply_template(value: str, ctx: RefContext, overrides=None) -> Any:
     if not isinstance(value, six.string_types):
         msg = f"Error rendering template: source must be a string, not {type(value)}"
         if ctx.strict:
@@ -302,7 +324,7 @@ def apply_template(value, ctx, overrides=None):
     if ctx.task:
         logger = ctx.task.logger
     else:
-        logger = logging.getLogger("unfurl")
+        logger = logging.getLogger("unfurl")  # type: ignore
 
     # local class to bind with logger and ctx
     class _UnfurlUndefined(DebugUndefined):
@@ -320,7 +342,7 @@ def apply_template(value, ctx, overrides=None):
         __getitem__ = __getattr__  # type: ignore
 
         def _log_message(self) -> None:
-            msg = "Template: %s", self._undefined_message
+            msg = "Template: %s" % self._undefined_message
             # XXX override _undefined_message: if self._undefined_obj is a Results then add its ctx._lastResource to the msg
             #     see https://github.com/pallets/jinja/blob/7d72eb7fefb7dce065193967f31f805180508448/src/jinja2/runtime.py#L824
             logger.warning(msg)
@@ -344,6 +366,9 @@ def apply_template(value, ctx, overrides=None):
         def __html__(self) -> str:
             return str(self)
 
+    if isinstance(ctx, SafeRefContext):
+        return _sandboxed_template(value, ctx, _UnfurlUndefined)
+
     # implementation notes:
     #   see https://github.com/ansible/ansible/test/units/template/test_templar.py
     #   dataLoader is only used by _lookup and to set _basedir (else ./)
@@ -367,14 +392,16 @@ def apply_template(value, ctx, overrides=None):
     templar.environment.trim_blocks = False
     # templar.environment.lstrip_blocks = False
     fail_on_undefined = ctx.strict
-    if not fail_on_undefined:
+    if not fail_on_undefined:  # note: strict is on by default
         templar.environment.undefined = _UnfurlUndefined
 
     vars = _VarTrackerDict(
         __unfurl=ctx, __python_executable=sys.executable, __now=time.time()
     )
     if hasattr(ctx.currentResource, "attributes"):
-        vars["SELF"] = ctx.currentResource.attributes
+        vars["SELF"] = ctx.currentResource.attributes  # type: ignore
+    if os.getenv("UNFURL_TEST_DEBUG_EX"):
+        logger.debug("template vars for %s: %s", value[:300], list(ctx.vars))
     vars.update(ctx.vars)
     vars.ctx = ctx
 
@@ -457,7 +484,7 @@ def _template_func(args, ctx):
     return apply_template(value, ctx, ctx.kw.get("overrides"))
 
 
-set_eval_func("template", _template_func)
+set_eval_func("template", _template_func, safe=True)
 
 
 def run_lookup(name, templar, *args, **kw):
@@ -494,7 +521,7 @@ def lookup_func(arg, ctx):
     assert_form(arg, test=arg)  # a map with at least one element
     name = None
     args = None
-    kwargs = {}
+    kwargs: Dict[str, Any] = {}
     for key, value in arg.items():
         if not name:
             name = key
@@ -539,7 +566,7 @@ def concat(args, ctx):
     return sep.join([str(a) for a in result])
 
 
-set_eval_func("concat", concat, True)
+set_eval_func("concat", concat, True, True)
 
 
 def token(args, ctx):
@@ -547,7 +574,7 @@ def token(args, ctx):
     return args[0].split(args[1])[args[2]]
 
 
-set_eval_func("token", token, True)
+set_eval_func("token", token, True, True)
 
 # XXX this doesn't work with node_filters, need an instance to get a specific result
 def get_tosca_property(args, ctx):
@@ -565,13 +592,13 @@ def has_env(arg, ctx):
     """
     {has_env: foo}
     """
-    return arg in os.environ
+    return arg in ctx.environ
 
 
 set_eval_func("has_env", has_env, True)
 
 
-def get_env(args, ctx: RefContext):
+def get_env(args, ctx: RefContext) -> Union[str, None, Dict[str, str]]:
     """
     Return the value of the given environment variable name.
     If NAME is not present in the environment, return the given default value if supplied or return None.
@@ -580,18 +607,22 @@ def get_env(args, ctx: RefContext):
 
     If the value of its argument is empty (e.g. [] or null), return the entire dictionary.
     """
-    env = os.environ
+    env = cast(Dict[str, str], ctx.environ)
     if not args:
         return env
 
     if isinstance(args, list):
         name = args[0]
-        default = args[1] if len(args) > 1 else None
+        default: Union[str, None] = args[1] if len(args) > 1 else None
     else:
         name = args
         default = None
 
-    return env.get(name, map_value(default, ctx))
+    if name in env:
+        return env[name]
+    else:
+        default = cast(Union[str, None], map_value(default, ctx))
+        return default
 
 
 set_eval_func("get_env", get_env, True)
@@ -599,7 +630,7 @@ set_eval_func("get_env", get_env, True)
 
 def set_context_vars(vars, resource):
     root = resource.root
-    ROOT = {}
+    ROOT: Dict[str, Any] = {}
     vars.update(dict(NODES=TopologyMap(root), ROOT=ROOT, TOPOLOGY=ROOT))
     if "inputs" in root._attributes:
         ROOT.update(
@@ -625,9 +656,14 @@ class _EnvMapper(dict):
     """Resolve environment variable name to instance properties via the root template's requirements.
     Pattern should match _generate_env_names in to_json.py and set_context_vars above.
     """
+    ctx: Optional[RefContext] = None
+
+    def copy(self):
+        return _EnvMapper(self)
 
     def __missing__(self, key):
         objname, sep, prop = key.partition("_")
+        assert self.ctx
         root = self.ctx.currentResource.root
         app = root.template.spec.substitution_template
         if app and objname and prop:
@@ -649,81 +685,148 @@ class _EnvMapper(dict):
 
 
 def to_env(args, ctx: RefContext):
-    env = None
-    if ctx.task:
-        env = ctx.task.get_environment(False)
+    env = ctx.environ
     sub = _EnvMapper(env or {})
-    sub.ctx = ctx  # type: ignore
+    sub.ctx = ctx
 
     rules = map_value(args or {}, ctx)
     assert isinstance(rules, Mapping)
     result = filter_env(rules, env, True, sub)
     if ctx.kw.get("update_os_environ"):
-        os.environ.update(result)
-        for key, value in rules.items():
-            if value is None and key in os.environ:
-                del os.environ[key]
+        log = ctx.task and ctx.task.logger or logger
+        log.debug("to_env is updating os.environ with %s using rules %s", result, rules)
+        # update all the copies of environ
+        envs = [ctx.environ, ctx.currentResource.root._environ, os.environ]
+        for env in envs:
+            env.update(result)
+            for key, value in rules.items():
+                if value is None and key in env:
+                    del env[key]
     return result
 
 
 set_eval_func("to_env", to_env)
 
 
-def to_label(arg, extra="", max=63, lower=False, replace=""):
+def to_label(arg, allowed=r"\w", max=63, case="any", replace="", start="a-zA-Z", start_prepend="x", **kw):
+    r"""Convert a string to q label with the given constraints. 
+        If a dictionary or list, all keys and string values are converted.
+
+    Args:
+        arg (str or dict or list): Convert to label
+        allowed (str, optional): Allowed characters. Regex character ranges and character classes. 
+                               Defaults to "\w"  (equivalent to [a-zA-Z0-9_])
+        replace (str, optional): String Invalidate. Defaults to "" (remove the characters).
+        start (str, optional): Allowed characters for the first character. Regex character ranges and character classes. 
+                               Defaults to "a-zA-Z"
+        start_prepend (str, optional): If the start character is invalid, prepend with this string (Default: "x")
+        max (int, optional): max length of label. Defaults to 63 (the maximum for a DNS name).
+        case (str, optional): "upper", "lower" or "any" (no conversion). Defaults to "any".
+    """
     if isinstance(arg, Mapping):
         return {
-            to_label(n, extra, max, lower, replace): to_label(
-                v, extra, max, lower, replace
+            to_label(n, allowed, max, case, replace): to_label(
+                v, allowed, max, case, replace
             )
             for n, v in arg.items()
             if v is not None
         }
+    elif isinstance(arg, list):
+        return [to_label(n, allowed, max, case, replace) for n in arg]
     elif isinstance(arg, str):
-        val = re.sub(f"[^\\w{extra}]", replace, arg)
-        if lower:
+        if arg and re.match(fr"[^{start}]", arg[0]):
+            val = start_prepend + arg
+        else:
+            val = arg
+        val = re.sub(fr"[^{allowed}]", replace, val)
+        if case == "lower":
             val = val.lower()
+        elif case == "upper":
+            val = val.upper()
         return val[:max]
+    else:
+        return arg
 
 
-def to_dns_label(arg):
+set_eval_func(
+    "to_label",
+    lambda arg, ctx: to_label(map_value(arg, ctx), **ctx.kw), safe=True
+)
+
+
+def to_dns_label(arg, max=63, **kw):
     """
-    The maximum length of each label is 63 characters, and a full domain name can have a maximum of 253 characters.
-    Alphanumeric characters and hyphens can be used in labels, but a domain name must not commence or end with a hyphen.
+    Convert the given argument (see `to_label` for full description) to a DNS label (a label is the name separated by "." in a domain name).
+    The maximum length of each label is 63 characters and can include
+    alphanumeric characters and hyphens but a domain name must not commence or end with a hyphen.
+
+    Invalid characters are replaced with "--".
     """
-    return to_label(arg, extra="-", replace="--")
+    return to_label(arg, allowed=r"\w-", start=r"\w", replace="--", max=max)
 
 
 set_eval_func(
     "to_dns_label",
-    lambda arg, ctx: to_dns_label(map_value(arg, ctx)),
+    lambda arg, ctx: to_dns_label(map_value(arg, ctx), **ctx.kw), safe=True
 )
 
 
 def to_kubernetes_label(arg):
     """
     See https://kubernetes.io/docs/concepts/overview/working-with-objects/labels/#syntax-and-character-set
+
+    Invalid characters are replaced with "__".
     """
-    return to_label(arg, extra="_.-", replace="__")
+    return to_label(arg, allowed=r"\w_.-", replace="__")
 
 
 set_eval_func(
     "to_kubernetes_label",
-    lambda arg, ctx: to_kubernetes_label(map_value(arg, ctx)),
+    lambda arg, ctx: to_kubernetes_label(map_value(arg, ctx)), safe=True
 )
 
 
-def to_googlecloud_label(arg):
+def to_googlecloud_label(arg, max=63, **kw):
     """
     See https://cloud.google.com/resource-manager/docs/creating-managing-labels#requirements
+
+    Invalid characters are replaced with "__".
     """
-    return to_label(arg, extra="_-", lower=True, replace="__")
+    return to_label(arg, allowed=r"\w_-", case="lower", replace="__", max=max)
 
 
 set_eval_func(
     "to_googlecloud_label",
-    lambda arg, ctx: to_googlecloud_label(map_value(arg, ctx)),
+    lambda arg, ctx: to_googlecloud_label(map_value(arg, ctx), **ctx.kw), safe=True
 )
 
+
+def get_ensemble_metadata(arg, ctx):
+    if not ctx.task:
+        return {}
+    ensemble = ctx.task._manifest
+    metadata = dict(
+        deployment=os.path.basename(os.path.dirname(ensemble.path)),
+        job=ctx.task.job.changeId
+    )
+    if ensemble.repo:
+        metadata["unfurlproject"] = ensemble.repo.project_path()
+        metadata["commit"] = ensemble.repo.revision[:8]
+    environment = ensemble.localEnv and ensemble.localEnv.manifest_context_name
+    if environment:
+        metadata["environment"] = environment
+    if arg:
+        key = map_value(arg, ctx)
+        if key == "project_namespace_subdomain" and ensemble.repo:
+            return ".".join(reversed(os.path.dirname(ensemble.repo.project_path()).split("/")))
+        return metadata.get(key, "")
+    else:
+        return metadata
+
+
+set_eval_func("get_ensemble_metadata", get_ensemble_metadata)
+
+
 _toscaKeywordsToExpr = {
     "SELF": ".",
     "SOURCE": ".source",
@@ -755,7 +858,7 @@ def get_attribute(args, ctx: RefContext):
     return ctx.query(query)
 
 
-set_eval_func("get_attribute", get_attribute, True)
+set_eval_func("get_attribute", get_attribute, True, True)
 
 
 def get_nodes_of_type(type_name, ctx: RefContext):
@@ -767,10 +870,10 @@ def get_nodes_of_type(type_name, ctx: RefContext):
     ]
 
 
-set_eval_func("get_nodes_of_type", get_nodes_of_type, True)
+set_eval_func("get_nodes_of_type", get_nodes_of_type, True, True)
 
 
-set_eval_func("_generate", lambda arg, ctx: get_random_password(10, ""), True)
+set_eval_func("_generate", lambda arg, ctx: get_random_password(10, ""), True, True)
 
 
 class ContainerImage(ExternalValue):
@@ -1008,9 +1111,6 @@ set_eval_func("external", get_import)
 
 
 class _Import:
-    if TYPE_CHECKING:
-        from .runtime import EntityInstance
-
     def __init__(
         self,
         external_instance: "EntityInstance",
@@ -1023,6 +1123,8 @@ class _Import:
 
 
 class Imports(collections.OrderedDict):
+    manifest: Optional["Manifest"] = None
+
     def find_import(self, qualified_name):
         # return a local shadow of the imported instance
         # or the imported instance itself if no local shadow exist (yet).
@@ -1030,12 +1132,13 @@ class Imports(collections.OrderedDict):
         if imported:
             return imported
         iName, sep, rName = qualified_name.partition(":")
+        assert self.manifest
         localEnv = self.manifest.localEnv
         if iName not in self and localEnv:
             project = localEnv.project or localEnv.homeProject
             tpl = project and project.find_ensemble_by_name(iName)
             if tpl:
-                self.manifest.load_external_ensemble(iName, dict(manifest=tpl))
+                self.manifest.load_external_ensemble(iName, dict(manifest=tpl))  # type: ignore
                 return self._find_import(qualified_name)
         return None
 
@@ -1085,7 +1188,7 @@ class ExternalResource(ExternalValue):
                 (message, schemaErrors) = messages
                 raise UnfurlValidationError(
                     f"schema validation failed for attribute '{name}': {schemaErrors}",
-                    schemaErrors,
+                    schemaErrors,  # type: ignore
                 )
 
     def _get_schema(self, name):
@@ -1164,6 +1267,8 @@ class DelegateAttributes:
                 return result
 
 
+AttributeChanges = NewType("AttributeChanges", Dict["InstanceKey", MutableMapping[str, Any]])
+
 class ResourceChanges(collections.OrderedDict):
     """
     Records changes made by configurations.
@@ -1181,17 +1286,23 @@ class ResourceChanges(collections.OrderedDict):
     addedIndex = 1
     attributesIndex = 2
 
-    def get_attribute_changes(self, key):
+    def get_attribute_changes(self, key: "InstanceKey"):
         record = self.get(key)
         if record:
             return record[self.attributesIndex]
         return {}
 
-    def sync(self, resource, changeId=None):
+    def get_changes_as_expr(self) -> Iterator[str]:
+        for key, record in self.items():
+            attribute_changes = record[self.attributesIndex]
+            if attribute_changes:
+                yield from (key + "::" + attr for attr in attribute_changes)
+
+    def sync(self, resource: "EntityInstance", changeId=None):
         """Update self to only include changes that are still live"""
         for k, v in list(self.items()):
-            current = Ref(k).resolve_one(RefContext(resource))
-            if current:
+            current = cast(Optional["EntityInstance"], Ref(k).resolve_one(RefContext(resource)))
+            if current is not None:
                 attributes = v[self.attributesIndex]
                 if attributes:
                     v[self.attributesIndex] = intersect_dict(
@@ -1205,7 +1316,7 @@ class ResourceChanges(collections.OrderedDict):
             else:
                 del self[k]
 
-    def add_changes(self, changes):
+    def add_changes(self, changes: AttributeChanges):
         for name, change in changes.items():
             old = self.get(name)
             if old:
@@ -1228,7 +1339,7 @@ class ResourceChanges(collections.OrderedDict):
         for resource in resources:
             self["::" + resource["name"]] = [None, resource, None]
 
-    def update_changes(self, changes, statuses, resource, changeId=None):
+    def update_changes(self, changes: AttributeChanges, statuses, resource, changeId=None):
         self.add_changes(changes)
         self.add_statuses(statuses)
         if resource:
@@ -1244,6 +1355,9 @@ class TopologyMap(dict):
     def __init__(self, resource):
         self.resource = resource
 
+    def copy(self):
+        return self
+
     def __getitem__(self, key):
         r = self.resource.find_resource(key)
         if r:
@@ -1258,6 +1372,11 @@ class TopologyMap(dict):
         return len(tuple(self.resource.get_self_and_descendents()))
 
 
+LiveDependencies = NewType('LiveDependencies',
+                           Dict["InstanceKey", Tuple["EntityInstance",
+                                           Dict[str, Tuple[bool, Union[Result, Any]]]]])
+
+
 class AttributeManager:
     """
     Tracks changes made to Resources
@@ -1270,11 +1389,12 @@ class AttributeManager:
     -- if a configurator wants to re-evaluate that attribute, it can create a dependency on it
     so to treat that as changed configuration.
     """
+    validate = True
 
     # what about an attribute that is added to the spec that already exists in status?
     # XXX2 tests for the above behavior
     def __init__(self, yaml=None, task=None):
-        self.attributes = {}
+        self.attributes: Dict[str, Tuple[EntityInstance, ResultsMap]] = {}
         self.statuses = {}
         self._yaml = yaml  # hack to safely expose the yaml context
         self.task = task
@@ -1307,7 +1427,7 @@ class AttributeManager:
                 resource.template is not template
                 and template not in resource.template._isReferencedBy
             ):
-                resource.template._isReferencedBy.append(template)
+                resource.template._isReferencedBy.append(template)  # type: ignore
 
     def _get_context(self, resource):
         if self._context_vars is None:
@@ -1315,28 +1435,25 @@ class AttributeManager:
             set_context_vars(self._context_vars, resource)
         return RefContext(resource, self._context_vars, task=self.task)
 
-    def get_attributes(self, resource):
+    def get_attributes(self, resource: "EntityInstance"):
         if resource.key not in self.attributes:
             if resource.shadow:
                 return resource.shadow.attributes
 
-            # deepcopy() because lazily created ResultMaps and ResultLists will mutate
-            # the underlying nested structures when resolving values
             if resource.template:
                 specAttributes = resource.template.defaultAttributes
                 _attributes = ChainMap(
-                    copy.deepcopy(resource._attributes),
-                    copy.deepcopy(resource.template.properties),
-                    copy.deepcopy(specAttributes),
+                    resource._attributes,
+                    resource.template.properties,
+                    specAttributes,
                 )
             else:
-                _attributes = ChainMap(copy.deepcopy(resource._attributes))
+                _attributes = ChainMap(resource._attributes)
             ctx = self._get_context(resource)
-            validate = True
             mode = os.getenv("UNFURL_VALIDATION_MODE")
             if mode is not None and "nopropcheck" in mode:
-                validate = False
-            attributes = ResultsMap(_attributes, ctx, validate=validate)
+                self.validate = False
+            attributes = ResultsMap(_attributes, ctx, validate=self.validate)
             self.attributes[resource.key] = (resource, attributes)
             return attributes
         else:
@@ -1371,26 +1488,23 @@ class AttributeManager:
         )
         return changed, live
 
-    def find_live_dependencies(self):
-        liveDependencies = {}
+    def find_live_dependencies(self) -> Dict["InstanceKey", List["Dependency"]]:
+        from .configurator import Dependency
+        dependencies: Dict[InstanceKey, List["Dependency"]] = {} 
         for resource, attributes in self.attributes.values():
             overrides, specd = attributes._attributes.split()
-            live = []
             # items in overrides of type Result have been accessed during this transaction
             for key, value in overrides.items():
                 if isinstance(value, Result):
                     changed, isLive = self._check_attribute(specd, key, value, resource)
                     if isLive:
-                        live.append(key)
+                        dep = Dependency(resource.key + "::" + key, value, target=resource)
+                        dependencies.setdefault(resource.key, []).append(dep)
+        return dependencies
 
-            if live:
-                liveDependencies[resource.key] = (resource, live)
-
-        return liveDependencies
-
-    def commit_changes(self):
-        changes = {}
-        liveDependencies = {}
+    def commit_changes(self) -> Tuple[AttributeChanges, LiveDependencies]:
+        changes: AttributeChanges = cast(AttributeChanges, {})
+        liveDependencies = cast(LiveDependencies, {})
         for resource, attributes in list(self.attributes.values()):
             overrides, specd = attributes._attributes.split()
             # overrides will only contain:
@@ -1399,7 +1513,7 @@ class AttributeManager:
             _attributes = {}
             defs = resource.template and resource.template.propertyDefs or {}
             foundSensitive = []
-            live = {}
+            live: Dict[str, Any] = {}
             # items in overrides of type Result have been accessed during this transaction
             for key, value in list(overrides.items()):
                 if not isinstance(value, Result):
