@@ -16,6 +16,7 @@ import os.path
 import sys
 import itertools
 import json
+import datetime
 from typing import Any, Dict, List, Optional, Union, cast, TYPE_CHECKING
 from collections import Counter
 from collections.abc import Mapping, MutableSequence
@@ -32,20 +33,28 @@ from toscaparser.elements.scalarunit import get_scalarunit_class
 from toscaparser.elements.datatype import DataType
 from toscaparser.elements.portspectype import PortSpec
 from toscaparser.activities import ConditionClause
-from .logs import sensitive, is_sensitive
+from .logs import sensitive, is_sensitive, getLogger
 from .tosca import is_function, get_nodefilters
-from .localenv import LocalEnv
 from .util import to_enum, UnfurlError
 from .support import Status, is_template
-import logging
+from .result import ChangeRecord
+from .localenv import LocalEnv
 
-logger = logging.getLogger("unfurl")
+logger = getLogger("unfurl")
 
 
 def _print(*args):
     print(*args, file=sys.stderr)
 
 
+def js_timestamp(ts: datetime.datetime) -> str:
+    jsts = ts.isoformat(' ', 'seconds')
+    if ts.utcoffset() is None:
+        # if no timezone assume utc and append the missing offset
+        return jsts + "+00:00"
+    return jsts
+
+
 numeric_constraints = {
     "greater_than": "exclusiveMinimum",
     "greater_or_equal": "minimum",
@@ -107,7 +116,7 @@ def map_constraints(jsonType, constraints, schema):
 ONE_TO_ONE_TYPES = ("string", "boolean", "map", "list")
 
 #  see constraints.PROPERTY_TYPES
-VALUE_TYPES = dict(
+VALUE_TYPES: Dict[str, Dict] = dict(
     integer={"type": "number"},
     string={"type": "string"},
     boolean={"type": "boolean"},
@@ -137,14 +146,13 @@ VALUE_TYPES.update(
 
 
 def tosca_type_to_jsonschema(spec, propdefs, toscatype):
-    propdefs = [p for p in propdefs if is_property_user_visible(p)]
     jsonschema = dict(
         type="object",
-        properties={p.name: tosca_schema_to_jsonschema(p, spec) for p in propdefs},
+        properties = {p.name: tosca_schema_to_jsonschema(p, spec) for p in propdefs},
     )
     # add typeid:
     if toscatype:
-        jsonschema["properties"].update({"$toscatype": dict(const=toscatype)})
+        jsonschema["properties"].update({"$toscatype": dict(const=toscatype)})  # type: ignore
     return jsonschema
 
 
@@ -220,8 +228,9 @@ def tosca_schema_to_jsonschema(p, spec):
             if valuetype_schema.constraints:  # merge constraints
                 constraints = valuetype_schema.constraints + constraints
         else:  # it's a complex type with properties:
+            propdefs = [p for p in dt.get_properties_def_objects() if is_property_user_visible(p)]
             type_schema = tosca_type_to_jsonschema(
-                spec, dt.get_properties_def_objects(), dt.type
+                spec, propdefs, dt.type
             )
     if tosca_type not in ONE_TO_ONE_TYPES and "properties" not in schema:
         schema["$toscatype"] = tosca_type
@@ -272,6 +281,10 @@ def _get_req(req_dict):
     return name, req
 
 
+def expand_prefix(spec, nodetype: str):
+    return nodetype.replace("tosca:", "tosca.nodes.")  # XXX
+
+
 def requirement_to_graphql(spec, req_dict):
     """
     type RequirementConstraint {
@@ -319,11 +332,15 @@ def requirement_to_graphql(spec, req_dict):
     nodetype = req.get("node")
     if nodetype:
         # req['node'] can be a node_template instead of a type
+        expand_prefix(spec, nodetype)
         if nodetype in spec.nodeTemplates:
             reqobj["match"] = nodetype
             nodetype = spec.nodeTemplates[nodetype].type
     else:
-        nodetype = req.get("capability") or "tosca.nodes.Root"
+        nodetype = req.get("capability")
+        if not nodetype:
+            logger.warning("skipping constraint %s, there was no type specified ", req_dict)
+            return None
     reqobj["resourceType"] = nodetype
     if req.get("node_filter"):
         reqobj["node_filter"] = req["node_filter"]
@@ -385,7 +402,7 @@ def is_computed(p):  # p: Property | PropertyDef
     return (
         p.name in ["tosca_id", "state", "tosca_name"]
         or is_value_computed(p.value)
-        or is_value_computed(p.default)
+        or (p.value is None and is_value_computed(p.default))
         or metadata.get("computed")
     )
 
@@ -512,6 +529,9 @@ def node_type_to_graphql(spec, type_definition, types: dict):
     jsontype["extends"] = extends
     if not type_definition.is_derived_from("tosca.nodes.Root"):
         return jsontype
+    if type_definition.defs is None:
+        logger.warning("%s is missing type definition", type_definition.type)
+        return jsontype
 
     # add capabilities types to extends
     for cap in type_definition.get_capability_typedefs():
@@ -568,7 +588,7 @@ def _make_typedef(typename, custom_defs):
 
 def to_graphql_nodetypes(spec):
     # node types are readonly, so mapping doesn't need to be bijective
-    types = {}
+    types: Dict[str, Dict[str, Any]] = {}
     custom_defs = spec.template.topology_template.custom_defs
     for typename in custom_defs:
         if typename in types:
@@ -618,8 +638,9 @@ def add_capabilities_as_properties(schema, nodetype, spec):
     for cap in nodetype.get_capability_typedefs():
         if not cap.get_definition("properties"):
             continue
+        propdefs = [p for p in cap.get_properties_def_objects() if is_property_user_visible(p)]
         schema[cap.name] = tosca_type_to_jsonschema(
-            spec, cap.get_properties_def_objects(), cap.type
+            spec, propdefs, cap.type
         )
 
 
@@ -628,8 +649,9 @@ def add_capabilities_as_attributes(schema, nodetype, spec):
     for cap in nodetype.get_capability_typedefs():
         if not cap.get_definition("attributes"):
             continue
+        propdefs = [p for p in cap.get_attributes_def_objects() if is_property_user_visible(p)]
         schema[cap.name] = tosca_type_to_jsonschema(
-            spec, cap.get_attributes_def_objects(), cap.type
+            spec, propdefs, cap.type
         )
 
 
@@ -715,7 +737,7 @@ def nodetemplate_to_json(nodetemplate, spec, types, for_resource=False):
     json = dict(
         type=nodetemplate.type,
         name=nodetemplate.name,
-        title=nodetemplate.name,
+        title=nodetemplate.entity_tpl.get("metadata", {}).get("title") or nodetemplate.name,
         description=nodetemplate.entity_tpl.get("description") or "",
         directives=nodetemplate.directives,
     )
@@ -725,12 +747,12 @@ def nodetemplate_to_json(nodetemplate, spec, types, for_resource=False):
     jsonnodetype = types[nodetemplate.type]
     visitor = PropertyVisitor()
     json["properties"] = list(template_properties_to_json(nodetemplate, visitor))
-    if visitor.redacted and "predefined" not in nodetemplate.directives:
-        json["directives"].append("predefined")
-        logger.warning(
-            "Adding 'predefined' directive to '%s' because it has redacted properties",
-            nodetemplate.name,
-        )
+    # if visitor.redacted and "predefined" not in nodetemplate.directives:
+    #     json["directives"].append("predefined")
+    #     logger.warning(
+    #         "Adding 'predefined' directive to '%s' because it has redacted properties",
+    #         nodetemplate.name,
+    #     )
     json["dependencies"] = []
     visibility = template_visibility(spec, nodetemplate, for_resource)
     if visibility != "inherit":
@@ -759,15 +781,16 @@ def nodetemplate_to_json(nodetemplate, spec, types, for_resource=False):
         if name in ["dependency", "installer"]:
             # skip artifact requirements and the base TOSCA relationship type that every node has
             continue
+        occurrences = req_dict.get("occurrences")
+        if occurrences and occurrences[1] == 0:
+            continue  # skip when max occurrences is 0
+
+        # we need to call _get_explicit_relationship() to make sure all the requirements relationships are created
         reqDef, rel_template = nodetemplate._get_explicit_relationship(name, req_dict)
         reqconstraint = _find_requirement_constraint(jsonnodetype["requirements"], name)
         reqjson = dict(constraint=reqconstraint, name=name, __typename="Requirement")
-        req_metadata = req_dict.get("metadata") or {}
-        visibility = (
-            req_metadata
-            and "constraint" in req_metadata
-            and req_metadata["constraint"].get("visibility")
-        )
+        req_metadata = req_dict.get("metadata") or req_dict.get("constraint") or {}
+        visibility = req_metadata and req_metadata.get("visibility")
         if req_dict.get("node") and not _get_typedef(req_dict["node"], spec):
             # it's not a type, assume it's a node template
             # (the node template name might not match a template if it is only defined in the deployment blueprints)
@@ -807,7 +830,7 @@ primary_name = "__primary"
 
 
 def _generate_primary(spec, db, node_tpl=None):
-    base_type = node_tpl["type"] if node_tpl else "tosca:Root"
+    base_type = node_tpl["type"] if node_tpl else "tosca.nodes.Root"
     topology = spec.template.topology_template
     # generate a node type and node template that represents root of the topology
     # generate a type that exposes the topology's inputs and outputs as properties and attributes
@@ -891,7 +914,7 @@ def _get_or_make_primary(spec, db):
     return root.name, root.type
 
 
-def to_graphql_blueprint(spec, db, deploymentTemplates=None):
+def to_graphql_blueprint(spec, db):
     """
     Returns json object as ApplicationBlueprint
 
@@ -914,7 +937,7 @@ def to_graphql_blueprint(spec, db, deploymentTemplates=None):
     title, name = _template_title(spec, root_name)
     blueprint = dict(__typename="ApplicationBlueprint", name=name, title=title)
     blueprint["primary"] = root_type
-    blueprint["deploymentTemplates"] = deploymentTemplates or []
+    blueprint["deploymentTemplates"] = []
     blueprint["description"] = spec.template.description
     metadata = spec.template.tpl.get("metadata") or {}
     blueprint["livePreview"] = metadata.get("livePreview")
@@ -1052,12 +1075,17 @@ def get_deployment_blueprints(manifest, blueprint, root_name, db):
     return deployment_templates
 
 
+def _project_path(url):
+    pp = urlparse(url).path.lstrip("/")
+    if pp.endswith(".git"):
+        return pp[:-4]
+    return pp
+
+
 def get_blueprint_from_topology(manifest, db):
     spec = manifest.tosca
-    title = os.getenv("DEPLOYMENT") or os.path.basename(os.path.dirname(manifest.path))
-    slug = slugify(title)
     # XXX cloud = spec.topology.primary_provider
-    blueprint, root_name = to_graphql_blueprint(spec, db, [title])
+    blueprint, root_name = to_graphql_blueprint(spec, db)
     templates = get_deployment_blueprints(manifest, blueprint, root_name, db)
 
     template = {}
@@ -1066,9 +1094,11 @@ def get_blueprint_from_topology(manifest, db):
         deployment_blueprint = templates[deployment_blueprint_name]
         template = deployment_blueprint.copy()
 
+    # the deployment template created for this deployment will have a "source" key
+    # so if it doesn't (or one wasn't set) create a new one and set the current one as its "source"
     if "source" not in template:
-        # the deployment template created for this deployment will have a "source" key
-        # so if it doesn't (or one wasn't set) create a new one and set the current one as its "source"
+        title = os.path.basename(os.path.dirname(manifest.path))
+        slug = slugify(title)
         template.update(
             dict(
                 __typename="DeploymentTemplate",
@@ -1080,18 +1110,22 @@ def get_blueprint_from_topology(manifest, db):
                 resourceTemplates=sorted(db["ResourceTemplate"]),
                 ResourceTemplate={},
                 source=deployment_blueprint_name,
-                projectPath=urlparse(manifest.repositories["spec"].url)
-                .path.lstrip("/")
-                .rstrip(".git"),
+                projectPath=_project_path(manifest.repositories["spec"].url),
             )
         )
+        templates[slug] = template
+
+    blueprint["deploymentTemplates"] = list(templates)
     template["blueprint"] = blueprint["name"]
     template["primary"] = root_name
     template["environmentVariableNames"] = list(_generate_env_names(spec, root_name))
+    last_commit_time = manifest.last_commit_time()
+    if last_commit_time:
+        template["commitTime"] = js_timestamp(datetime.datetime.fromtimestamp(last_commit_time))
     return blueprint, template
 
 
-def to_graphql(localEnv):
+def _to_graphql(localEnv):
     # set skip_validation because we want to be able to dump incomplete service templates
     manifest = localEnv.get_manifest(skip_validation=True)
     db = {}
@@ -1157,13 +1191,14 @@ def add_graphql_deployment(manifest, db, dtemplate):
       status: Status
       summary: String
       workflow: String
+      deployTime: String
     }
     """
     name = dtemplate["name"]
     title = dtemplate.get("title") or name
     deployment = dict(name=name, title=title, __typename="Deployment")
     templates = db["ResourceTemplate"]
-    relationships = {}
+    relationships: Dict[str, List[Dict]] = {}
     for t in templates.values():
         if "dependencies" in t:
             for c in t["dependencies"]:
@@ -1188,10 +1223,9 @@ def add_graphql_deployment(manifest, db, dtemplate):
             )
             if workflow == "undeploy" and deployment["status"] == Status.ok:
                 deployment["status"] = Status.absent
-
         deployment["summary"] = manifest.lastJob.get("summary")
-    deployment["ci_job_id"] = os.getenv("CI_JOB_ID")
-    deployment["ci_pipeline_id"] = os.getenv("CI_PIPELINE_ID")
+        deployment["deployTime"] = js_timestamp(datetime.datetime.strptime(
+            manifest.lastJob["startTime"], ChangeRecord.DateTimeFormat))
 
     url = manifest.rootResource.attributes["outputs"].get("url")
     primary_resource = db["Resource"].get(primary_name)
@@ -1209,7 +1243,7 @@ def add_graphql_deployment(manifest, db, dtemplate):
 
 
 def to_blueprint(localEnv, existing=None):
-    db, manifest, env, env_types = to_graphql(localEnv)
+    db, manifest, env, env_types = _to_graphql(localEnv)
     blueprint, root_name = to_graphql_blueprint(manifest.tosca, db)
     deployment_blueprints = get_deployment_blueprints(
         manifest, blueprint, root_name, db
@@ -1222,13 +1256,15 @@ def to_blueprint(localEnv, existing=None):
 
 # NB! to_deployment is the default export format used by __main__.export (but you won't find that via grep)
 def to_deployment(localEnv, existing=None):
-    db, manifest, env, env_types = to_graphql(localEnv)
+    logger.debug("exporting deployment %s", localEnv.manifestPath)
+    db, manifest, env, env_types = _to_graphql(localEnv)
     blueprint, dtemplate = get_blueprint_from_topology(manifest, db)
     db["DeploymentTemplate"] = {dtemplate["name"]: dtemplate}
     db["ApplicationBlueprint"] = {blueprint["name"]: blueprint}
-    deployment = add_graphql_deployment(manifest, db, dtemplate)
+    add_graphql_deployment(manifest, db, dtemplate)
     # don't include base node type:
     db["ResourceType"].pop("tosca.nodes.Root")
+    logger.debug("finished exporting deployment %s", localEnv.manifestPath)
     return db
 
 
@@ -1236,7 +1272,7 @@ def mark_leaf_types(types):
     # types without implementations can't be instantiated by users
     # treat non-leaf types as abstract types that shouldn't be instatiated
     # treat leaf types that have requirements as having implementations
-    counts = Counter()
+    counts: Counter = Counter()
     for rt in types.values():
         for name in rt["extends"]:
             counts[name] += 1
@@ -1286,6 +1322,17 @@ def map_nodefilter(filters, jsonprops):
             schema.update(map_constraints(schema["type"], constraints, schema))
 
 
+def _set_shared_instances(instances):
+    env_instances = {}
+    # imported (aka shared) resources are just a reference and won't be part of the manifest unless referenced
+    # so just copy them over now
+    if instances:
+        for instance_name, value in instances.items():
+            if "imported" in value:
+                env_instances[instance_name] = value
+    return env_instances
+
+
 def to_environments(localEnv, existing=None):
     """
       Map environments in unfurl.yaml to DeploymentEnvironments
@@ -1316,7 +1363,7 @@ def to_environments(localEnv, existing=None):
     """
 
     # XXX one manifest and blueprint per environment
-    if existing:
+    if existing and os.path.exists(existing):
         with open(existing) as f:
             db = json.load(f)
     else:
@@ -1324,35 +1371,74 @@ def to_environments(localEnv, existing=None):
     environments = {}
     all_connection_types = {}
     blueprintdb = None
+    deployments = set_deploymentpaths(localEnv.project)["DeploymentPath"]
+    env_deployments = {}
+    for ensemble_info in deployments.values():
+        env_deployments[ensemble_info["environment"]] = ensemble_info["name"]
+    defaults = localEnv.project.contexts.get("defaults")
+    default_imported_instances = _set_shared_instances(defaults and defaults.get("instances"))
     for name in localEnv.project.contexts:
-        # we create new LocalEnv for each context because we need to instantiate a different ToscaSpec object
-        blueprintdb, manifest, env, env_types = to_graphql(
-            LocalEnv(
-                localEnv.manifestPath,
-                project=localEnv.project,
-                override_context=name,
-            )
-        )
-        env["name"] = name
-        environments[name] = env
-        all_connection_types.update(env_types)
+        try:
+            if name in env_deployments:
+                # we can reuse the localEnv if there's a distinct manifest that uses this environment
+                localEnv.manifest_context_name = name
+                localEnv.manifestPath = os.path.join(localEnv.project.projectRoot, env_deployments[name], "ensemble.yaml")
+                localLocalEnv = localEnv
+            else:
+                # this environment doesn't have any deployments so we have to create a new localEnv
+                # because we need to instantiate a different ToscaSpec object
+                localLocalEnv = LocalEnv(
+                    localEnv.manifestPath,
+                    project=localEnv.project,
+                    override_context=name,
+                    readonly=True,
+                )
+            blueprintdb, manifest, env, env_types = _to_graphql(localLocalEnv)
+            env["name"] = name
+            env["instances"].update(default_imported_instances)
+            instances = localEnv.project.contexts[name].get("instances")
+            env["instances"].update(_set_shared_instances(instances))
+            environments[name] = env
+            all_connection_types.update(env_types)
+        except Exception:
+            logger.error("error exporting environment %s", name, exc_info=True)
 
     db["DeploymentEnvironment"] = environments
     if blueprintdb:
-        if blueprintdb.get("repositories", {}).get("types"):
-            # don't include ResourceTypes if we are including a types repository
-            db["ResourceType"] = {}
-            return db
+        # XXX re-enable this?
+        # if blueprintdb.get("repositories", {}).get("types"):
+        #     # don't include ResourceTypes if we are including a types repository
+        #     db["ResourceType"] = all_connection_types
+        #     return db
 
         # add the rest of the types too
         # XXX is it safe to only include types with "connect" implementations?
         all_connection_types.update(blueprintdb["ResourceType"])
     db["ResourceType"] = all_connection_types
+    db["DeploymentPath"] = deployments
     return db
 
 
 def to_deployments(localEnv, existing=None):
-    return set_deploymentpaths(localEnv.project, existing)
+    db = set_deploymentpaths(localEnv.project)
+    deployments = db["deployments"] = []
+    for manifest_path, dp in db["DeploymentPath"].items():
+        try:
+            # optimization: reuse localenv
+            localEnv.manifestPath = os.path.join(localEnv.project.projectRoot, manifest_path, "ensemble.yaml")
+            environment = dp.get("environment") or "defaults"
+            localEnv.manifest_context_name = environment
+            deployments.append(to_deployment(localEnv))
+        except Exception:
+            logger.error("error exporting deployment %s", manifest_path, exc_info=True)
+
+    # from .yamlloader import yaml_perf
+    # _print(f"exported {len(deployments)} deployments, {yaml_perf} seconds parsing yaml")
+    return db
+
+
+def get_deploymentpaths(project):
+    return set_deploymentpaths(project)["DeploymentPath"]
 
 
 def set_deploymentpaths(project, existing=None):
@@ -1366,15 +1452,21 @@ def set_deploymentpaths(project, existing=None):
         if "environment" in ensemble_info and "project" not in ensemble_info:
             # exclude external ensembles
             path = os.path.dirname(ensemble_info["file"])
-            if path in deployment_paths:
-                continue  # don't overwrite
-            deployment_paths[path] = {
+            if os.path.isabs(path):
+                path = project.get_relative_path(path)
+            obj = {
                 "__typename": "DeploymentPath",
                 "name": path,
-                "project_id": None,
-                "pipelines": [],
+                "project_id": ensemble_info.get("project_id"),
+                "pipelines": ensemble_info.get("pipelines", []),
                 "environment": ensemble_info["environment"],
+                "incremental_deploy": ensemble_info.get("incremental_deploy", False)
             }
+            if path in deployment_paths:
+                # merge duplicate entries
+                deployment_paths[path].update(obj)
+            else:
+                deployment_paths[path] = obj
     return db
 
 
