@@ -13,7 +13,7 @@ from .util import (
     check_class_registry,
     env_var_value,
 )
-from .eval import Ref, RefContext, map_value
+from .eval import Ref, RefContext, SafeRefContext, map_value
 from .result import ExternalValue, ResourceRef, ResultsList
 from .merge import patch_dict, merge_dicts
 from .logs import get_console_log_level
@@ -32,7 +32,7 @@ import os
 from .logs import getLogger
 import logging
 import re
-from typing import Dict, Optional
+from typing import Dict, Optional, Sequence
 
 from ruamel.yaml.comments import CommentedMap
 
@@ -92,22 +92,27 @@ def find_standard_interface(op):
         return ""
 
 
-@functools.lru_cache(maxsize=None)
-def create_default_topology():
-    tpl = dict(
-        tosca_definitions_version=TOSCA_VERSION,
-        topology_template=dict(
-            node_templates={"_default": {"type": "tosca.nodes.Root"}},
-            relationship_templates={"_default": {"type": "tosca.relationships.Root"}},
-        ),
-    )
-    return ToscaTemplate(yaml_dict_tpl=tpl)
+__default_topology = None
+
+
+def get_default_topology():
+    global __default_topology
+    if __default_topology is None:
+        tpl = dict(
+            tosca_definitions_version=TOSCA_VERSION,
+            topology_template=dict(
+                node_templates={"_default": {"type": "tosca.nodes.Root"}},
+                relationship_templates={"_default": {"type": "tosca.relationships.Root"}},
+            ),
+        )
+        __default_topology = ToscaTemplate(yaml_dict_tpl=tpl)
+    return __default_topology
 
 
 def _patch(node, patchsrc, quote=False, tpl=None):
     if tpl is None:
         tpl = node.toscaEntityTemplate.entity_tpl
-    ctx = RefContext(node, dict(template=tpl))
+    ctx = SafeRefContext(node, dict(template=tpl))
     ctx.base_dir = getattr(patchsrc, "base_dir", ctx.base_dir)
     if quote:
         patch = copy.deepcopy(patchsrc)
@@ -133,8 +138,9 @@ class ToscaSpec:
                 modified.append(import_tpl)
                 continue
 
+            assert self.topology
             match = Ref(import_tpl["when"]).resolve_one(
-                RefContext(self.topology, trace=0)
+                SafeRefContext(self.topology, trace=0)
             )
             if match:
                 logger.debug(
@@ -183,11 +189,12 @@ class ToscaSpec:
 
     def _overlay(self, overlays):
         def _find_matches():
+            assert self.topology
             ExceptionCollector.start()  # clears previous errors
             for expression, _tpl in overlays.items():
                 try:
                     match = Ref(expression).resolve_one(
-                        RefContext(self.topology, trace=0)
+                        SafeRefContext(self.topology, trace=0)
                     )
                     if not match:
                         continue
@@ -196,7 +203,7 @@ class ToscaSpec:
                             yield (item, _tpl)
                     else:
                         yield (match, _tpl)
-                except:
+                except Exception:
                     ExceptionCollector.appendException(
                         UnfurlValidationError(
                             f'error evaluating decorator match expression "{expression}"',
@@ -405,7 +412,7 @@ class ToscaSpec:
                 workflows.setdefault(name, []).append(Workflow(w))
         self._workflows = workflows
 
-    def get_workflow(self, workflow):
+    def get_workflow(self, workflow: str) -> Optional["Workflow"]:
         # XXX need api to get all the workflows with the same name
         wfs = self._workflows.get(workflow)
         if wfs:
@@ -563,6 +570,8 @@ class ToscaSpec:
             if template.name not in self.relationshipTemplates:  # not defined yet
                 self.relationshipTemplates[template.name] = relTemplate
 
+    def get_repository(self, name: str):
+        return self.template and self.template.repositories.get(name)
 
 def find_env_vars(props_iter):
     for propdef, value in props_iter:
@@ -627,10 +636,10 @@ def find_props(attributes, propertyDefs, flatten=False):
 # represents a node, capability or relationship
 class EntitySpec(ResourceRef):
     # XXX need to define __eq__ for spec changes
-    def __init__(self, toscaNodeTemplate=None, spec=None):
+    def __init__(self, toscaNodeTemplate: Optional[EntityTemplate], spec: ToscaSpec):
         if not toscaNodeTemplate:
             toscaNodeTemplate = next(
-                iter(create_default_topology().topology_template.nodetemplates)
+                iter(get_default_topology().topology_template.nodetemplates)
             )
         self.toscaEntityTemplate = toscaNodeTemplate
         self.spec = spec
@@ -644,7 +653,7 @@ class EntitySpec(ResourceRef):
             )
 
         self.type = toscaNodeTemplate.type
-        self._isReferencedBy = (
+        self._isReferencedBy: Sequence[EntitySpec] = (
             []
         )  # this is referenced by another template or via property traversal
         # nodes have both properties and attributes
@@ -653,13 +662,9 @@ class EntitySpec(ResourceRef):
         # XXX user should be able to declare default attribute values
         self.propertyDefs = toscaNodeTemplate.get_properties()
         self.attributeDefs = {}
-        # XXX test_helm.py fails without making a deepcopy
-        # some how chart_values is being modifying outside of a task transaction
-        self.properties = copy.deepcopy(
-            CommentedMap(
+        self.properties = CommentedMap(
                 [(prop.name, prop.value) for prop in self.propertyDefs.values()]
             )
-        )
         if toscaNodeTemplate.type_definition:
             # add attributes definitions
             attrDefs = toscaNodeTemplate.type_definition.get_attributes_def()
@@ -816,7 +821,7 @@ class EntitySpec(ResourceRef):
     def aggregate_only(self):
         "The template is only the sum of its parts."
         for iDef in self.get_interfaces():
-            if iDef.interfacename == "Standard":
+            if iDef.interfacename in ("Standard", "Configure"):
                 return False
             if iDef.interfacename == "Install" and iDef.name == "discover":
                 return False
@@ -830,7 +835,7 @@ class EntitySpec(ResourceRef):
         pass
 
     @property
-    def required(self):
+    def required(self) -> bool:
         # check if this template is required by another template
         for root in _get_roots(self):
             if self.spec.substitution_template:
@@ -842,8 +847,12 @@ class EntitySpec(ResourceRef):
                 return True
         return False
 
+    @property
+    def environ(self):
+        return os.environ
 
-def _get_roots(node, seen=None):
+
+def _get_roots(node: EntitySpec, seen=None):
     # node can reference each other's properties, so we need to handle circular references
     if seen is None:
         seen = set()
@@ -859,14 +868,14 @@ class NodeSpec(EntitySpec):
     def __init__(self, template=None, spec=None):
         if not template:
             template = next(
-                iter(create_default_topology().topology_template.nodetemplates)
+                iter(get_default_topology().topology_template.nodetemplates)
             )
-            spec = ToscaSpec(create_default_topology())
+            spec = ToscaSpec(get_default_topology())
         else:
             assert spec
         EntitySpec.__init__(self, template, spec)
         self._capabilities = None
-        self._requirements = None
+        self._requirements: Optional[Dict[str, "RequirementSpec"]] = None
         self._relationships = []
         self._artifacts = None
 
@@ -980,7 +989,7 @@ class NodeSpec(EntitySpec):
     def get_capability(self, name):
         return self.capabilities.get(name)
 
-    def add_relationship(self, reqSpec):
+    def add_relationship(self, reqSpec: "RequirementSpec"):
         # find the relationship for this requirement:
         for relSpec in self._get_relationship_specs():
             # the RelationshipTemplate should have had the source node assigned by the tosca parser
@@ -1006,6 +1015,7 @@ class NodeSpec(EntitySpec):
                 )
                 if not relSpec.requirement:
                     relSpec.requirement = reqSpec
+                    relSpec._isReferencedBy.append(self)  # type: ignore
                 break
         else:
             msg = f'relationship not found for requirement "{reqSpec.name}" on "{reqSpec.parentNode}" targeting "{self.name}"'
@@ -1045,9 +1055,9 @@ class RelationshipSpec(EntitySpec):
         # its RelationshipType has valid_target_types
         if not template:
             template = (
-                create_default_topology().topology_template.relationship_templates[0]
+                get_default_topology().topology_template.relationship_templates[0]
             )
-            spec = ToscaSpec(create_default_topology())
+            spec = ToscaSpec(get_default_topology())
         else:
             assert spec
         EntitySpec.__init__(self, template, spec)
@@ -1120,12 +1130,12 @@ class RequirementSpec:
     """
 
     # XXX need __eq__ since this doesn't derive from EntitySpec
-    def __init__(self, name, req, parent, type_tpl):
-        self.source = self.parentNode = parent  # NodeSpec
+    def __init__(self, name: str, req: dict, parent: NodeSpec, type_tpl: dict):
+        self.source = self.parentNode = parent
         self.spec = parent.spec
         self.name = name
         self.entity_tpl = req
-        self.relationship = None
+        self.relationship: Optional[RelationshipSpec] = None
         self.type_tpl = type_tpl
         # entity_tpl may specify:
         # capability (definition name or type name), node (template name or type name), and node_filter,
@@ -1224,8 +1234,8 @@ class TopologySpec(EntitySpec):
             self.spec = spec
             template = spec.template.topology_template
         else:
-            template = create_default_topology().topology_template
-            self.spec = ToscaSpec(create_default_topology())
+            template = get_default_topology().topology_template
+            self.spec = ToscaSpec(get_default_topology())
             self.spec.topology = self
 
         inputs = inputs or {}
@@ -1344,10 +1354,11 @@ class ArtifactSpec(EntitySpec):
         "file_extensions",
     )
 
-    def __init__(self, artifact_tpl, template=None, spec=None, path=None):
+    def __init__(self, artifact_tpl, template: Optional[EntitySpec] = None, spec: Optional[ToscaSpec] = None, path=None):
         # 3.6.7 Artifact definition p. 84
         self.parentNode = template
         spec = template.spec if template else spec
+        assert spec
         if isinstance(artifact_tpl, toscaparser.artifacts.Artifact):
             artifact = artifact_tpl
         else:
@@ -1362,7 +1373,7 @@ class ArtifactSpec(EntitySpec):
         self.repository: Optional[toscaparser.repositories.Repository] = (
             spec
             and artifact.repository
-            and spec.template.repositories.get(artifact.repository)
+            and spec.get_repository(artifact.repository)
             or None
         )
         # map artifacts fields into properties
