@@ -556,6 +556,10 @@ class ChainMap(MutableMapping):
     def __init__(self, *maps: MutableMapping) -> None:
         self._maps = maps
 
+    def copy(self):
+         # assume map implements copy()
+        return ChainMap(*(m.copy() for m in self._maps))  # type: ignore
+
     def split(self) -> Tuple[MutableMapping, "ChainMap"]:
         return self._maps[0], ChainMap(*self._maps[1:])
 
@@ -610,7 +614,16 @@ class Generate:
             return False
 
 
-def substitute_env(contents, env=None):
+def taketwo(seq):
+    last = None
+    for i, x in enumerate(seq):
+        if (i + 1) % 2 == 0:
+            yield last, x
+        else:
+            last = x
+
+
+def substitute_env(contents, env=None, preserve_missing=False):
     """
     Replace ${NAME} or ${NAME:default value} with the value of the environment variable $NAME
     Use \${NAME} to ignore
@@ -620,15 +633,18 @@ def substitute_env(contents, env=None):
 
     def replace(m):
         if m.group(1):  # \ found
-            return m.group(0)[len(m.group(1)) - 1 or 1 :]
-        default_value = m.group(3)[1:] if m.group(3) else ""
+            return m.group(0)[len(m.group(1)) - 1 or 1:]
         for name in m.group(2).split("|"):
             if name in env:
                 value = env[name]
                 if callable(value):
                     return value()
                 return value
-        return default_value
+        # can't resolve, use default
+        if preserve_missing:
+            return m.group(0)
+        else:  # return the default value or an empty ""
+            return m.group(3)[1:] if m.group(3) else ""
 
     return re.sub(r"(\\+)?\$\{([\w|]+)(\:.+?)?\}", replace, contents)
 
@@ -743,8 +759,13 @@ required_envvars = [
     "UNFURL_APPROVE",
     "UNFURL_MOCK_DEPLOY",
     "UNFURL_LOGFILE",
-    "UNFURL_VAULT_SKIP_DECRYPT",
+    "UNFURL_SKIP_VAULT_DECRYPT",
+    "UNFURL_SKIP_UPSTREAM_CHECK",
     "UNFURL_VALIDATION_MODE",
+    "UNFURL_PACKAGE_RULES",
+    "UNFURL_LOG_FORMAT",
+    "UNFURL_RAISE_LOGGING_EXCEPTIONS",
+    "PY_COLORS",
 ]
 # hack for sphinx ext documentedlist
 _sphinx_envvars = [(i,) for i in required_envvars]
