@@ -1,17 +1,19 @@
 # Copyright (c) 2020 Adam Souzis
 # SPDX-License-Identifier: MIT
+import io
 import os.path
 import sys
 import codecs
 import json
 import os
-from typing import Union, Tuple, List
-import six
+from typing import Optional, TextIO, Union, Tuple, List, cast, TYPE_CHECKING, Dict
 import urllib
 from urllib.parse import urljoin, urlsplit
 import ssl
 import certifi
 
+if TYPE_CHECKING:
+    from .manifest import Manifest
 
 pathname2url = urllib.request.pathname2url
 from jsonschema import RefResolver
@@ -42,7 +44,8 @@ from .merge import (
     _cache_anchors,
     restore_includes,
 )
-from .repo import is_url_or_git_path
+from .repo import RepoView, GitRepo, is_url_or_git_path, split_git_url
+from .packages import resolve_package
 from .logs import getLogger
 from toscaparser.common.exception import URLException, ExceptionCollector
 from toscaparser.utils.gettextutils import _
@@ -50,14 +53,41 @@ import toscaparser.imports
 from toscaparser.repositories import Repository
 
 from ansible.parsing.vault import VaultLib, VaultSecret
+from ansible.parsing.yaml.objects import AnsibleMapping
+from ansible.parsing.yaml.loader import AnsibleLoader, AnsibleConstructor
 from ansible.utils.unsafe_proxy import AnsibleUnsafeText, AnsibleUnsafeBytes
+from time import perf_counter
 
 logger = getLogger("unfurl")
 
+yaml_perf = 0.0
 
 CLEARTEXT_VAULT = VaultLib(secrets=[("cleartext", b"")])
 
 
+def yaml_dict_type(readonly: bool) -> type:
+    if readonly:
+        return AnsibleMapping
+    else:
+        return CommentedMap
+
+
+def load_yaml(yaml, stream, path=None, readonly: bool = False):
+    global yaml_perf
+    start_time = perf_counter()
+    if not readonly:
+        if path and isinstance(stream, str):
+            stream = io.StringIO(stream)
+            stream.name = path
+        doc = yaml.load(stream)
+    else:
+        loader = AnsibleLoader(stream, path, yaml.constructor.vault.secrets)
+        loader.vault = yaml.constructor.vault
+        doc = loader.get_single_data()
+    yaml_perf += (perf_counter() - start_time)
+    return doc
+
+
 def _use_clear_text(vault):
     clear_id = CLEARTEXT_VAULT.secrets[0][0]
     return vault.secrets and all(s[0] == clear_id for s in vault.secrets)
@@ -103,7 +133,7 @@ def represent_sensitive_bytes(dumper, data):
 
 def _construct_vault(constructor, node, tag):
     value = constructor.construct_scalar(node)
-    if os.getenv("UNFURL_VAULT_SKIP_DECRYPT"):
+    if getattr(constructor.vault, "skip_decode", False):
         return sensitive_str.redacted_str
     if not constructor.vault.secrets:
         raise ConstructorError(
@@ -140,6 +170,11 @@ def construct_vaultbinary(constructor, node):
     return sensitive_bytes(cleartext)
 
 
+AnsibleConstructor.add_constructor("!vault", construct_vault)
+AnsibleConstructor.add_constructor("!vault-json", construct_vaultjson)
+AnsibleConstructor.add_constructor("!vault-binary", construct_vaultbinary)
+
+
 def make_yaml(vault=None):
     if not vault:
         vault = VaultLib(secrets=None)
@@ -172,19 +207,23 @@ yaml = make_yaml()
 cleartext_yaml = make_yaml(CLEARTEXT_VAULT)
 
 
-def make_vault_lib(passwordBytes, vaultId="default"):
+class UnfurlVaultLib(VaultLib):
+    skip_decode = False
+
+
+def make_vault_lib(passwordBytes: Union[str, bytes], vaultId="default") -> Optional[VaultLib]:
     if passwordBytes:
-        if isinstance(passwordBytes, six.string_types):
+        if isinstance(passwordBytes, str):
             passwordBytes = to_bytes(passwordBytes)
         vault_secrets = [(vaultId, VaultSecret(passwordBytes))]
         return VaultLib(secrets=vault_secrets)
     return None
 
 
-def make_vault_lib_ex(secrets: List[Tuple[str, Union[str, bytes]]]):
+def make_vault_lib_ex(secrets: List[Tuple[str, Union[str, bytes]]]) -> VaultLib:
     vault_secrets = []
     for vaultId, passwordBytes in secrets:
-        if isinstance(passwordBytes, six.string_types):
+        if isinstance(passwordBytes, str):
             passwordBytes = to_bytes(passwordBytes)
         vault_secrets.append((vaultId, VaultSecret(passwordBytes)))
     return VaultLib(secrets=vault_secrets)
@@ -198,26 +237,28 @@ def urlopen(url):
 
 _refResolver = RefResolver("", None)
 
+GetURLType = Optional[Tuple[str, bool, Optional[str]]]
+
 
 class ImportResolver(toscaparser.imports.ImportResolver):
-    def __init__(self, manifest, ignoreFileNotFound=False, expand=False, config=None):
+    def __init__(self, manifest: "Manifest", ignoreFileNotFound=False, expand=False, config=None):
         self.manifest = manifest
+        self.readonly = bool(manifest.localEnv and manifest.localEnv.readonly)
         self.ignoreFileNotFound = ignoreFileNotFound
-        self.loader = manifest.loader
+        self.yamlloader = manifest.loader
         self.expand = expand
         self.config = config or {}
+        self.repo_refs: Dict[str, GitRepo] = {}
 
     def __getstate__(self):
         state = self.__dict__.copy()
-        state["loader"] = None
+        state["yamlloader"] = None
         return state
 
-    def get_repository(self, name, tpl):
-        if self.manifest:
-            # don't create another Repository instance
-            repo_view = self.manifest.repositories.get(name)
-            if repo_view:
-                return repo_view.repository
+    def get_repository(self, name: str, tpl: dict) -> Repository:
+        # don't create another Repository instance
+        if name in self.manifest.repositories:
+            return self.manifest.repositories[name].repository
 
         if isinstance(tpl, dict) and "url" in tpl:
             url = tpl["url"]
@@ -228,7 +269,7 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                 url = "ssh://" + url.replace(":", "/", 1)
             tpl["url"] = url
 
-        if self.manifest and tpl.get('credential'):
+        if tpl.get('credential'):
             credential = tpl['credential']
             # support expressions to resolve credential secrets
             if self.manifest.rootResource:
@@ -236,92 +277,110 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                 tpl['credential'] = map_value(credential, self.manifest.rootResource)
             elif self.manifest.localEnv:
                 # we're including or importing before we finished initializing
-                # update os.environ so get_env works
                 context = self.manifest.localEnv.get_context(self.config.get("environment"))
-                if context.get("variables"):
-                    os.environ.update(filter_env(context["variables"]))
-                tpl['credential'] = self.manifest.localEnv.map_value(credential)
+                tpl['credential'] = self.manifest.localEnv.map_value(credential, context.get("variables"))
 
         return Repository(name, tpl)
 
-    def get_url(self, importLoader, repository_name, file_name, isFile=None):
-        # returns url or path, isFile, fragment
-        importLoader.stream = None
-        if repository_name:
-            path = self.get_repository_url(importLoader, repository_name)
-            file_name, sep, fragment = file_name.partition("#")
-            if path.startswith("file:"):
-                path = path[5:]
-                if path.startswith("//"):
-                    path = path[2:]
-                isFile = True
-        else:
-            url_info = super().get_url(importLoader, repository_name, file_name, isFile)
-            if not url_info:
-                return url_info
-            file_name = ""
-            path, isFile, fragment = url_info
+    def _get_bare_path(self, path, repo, filePath, revision, file_name, importsLoader):
+        # # XXX support empty filePath or when filePath is a directory -- need to checkout the tree
+        if not filePath:
+            raise UnfurlError(
+                f"local repository for '{path}' can not checkout revision '{revision}'"
+            )
+        elif repo.repo.rev_parse(revision + ":" + filePath).type != "blob":
+            raise UnfurlError(
+                f"can't retrieve '{filePath}' with revision '{revision}' from local repository for '{path}'"
+            )
+        url = "git-ref:" + repo.url
+        self.repo_refs[url] = repo
+        return f"{url}#{revision}:{filePath}"
 
-        if not isFile or is_url_or_git_path(path):
-            if not self.manifest:
-                raise UnfurlError("Could not resolve git URL: " + path)
-            # only support urls to git repos for now
-            repo_view = self.manifest.repositories.get(repository_name)
+    def _get_repository_path(self, importsLoader: toscaparser.imports.ImportsLoader,
+                             repository_name: str) -> Tuple[str, bool, Optional["RepoView"]]:
+        path = cast(str, self.get_repository_url(importsLoader, repository_name))
+        if path.startswith("file:"):
+            path = path[5:]
+            if path.startswith("//"):
+                path = path[2:]
+            return path, True, None
+        else:
+            repoview = self.manifest.repositories.get(repository_name)
+            if not repoview:
+                repository = self.get_repository(repository_name, importsLoader.repositories[repository_name])
+                repoview = self.manifest.add_repository(None, repository, "")
+            assert repoview
+            if repoview.package is None:
+                # need to resolve if its a package
+                # if repoview.repository references a package, set the repository's url
+                # and register this reference with the package
+                # might raise error if version conflict
+                resolve_package(repoview, self.manifest.packages, self.manifest.package_specs)
+            return repoview.url, False, repoview
+
+    def get_url(self, importsLoader: toscaparser.imports.ImportsLoader,
+                repository_name: Optional[str], file_name: str, isFile: Optional[bool] = None) -> GetURLType:
+        # called by ImportsLoader when including or importing a file
+        # or resolving path to an artifact in a repository (see ToscaSpec.get_repository_path)
+        # returns url or path, isFile, fragment
+        fragment: Optional[str] = None
+        repo_view = None
+        if repository_name:
+            file_name, sep, fragment = file_name.partition("#")
+            path, isFile, repo_view = self._get_repository_path(importsLoader, repository_name)
+            # if not isFile path will be git url possibly including revision in fragment
             if repo_view and repo_view.repo:
                 return os.path.join(repo_view.working_dir, file_name), True, fragment
+        else:
+            url_info = cast(GetURLType, super().get_url(importsLoader, None, file_name, isFile))
+            if not url_info:
+                return url_info
+            path, isFile, fragment = url_info
+            file_name = ""  # was included in path
 
+        if not isFile or is_url_or_git_path(path):
+            # only support urls to git repos for now
+            # this may clone a remote git repo
             repo, filePath, revision, bare = self.manifest.find_repo_from_git_url(
-                path, isFile, importLoader
+                path, isFile, importsLoader
             )
             if not repo:
                 raise UnfurlError("Could not resolve git URL: " + path)
-            isFile = True
             if bare:
-                # XXX support empty filePath or when filePath is a directory -- need to checkout the tree
-                if not filePath:
-                    raise UnfurlError(
-                        f"local repository for '{path}' can not checkout revision '{revision}'"
-                    )
-                elif repo.repo.rev_parse(revision + ":" + filePath).type != "blob":
-                    raise UnfurlError(
-                        f"can't retrieve '{filePath}' with revision '{revision}' from local repository for '{path}'"
-                    )
-                importLoader.stream = six.StringIO(repo.show(filePath, revision))
-                path = os.path.join(filePath, file_name or "")
+                path = self._get_bare_path(path, repo, filePath, revision, file_name, importsLoader)
+                return path, False, fragment
             else:
-                path = os.path.join(repo.working_dir, filePath, file_name or "").rstrip(
-                    "/"
-                )
                 if repo_view:
                     assert not repo_view.repo
-                    repo_view.repo = repo
-                elif repository_name:
-                    # first time seeing this repository
-                    repository = self.get_repository(repository_name, importLoader.repositories[repository_name])
-                    self.manifest.add_repository(repo, repository, filePath)
+                    repo_view.set_repo_and_path(repo, filePath)
+                path = os.path.join(repo.working_dir, filePath, file_name or "").rstrip("/")
         elif file_name:
             path = os.path.join(path, file_name)
-        return path, isFile, fragment
 
-    def load_yaml(self, importLoader, path, isFile=True, fragment=None):
+        # always a local file path at this point
+        return path, True, fragment
+
+    def load_yaml(self, importsLoader, path, isFile=True, fragment=None):
         try:
             logger.trace(
                 "attempting to load YAML %s: %s", "file" if isFile else "url", path
             )
             originalPath = path
-            f = importLoader.stream
 
-            if not f:
+            if path.startswith("git-ref:"):
+                url, filePath, revision = split_git_url(path)
+                f: TextIO = io.StringIO(self.repo_refs[url].show(filePath, revision))
+            else:
                 try:
                     if isFile:
                         ignoreFileNotFound = self.ignoreFileNotFound
                         if ignoreFileNotFound and not os.path.isfile(path):
                             return None
 
-                        if self.loader:
+                        if self.yamlloader:
                             # show == True if file was decrypted
-                            contents, show = self.loader._get_file_contents(path)
-                            f = six.StringIO(codecs.decode(contents))
+                            contents, show = self.yamlloader._get_file_contents(path)
+                            f = io.StringIO(codecs.decode(contents))
                         else:
                             f = codecs.open(path, encoding="utf-8", errors="strict")
                     else:
@@ -344,27 +403,28 @@ class ImportResolver(toscaparser.imports.ImportResolver):
                             )
                         ) % {
                             "path": path,
-                            "code": e.code,
-                        }  # type: ignore
+                            "code": e.code,  # type: ignore
+                        }
                         ExceptionCollector.appendException(URLException(what=msg))
                         return
                     else:
                         raise
             with f:
-                doc = yaml.load(f.read())
-                if isinstance(doc, CommentedMap):
+                contents = f.read()
+                doc = load_yaml(yaml, contents, path, self.readonly)
+                yaml_dict = yaml_dict_type(self.readonly)
+                if isinstance(doc, yaml_dict):
                     if self.expand:
                         includes, doc = expand_doc(
-                            doc, cls=make_map_with_base(doc, get_base_dir(path))
+                            doc, cls=make_map_with_base(doc, get_base_dir(path), yaml_dict)
                         )
                         doc.includes = includes
                     doc.path = path
-
             if fragment and doc:
                 return _refResolver.resolve_fragment(doc, fragment)
             else:
                 return doc
-        except:
+        except Exception:
             if path != originalPath:
                 msg = f'Could not load "{path}" (originally "{originalPath}")'
             else:
@@ -381,12 +441,14 @@ class YamlConfig:
         schema=None,
         loadHook=None,
         vault=None,
+        readonly=False
     ):
         try:
             self._yaml = None
             self.vault = vault
             self.path = None
             self.schema = schema
+            self.readonly = bool(readonly)
             self.lastModified = None
             if path:
                 self.path = os.path.abspath(path)
@@ -399,17 +461,13 @@ class YamlConfig:
             else:
                 self.path = None
 
-            if isinstance(config, six.string_types):
-                if self.path:
-                    # set name on a StringIO so parsing error messages include the path
-                    config = six.StringIO(config)
-                    config.name = self.path
-                self.config = self.yaml.load(config)
+            if isinstance(config, str):
+                self.config = load_yaml(self.yaml, config, self.path, self.readonly)
             elif isinstance(config, dict):
                 self.config = CommentedMap(config.items())
             else:
                 self.config = config
-            if not isinstance(self.config, CommentedMap):
+            if not isinstance(self.config, dict):
                 raise UnfurlValidationError(
                     f'invalid YAML document with contents: "{self.config}"'
                 )
@@ -418,12 +476,13 @@ class YamlConfig:
             self._cachedDocIncludes = {}
             # schema should include defaults but can't validate because it doesn't understand includes
             # but should work most of time
-            self.config.loadTemplate = self.load_include
+            self.config.loadTemplate = self.load_include  # type: ignore
             self.loadHook = loadHook
 
             self.baseDirs = [self.get_base_dir()]
+            yaml_dict = yaml_dict_type(self.readonly)
             self.includes, expandedConfig = expand_doc(
-                self.config, cls=make_map_with_base(self.config, self.baseDirs[0])
+                self.config, cls=make_map_with_base(self.config, self.baseDirs[0], yaml_dict)
             )
             self.expanded = expandedConfig
             errors = schema and self.validate(expandedConfig)
@@ -434,7 +493,7 @@ class YamlConfig:
                 )
             else:
                 self.valid = not not errors
-        except Exception as e:
+        except Exception:
             if self.path:
                 msg = f"Unable to load yaml config at {self.path}"
             else:
@@ -465,7 +524,7 @@ class YamlConfig:
             f = open(path, "r")
 
         with f:
-            config = self.yaml.load(f)
+            config = load_yaml(self.yaml, f, path, self.readonly)
         if fragment and config:
             return path, _refResolver.resolve_fragment(config, fragment)
         return path, config
@@ -486,7 +545,11 @@ class YamlConfig:
             raise UnfurlError(f"Error saving {self.path}", True)
 
     def save(self):
-        output = six.StringIO()
+        if self.readonly:
+            raise UnfurlError(
+                f'Can not save "{self.path}", it is set to readonly'
+            )
+        output = io.StringIO()
         self.dump(output)
         if self.path:
             if self.lastModified:
@@ -521,7 +584,7 @@ class YamlConfig:
         return state
 
     def validate(self, config):
-        if isinstance(self.schema, six.string_types):
+        if isinstance(self.schema, str):
             # assume its a file path
             path = self.schema
             with open(path) as fp:
@@ -547,7 +610,11 @@ class YamlConfig:
 
     def save_include(self, key):
         path, template = self._cachedDocIncludes[key]
-        output = six.StringIO()
+        if self.readonly:
+            raise UnfurlError(
+                f'Can not save include at "{path}", config "{self.path}" is readonly'
+            )
+        output = io.StringIO()
         try:
             self.yaml.dump(template, output)
         except Exception:
@@ -625,8 +692,8 @@ class YamlConfig:
                 newBaseDir = baseDir
             else:
                 newBaseDir = os.path.dirname(path)
-            if isinstance(template, CommentedMap):
-                template.base_dir = newBaseDir
+            if isinstance(template, dict):
+                template.base_dir = newBaseDir  # type: ignore
             _cache_anchors(self.config._anchorCache, template)
         except Exception:
             raise UnfurlError(
