@@ -3,6 +3,7 @@
 """Loads and saves a ensemble manifest.
 """
 from __future__ import absolute_import
+from typing import Dict, List, Optional, Tuple
 import six
 import sys
 from collections.abc import MutableSequence, Mapping
@@ -15,23 +16,21 @@ from . import DefaultNames
 from .util import UnfurlError, to_yaml_text, filter_env
 from .merge import patch_dict, intersect_dict
 from .yamlloader import YamlConfig, make_yaml
-from .result import serialize_value, ChangeRecord
+from .result import serialize_value
 from .support import ResourceChanges, Defaults, Status
 from .localenv import LocalEnv
 from .lock import Lock
-from .manifest import Manifest, relabel_dict
+from .manifest import Manifest, relabel_dict, ChangeRecordRecord
 from .tosca import ArtifactSpec, find_env_vars
 from .runtime import TopologyInstance
 from .eval import map_value
 from .planrequests import create_instance_from_spec
-
+from .logs import getLogger
 from ruamel.yaml.comments import CommentedMap
 from codecs import open
 from ansible.parsing.dataloader import DataLoader
 
-import logging
-
-logger = logging.getLogger("unfurl")
+logger = getLogger("unfurl")
 
 _basepath = os.path.abspath(os.path.dirname(__file__))
 
@@ -178,7 +177,7 @@ class ReadOnlyManifest(Manifest):
     """Loads an ensemble from a manifest but doesn't instantiate the instance model."""
 
     def __init__(
-        self, manifest=None, path=None, validate=True, localEnv=None, vault=None
+        self, manifest=None, path: Optional[str]=None, validate=True, localEnv=None, vault=None
     ):
         path = path or localEnv and localEnv.manifestPath
         if path:
@@ -191,6 +190,7 @@ class ReadOnlyManifest(Manifest):
             os.path.join(_basepath, "manifest-schema.json"),
             self.load_yaml_include,
             vault,
+            localEnv and localEnv.readonly
         )
         if self.manifest.path:
             logger.debug("loaded ensemble manifest at %s", self.manifest.path)
@@ -209,8 +209,8 @@ class ReadOnlyManifest(Manifest):
         self._update_repositories(manifest)
 
     @property
-    def uris(self):
-        uris = []
+    def uris(self) -> List[str]:
+        uris: List[str] = []
         if "metadata" in self.manifest.config:
             uri = self.metadata.get("uri")
             uris = self.metadata.get("aliases") or []
@@ -268,7 +268,7 @@ def clone(localEnv, destPath):
 
 
 class YamlManifest(ReadOnlyManifest):
-    _operationIndex = None
+    _operationIndex: Optional[Dict[Tuple[str, str], str]] = None
     lockfilepath = None
     lockfile = None
 
@@ -282,11 +282,11 @@ class YamlManifest(ReadOnlyManifest):
         skip_validation=False,
     ):
         super().__init__(manifest, path, validate, localEnv, vault)
+        self.validate = not skip_validation  # see AttributeManager.validate
         # instantiate the tosca template
         manifest = self.manifest.expanded
         if self.manifest.path:
             self.lockfilepath = self.manifest.path + ".lock"
-
         spec = manifest.get("spec", {})
         more_spec = self._load_context(self.context, localEnv)
         deployment_blueprint = self.context.get("deployment_blueprint")
@@ -294,10 +294,10 @@ class YamlManifest(ReadOnlyManifest):
             manifest.get("spec", {}).get("deployment_blueprints") or {}
         )
         if deployment_blueprint:
-            self._add_deployment_blueprint_template(
+            if self._add_deployment_blueprint_template(
                 deployment_blueprints, deployment_blueprint, more_spec
-            )
-            logger.info('Using deployment blueprint "%s"', deployment_blueprint)
+            ):
+                logger.info('Using deployment blueprint "%s"', deployment_blueprint)
         elif deployment_blueprints:
             logger.warning(
                 "This ensemble contains deployment blueprints but none were specified for use."
@@ -305,7 +305,7 @@ class YamlManifest(ReadOnlyManifest):
         if self.context.get("instances"):
             # add context instances to spec instances but skip ones that are just in there because they were shared
             env_instances = {
-                k: v
+                k: v.copy()
                 for k, v in self.context["instances"].items()
                 if "imported" not in v
             }
@@ -321,8 +321,8 @@ class YamlManifest(ReadOnlyManifest):
             logger.info(msg)
 
         status = manifest.get("status", {})
-        self.changeLogPath = manifest.get("jobsLog")
-        self.jobsFolder = manifest.get("jobsFolder", "jobs")
+        self.changeLogPath: str = manifest.get("jobsLog") or ""
+        self.jobsFolder: str = manifest.get("jobsFolder", "jobs")
         if not self.changeLogPath and localEnv and manifest.get("changes") is None:
             # save changes to a separate file if we're in a local environment
             self.changeLogPath = DefaultNames.JobsLog
@@ -336,7 +336,8 @@ class YamlManifest(ReadOnlyManifest):
                 self.imports.add_import(name.rstrip("s"), instance, local_spec)
 
         rootResource = self.create_topology_instance(status)
-
+        for key, val in status.get("instances", {}).items():
+            self.create_node_instance(key, val, rootResource)
         # create an new instances declared in the spec:
         for name, instance in spec.get("instances", {}).items():
             if not rootResource.find_resource(name):
@@ -351,9 +352,12 @@ class YamlManifest(ReadOnlyManifest):
         self, deployment_blueprints, deployment_blueprint, more_spec
     ):
         if deployment_blueprint not in deployment_blueprints:
-            raise UnfurlError(
-                f"Can not find requested deployment blueprint: '{deployment_blueprint}' is missing from the ensemble."
-            )
+            msg = f"Can not find requested deployment blueprint: '{deployment_blueprint}' is missing from the ensemble."
+            if self.validate:
+                raise UnfurlError(msg)
+            else:
+                logger.error(msg)
+            return False
         deployment_blueprint_tpl = deployment_blueprints[deployment_blueprint]
         resource_templates = deployment_blueprint_tpl.get("resource_templates")
         resourceTemplates = deployment_blueprint_tpl.get("resourceTemplates")
@@ -372,6 +376,7 @@ class YamlManifest(ReadOnlyManifest):
         if resource_templates:
             node_templates = more_spec["topology_template"]["node_templates"]
             self._load_resource_templates(resource_templates, node_templates, False)
+        return True
 
     def _configure_root(self, rootResource):
         rootResource.imports = self.imports
@@ -379,10 +384,11 @@ class YamlManifest(ReadOnlyManifest):
             self.manifest.vault and self.manifest.vault.secrets
         ):  # setBaseDir() may create a new templar
             rootResource._templar._loader.set_vault_secrets(self.manifest.vault.secrets)
-        rootResource.envRules = self.context.get("variables") or CommentedMap()
         if not self.localEnv:
             return
 
+        if self.localEnv.overrides.get("UNFURL_SKIP_VAULT_DECRYPT"):
+            return
         # use the password associated with the project the repository appears in.
         repos = set(self.repositories.values())
         project = self.localEnv.project or self.localEnv.homeProject
@@ -404,15 +410,37 @@ class YamlManifest(ReadOnlyManifest):
             repository.load_secrets(rootResource._templar._loader)
             repository.yaml = self.yaml
 
-    def create_topology_instance(self, status):
+    def _set_root_environ(self) -> None:
+        # We need to set the environment as early as possible but not too early
+        # Each ensemble maintains its own set of environment variables that is used when evaluating expressions (e.g get_env)
+        # but os.environ is only set to this when a task is active.
+        root = self.rootResource
+        assert root
+        rules = self.context.get("variables") or CommentedMap()
+        for rel in root.requirements:
+            rules.update(rel.merge_props(find_env_vars, True))
+        rules = serialize_value(
+            map_value(rules, root), resolveExternal=True
+        )
+        root._environ = filter_env(rules, os.environ)
+        paths = self.localEnv and self.localEnv.get_paths()
+        if paths:
+            path = os.pathsep.join(paths)
+            if path not in root._environ["PATH"]:  # avoid setting twice
+                root._environ["PATH"] = (
+                    path + os.pathsep + root._environ.get("PATH", "")
+                )
+                logger.debug("PATH set to %s", root._environ["PATH"])
+
+    def create_topology_instance(self, status: dict) -> TopologyInstance:
         """
         If an instance of the toplogy is recorded in status, load it,
         otherwise create a new resource using the the topology as its template
         """
         # XXX use the substitution_mapping (3.8.12) represent the resource
-        template = self.tosca.topology
         operational = self.load_status(status)
-        root = TopologyInstance(template, operational)
+        root = TopologyInstance(self.tosca and self.tosca.topology, operational)
+        root.attributeManager = self
         if os.environ.get("UNFURL_WORKDIR"):
             root.set_base_dir(os.environ["UNFURL_WORKDIR"])
         elif not self.path:
@@ -420,27 +448,9 @@ class YamlManifest(ReadOnlyManifest):
         else:
             root.set_base_dir(self.get_base_dir())
 
-        # We need to set the environment as early as possible but not too early
-        # and only once.
-        # Now that we loaded the main manifest and set the root's baseDir
-        # let's do it before we import any other manifests.
-        # But only if we're the main manifest.
-        if not self.localEnv or not self.localEnv.parent:
-            if self.context.get("variables"):
-                env = filter_env(map_value(self.context["variables"], root))
-            else:
-                env = os.environ.copy()
-            for rel in root.requirements:
-                t = lambda datatype: datatype.type == "unfurl.datatypes.EnvVar"
-                env.update(rel.merge_props(find_env_vars, True))
-            intersect_dict(os.environ, env)  # remove keys not in env
-            os.environ.update(env)
-            paths = self.localEnv and self.localEnv.get_paths()
-            if paths:
-                os.environ["PATH"] = (
-                    os.pathsep.join(paths) + os.pathsep + os.environ.get("PATH", [])
-                )
-                logger.debug("PATH set to %s", os.environ["PATH"])
+        # need to set rootResource before createNodeInstance() is called
+        self.rootResource = root
+        self._set_root_environ()
 
         # self.load_external_ensemble("localhost", tpl) # declared in templates/home/unfurl.yaml.j2
         importsSpec = self.context.get("external", {})
@@ -448,10 +458,6 @@ class YamlManifest(ReadOnlyManifest):
         for name, value in importsSpec.items():
             self.load_external_ensemble(name, value)
 
-        # need to set rootResource before createNodeInstance() is called
-        self.rootResource = root
-        for key, val in status.get("instances", {}).items():
-            self.create_node_instance(key, val, root)
         return root
 
     def _load_resource_templates(self, templates, node_templates, virtual):
@@ -462,7 +468,9 @@ class YamlManifest(ReadOnlyManifest):
         for name, tpl in templates.items():
             # hacky way to exclude being part of the deployment plan and the manifest's status
             if virtual:
-                tpl.setdefault("directives", []).append("virtual")
+                directives = tpl.setdefault("directives", [])
+                if "virtual" not in directives:
+                    directives.append("virtual")
             node_templates[name] = tpl
 
     def _load_context(self, context, localEnv):
@@ -495,7 +503,7 @@ class YamlManifest(ReadOnlyManifest):
             )
 
         if "project" in location:
-            importedManifest = self.localEnv.get_external_manifest(location)
+            importedManifest = self.localEnv and self.localEnv.get_external_manifest(location)
             if not importedManifest:
                 raise UnfurlError(
                     f"Can not import external ensemble '{name}': can't find project '{location['project']}'"
@@ -513,11 +521,10 @@ class YamlManifest(ReadOnlyManifest):
                 parent=self.localEnv,
                 override_context=location.get("environment", ""),
             )
-            if self.path and os.path.abspath(self.path) == os.path.abspath(
-                localEnv.manifestPath
-            ):
+            if self.is_path_to_self(localEnv.manifestPath):
                 # don't import self (might happen when context is shared)
                 return
+            logger.verbose("loading external ensemble at %s", localEnv.manifestPath)
             importedManifest = localEnv.get_manifest()
 
         uri = value.get("uri")
@@ -539,8 +546,7 @@ class YamlManifest(ReadOnlyManifest):
         self.imports.add_import(name, resource, value)
         self._importedManifests[id(root)] = importedManifest
 
-    def load_changes(self, changes, changeLogPath):
-        # self.changeSets[changeid => ChangeRecords]
+    def load_changes(self, changes: Optional[List[dict]], changeLogPath: str) -> bool:
         if changes is not None:
             self.changeSets = {
                 c.changeId: c
@@ -553,7 +559,7 @@ class YamlManifest(ReadOnlyManifest):
                     self.changeSets = {
                         c.changeId: c
                         for c in (
-                            ChangeRecord(parse=line.strip())
+                            ChangeRecordRecord(parse=line.strip())
                             for line in f.readlines()
                             if not line.strip().startswith("#")
                         )
@@ -580,11 +586,11 @@ class YamlManifest(ReadOnlyManifest):
         else:
             # ok if we race here, we'll just raise an error
             self.lockfile = open(self.lockfilepath, "xb", buffering=0)
-            self.lockfile.write(bytes(str(os.getpid()), "ascii"))
+            self.lockfile.write(bytes(str(os.getpid()), "ascii"))  # type: ignore
             return True
 
     def unlock(self):
-        if self.lockfile:
+        if self.lockfile and self.lockfilepath:
             # unlink first to avoid race (this will fail on Windows)
             os.unlink(self.lockfilepath)
             self.lockfile.close()
@@ -592,27 +598,25 @@ class YamlManifest(ReadOnlyManifest):
             return True
         return False
 
-    def find_last_operation(self, target, operation):
+    def find_last_operation(self, target, operation) -> Optional[ChangeRecordRecord]:
         if self._operationIndex is None:
-            operationIndex = {}
+            operationIndex: Dict[Tuple[str, str], str] = {}
             if self.changeSets:
                 # add list() for 3.7
                 for change in reversed(list(self.changeSets.values())):
-                    if not hasattr(change, "target") or not hasattr(
-                        change, "operation"
-                    ):
+                    if not change.target or not change.operation:
                         continue
                     key = (change.target, change.operation)
                     last = operationIndex.setdefault(key, change.changeId)
                     if last < change.changeId:
-                        operationIndex[key] = change
+                        operationIndex[key] = change.changeId
             self._operationIndex = operationIndex
         changeId = self._operationIndex.get((target, operation))
-        if changeId is not None:
+        if changeId is not None and self.changeSets:
             return self.changeSets[changeId]
         return None
 
-    def save_entity_instance(self, resource):
+    def save_entity_instance(self, resource) -> Tuple[str, Dict]:
         status = CommentedMap()
         status["template"] = resource.template.get_uri()
 
@@ -632,17 +636,18 @@ class YamlManifest(ReadOnlyManifest):
             status["protected"] = resource.protected
         return (resource.name, status)
 
-    def save_requirement(self, resource):
+    def save_requirement(self, resource) -> Optional[Dict[str, Dict]]:
         if not resource.last_change and (
             not resource.local_status or resource.local_status <= Status.ok
+            or resource.local_status == Status.pending
         ):
             # no reason to serialize requirements that haven't been instantiated
             return None
         name, status = self.save_entity_instance(resource)
         status["capability"] = resource.parent.key
-        return [{name: status}]
+        return {name: status}
 
-    def _save_entity_if_instantiated(self, resource, checkstatus=True):
+    def _save_entity_if_instantiated(self, resource, checkstatus=True) -> Optional[Tuple[str, Dict]]:
         if "virtual" in resource.template.directives:
             return None
         if not resource.last_change and (
@@ -667,7 +672,7 @@ class YamlManifest(ReadOnlyManifest):
             return ret
         name, status = ret
 
-        if self.tosca.discovered and resource.template.name in self.tosca.discovered:
+        if self.tosca and self.tosca.discovered and resource.template.name in self.tosca.discovered:
             discovered[resource.template.name] = self.tosca.discovered[
                 resource.template.name
             ]
@@ -713,6 +718,7 @@ class YamlManifest(ReadOnlyManifest):
 
     def save_root_resource(self, discovered):
         resource = self.rootResource
+        assert resource
         status = CommentedMap()
 
         # record the input and output values
@@ -804,7 +810,7 @@ class YamlManifest(ReadOnlyManifest):
             changes = list(
                 map(lambda t: save_task(t, exclude_result), job.workDone.values())
             )
-            if self.changeLogPath:
+            if self.changeLogPath and self.path is not None:
                 self.manifest.config["jobsLog"] = self.changeLogPath
 
                 jobLogPath = self.get_job_log_path(jobRecord["startTime"])
@@ -847,7 +853,11 @@ class YamlManifest(ReadOnlyManifest):
                 message = job.message
             else:
                 message = self.get_default_commit_message()
-            self.commit(message, True)
+
+            # only commit the ensemble repository:
+            ensembleRepo = self.repositories["self"]
+            if ensembleRepo.is_dirty():
+                ensembleRepo.commit(message, True)
 
     def get_default_commit_message(self):
         jobRecord = self.manifest.config.get("lastJob")
@@ -856,21 +866,21 @@ class YamlManifest(ReadOnlyManifest):
         else:
             return "Commit by Unfurl"
 
-    def get_repo_status(self, dirty=False):
+    def get_repo_status(self, dirty=False) -> str:
         return "".join([r.get_repo_status(dirty) for r in self.repositories.values()])
 
-    def add_all(self):
+    def add_all(self) -> None:
         for repository in self.repositories.values():
             if not repository.readOnly and repository.is_dirty():
                 repository.add_all()
 
-    def commit(self, msg, addAll):
+    def commit(self, msg: str, add_all: bool = False) -> int:
         committed = 0
         for repository in self.repositories.values():
             if repository.repo == self.repo:
                 continue
             if not repository.readOnly and repository.is_dirty():
-                retVal = repository.commit(msg, addAll)
+                retVal = repository.commit(msg, add_all)
                 committed += 1
                 logger.info(
                     "committed %s to %s: %s", retVal, repository.working_dir, msg
@@ -880,7 +890,7 @@ class YamlManifest(ReadOnlyManifest):
         #    (note: endCommit will be omitted as changes.yaml isn't updated)
         ensembleRepo = self.repositories["self"]
         if ensembleRepo.is_dirty():
-            retVal = ensembleRepo.commit(msg, addAll)
+            retVal = ensembleRepo.commit(msg, add_all)
             committed += 1
             logger.info("committed %s to %s: %s", retVal, ensembleRepo.working_dir, msg)
 
@@ -940,16 +950,17 @@ class YamlManifest(ReadOnlyManifest):
                     if key.startswith("digest"):
                         attrs[key] = change[key]
                 attrs["summary"] = change["summary"]
-                line = ChangeRecord.format_log(change["changeId"], attrs)
+                line = ChangeRecordRecord.format_log(change["changeId"], attrs)
                 f.write(line)
 
     def save_change_log(self, jobRecord, newChanges):
         try:
             changelog = CommentedMap()
             fullPath = self.get_job_log_path(jobRecord["startTime"])
-            changelog["manifest"] = os.path.relpath(
-                self.manifest.path, os.path.dirname(fullPath)
-            )
+            if self.manifest.path is not None:
+                changelog["manifest"] = os.path.relpath(
+                    self.manifest.path, os.path.dirname(fullPath)
+                )
             changes = itertools.chain([jobRecord], newChanges)
             changelog["changes"] = list(changes)
             output = six.StringIO()
